/**************************************************************************
*   Copyright (C) 2016 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions for spare files
 */
#include <stdio.h>
#include <malloc.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>
#include "common.h"
#include "macros.h"

/*
 * Read a binary file. The structure of the file must be:
 *
 * "SPARSE" or "DENSE "
 * instances (int)
 * inputs (int)
 * classes (int)
 * data[i] (scalar)
 * index[i] (int)   in case of sparse format
 */

#define MAX_LINE_LENGTH 1000000

scalar ReadBinaryDataAndLabels(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, int ***index, bool show_progress, bool sparse_format)
{
    FILE *in;
    char t[10];
    scalar size = 0.0;

    if ((in = fopen(file, "rb")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open input file");

    // Read type of file
    fread(t, sizeof(char), 7, in);

    // Read generic data values
    fread(instances, sizeof(int), 1, in);
    fread(inputs, sizeof(int), 1, in);
    fread(classes, sizeof(int), 1, in);

    MALLOC(*data, scalar *, *instances, "ReadBinaryDataAndLabels: *data");
    MALLOC(*labels, int, *instances, "ReadBinaryDataAndLabels: *labels");

    // Sparse or dense?
    if (strncmp(t, "SPARSE", 6) == 0) {
        int c;

        if (sparse_format) {
            size += *instances * (sizeof(scalar *) + sizeof(int *));

            MALLOC(*index, int *, *instances, "ReadBinaryDataAndLabels: *index");

            for (int i = 0; i < *instances; i++) {
                fread(&c, sizeof(int), 1, in);

                MALLOC((*data)[i], scalar, c, "ReadBinaryDataAndLabels: (*data)[i]");
                MALLOC((*index)[i], int, c+1, "ReadBinaryDataAndLabels; (*index)[i]");


                fread((*data)[i], sizeof(scalar), c, in);
                fread((*index)[i], sizeof(int), c, in);
                (*index)[i][c] = -1;

                fread(&((*labels)[i]), sizeof(int), 1, in);

                size += c*sizeof(scalar) + (c+1)*sizeof(int);

                SHOWPROGRESS(show_progress, "Read instances: %*d/%d\r", WIDTH(*instances), i+1, *instances);
            }

            SHOWPROGRESS(show_progress, "%c", '\n');
        } else {
            // Convert to dense format
            int *l_index;
            scalar *sparse;

            MALLOC(l_index, int, *inputs+1, "ReadBinaryDataAndLabels: l_index");
            MALLOC(sparse, scalar, *inputs, "ReadBinaryDataAndLabels: sparse");

            for (int i = 0; i < *instances; i++) {
                fread(&c, sizeof(int), 1, in);

                CALLOC((*data)[i], scalar, *inputs, "ReadBinaryDataAndLabels: (*data)[i]");

                fread(sparse, sizeof(scalar), c, in);
                fread(l_index, sizeof(int), c, in);
                fread(&((*labels)[i]), sizeof(int), 1, in);

                for (int j = 0; j < c; j++)
                    (*data)[i][l_index[j]-1] = sparse[j];
            }

            FREE(sparse);
            FREE(l_index);
        }
    } else if (strncmp(t, "DENSE", 5) == 0) {
        for (int i = 0; i < *instances; i++) {
            MALLOC((*data)[i], scalar, *inputs, "ReadBinaryDataAndLabels: (*data)[i]");

            fread((*data)[i], sizeof(scalar), *inputs, in);
            fread(&((*labels)[i]), sizeof(int), 1, in);

            SHOWPROGRESS(show_progress, "Read instances: %*d/%d\r", WIDTH(*instances), i+1, *instances);
        }

        SHOWPROGRESS(show_progress, "%c", '\n');

        // Convert to sparse format?
        if (sparse_format) {
            scalar *sparse;

            size += *instances * (sizeof(scalar *) + sizeof(int *));

            MALLOC(*index, int *, *instances, "ReadBinaryDataAndLabels: *index");

            for (int i = 0; i < *instances; i++) {
                int c = 0;
                for (int j = 0; j < *inputs; j++)
                    if (fabs((*data)[i][j]) > ZERO)
                        c++;

                MALLOC(sparse, scalar, c, "ReadBinaryDataAndLabels: sparse");
                MALLOC((*index)[i], int, c+1, "ReadBinaryDataAndLabels: (*index)[i]");

                c = 0;
                for (int j = 0; j < *inputs; j++)
                    if (fabs((*data)[i][j]) > ZERO) {
                        sparse[c] = (*data)[i][j];
                        (*index)[i][c] = j+1;
                        c++;
                    }

                size += c*sizeof(scalar) + (c+1)*sizeof(int);

                (*index)[i][c] = -1;
                FREE((*data)[i]);
                (*data)[i]= sparse;
            }
        }

    } else
        OnErrorExit(ERR_FILE, "Unknown format for binary file");

    fclose(in);

    return size/pow(2.0, 30.0);
}
