/***************************************************************************
*   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions.
 */
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <stdbool.h>
#include <unistd.h>
#include "common.h"
#include "macros.h"
#include <string.h>

/*
 * Read a data file. The structure of the file must be:
 * [ONE LABEL] or [SPARSE]
 * $ Number of lines
 * $ Number of inputs
 * $ Number of outputs
 * One pattern per line, variables separated by spaces, tabs or new lines.
 */


// If huge dataset, labels are read, but actual data is not
void ReadDataAndLabels(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, bool huge)
{
    int i, j;
    FILE *f;
    double outputs;
    scalar *next = NULL;
    char line[100];
    bool one_label = false;

    // If binary file exist read it
    char file_name[MAX_FILE_NAME];
    sprintf(file_name, "%s.bin", file);
    if (!access(file_name, R_OK)) {
        ReadBinaryDataAndLabels(data, labels, instances, inputs, classes, file_name, NULL, false, false);
        return;
    }

    // Opens the file and read number of elements
    if ((f = fopen(file, "rt")) == NULL) {
        printf("Dataset %s not found\n", file);
        OnErrorExit(ERR_FILE, "ReadDataAndLabels");
    }

    // Check sparse label
    if (fgets(line, 100, f) == NULL)
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (!strncmp(line, "[SPARSE]", 8)) {
        fclose(f);
        ReadSparseDataAndLabels(data, labels, instances, inputs, classes, file, huge);
        return;
    } else if (!strncmp(line, "[ONE LABEL]", 11)) {
        // One label in[0, K-1] for K classes
        one_label = true;
    } else
        rewind(f);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");

    // Memory allocation
    if (!huge) {
        MALLOC(*data, scalar *, *instances, "ReadDataAndLabels");

        for (i = 0; i < *instances; i++)
            MALLOC((*data)[i], scalar, *inputs, "ReadData: (*data)[i]");
    } else {
        MALLOC(next, scalar, *inputs + *classes, "ReadDataAndLabels");
    }

    MALLOC(*labels, int, *instances, "ReadDataAndLabels");

    // Read data
    for (i = 0; i < *instances; i++) {
        if (!huge)
            next = (*data)[i];

        for (j = 0; j < *inputs; j++)
#ifdef SCALAR_IS_FLOAT
            if (!fscanf(f, "%f", & (next[j])))
#else
            if (!fscanf(f, "%lf", & (next[j])))
#endif
                OnErrorExit(ERR_READ, "ReadDataAndLabels");

        // Read labels
        if (one_label) {
            if (!fscanf(f, "%d", &(*labels)[i]))
                OnErrorExit(ERR_READ, "ReadDataAndLabels");
        } else
            for (int j = 0; j < *classes; j++) {
                if (!fscanf(f, "%lf", &outputs))
                    OnErrorExit(ERR_READ, "ReadDataAndLabels");
                if (outputs == 1.0)
                    (*labels)[i] = j;
            }
    }

    fclose(f);

    if (huge)
        FREE(next);
}

int ReadNextInstanceAndLabel(scalar *instance, int *label, int *inputs, int *classes, FILE *test, bool *first_time)
{
    int instances;
    char line[100];

    // If first time read info
    if (*first_time) {
        rewind(test);

        // Check sparse label
        if (fgets(line, 100, test) == NULL)
            OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");
        if (strncmp(line, "[ONE LABEL]", 11))
            OnErrorExit(1, "Read next instance and label function only for one label datasets");

        // Read number of lines and columns.
        if (!fscanf(test, "$ %d\n", &instances))
            OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");
        if (!fscanf(test, "$ %d\n", inputs))
            OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");
        if (!fscanf(test, "$ %d\n", classes))
            OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");

        *first_time = false;

        // First time return number of instances and read nothing else
        return instances;
    }

    // Read instance
    for (int j = 0; j < *inputs; j++)
#ifdef SCALAR_IS_FLOAT
        if (!fscanf(test, "%f", & (instance[j])))
#else
        if (!fscanf(test, "%lf", & (instance[j])))
#endif
            OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");

    // Read label
    if (!fscanf(test, "%d", label))
        OnErrorExit(ERR_READ, "ReadNextInstanceAndLabel");

    return -1;
}

// Read the info of the data file
void ReadDataInfo(int *instances, int *inputs, int *classes, char *file)
{
    FILE *f;
    char line[100];

    // Opens the file and read number of elements.
    if ((f = fopen(file, "rt")) == NULL) {
        printf("Dataset %s not found\n", file);
        OnErrorExit(ERR_FILE, "ReadDataAndLabels");
    }

    // Check labels
    if (fgets(line, 100, f) == NULL)
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (strncmp(line, "[SPARSE]", 8) && strncmp(line, "[ONE LABEL]", 11))
        rewind(f);

    // Read number of lines and columns.
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadDataAndLabels");

    fclose(f);
}

// Sort classes for even distribution of folds
void SortClasses(scalar **data, int instances, int *labels, int classes, int n_folds)
{
    int *list, *fold_labels[n_folds], fold_count[n_folds];
    scalar **fold_data[n_folds];

    int fold_size = instances%n_folds ? instances/n_folds+1 : instances/n_folds;

    // Initialize instance count per fold
    for (int i = 0; i < n_folds; i++) {
        fold_count[i] = 0;
        MALLOC(fold_data[i], scalar *, fold_size, "SortClasses: fold_data[i]");
        MALLOC(fold_labels[i], int, fold_size, "SortClasses: fold_labels[i]");
    }

    // Assign instances to folds in a round-robin fashion starting from last fold to get balanced strata
    MALLOC(list, int, instances, "SortClasses: list");

    int next = RANDOM(0, n_folds);
    for (int i = 0; i < classes; i++) {
        int count = 0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                list[count] = j;
                count++;
            }

        RandomizeVector(list, count);

        for (int j = 0; j < count; j++) {
            fold_data[next][fold_count[next]] = data[list[j]];
            fold_labels[next][fold_count[next]] = labels[list[j]];
            fold_count[next]++;

            next = (next+1) % n_folds;
        }
    }

    next = 0;
    for (int i = 0; i < n_folds; i++)
        for (int j = 0; j < fold_count[i]; j++) {
            data[next] = fold_data[i][j];
            labels[next] = fold_labels[i][j];
            next++;
        }

    if (next != instances)
        OnErrorExit(1, "Something went wrong in dataset reordering");

    FREE(list);
    for (int i = 0; i < n_folds; i++) {
        FREE(fold_data[i]);
        FREE(fold_labels[i]);
    }
}

int ReadNextBlockOfInstancesAndLabels(scalar **instances, int *label, int *inputs, int *classes, FILE *test, bool *first_time, int block_size)
{
    static int n_instances, read_instances, local_inputs;
    char line[100];

    // If first time read info
    if (*first_time) {
        rewind(test);

        // Check sparse label
        if (fgets(line, 100, test) == NULL)
            OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");
        if (strncmp(line, "[ONE LABEL]", 11))
            OnErrorExit(1, "Read next instance and label block function only for one label datasets");

        // Read number of lines and columns.
        if (!fscanf(test, "$ %d\n", &n_instances))
            OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");
        if (!fscanf(test, "$ %d\n", inputs))
            OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");
        if (!fscanf(test, "$ %d\n", classes))
            OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");

        *first_time = false;

        read_instances = 0;
        local_inputs = *inputs;

        // First time return number of instances and read nothing else
        return n_instances;
    }

    // Read instance
    for (int i = 0; i < block_size && read_instances < n_instances; i++, read_instances++) {
        for (int j = 0; j < local_inputs; j++)
#ifdef SCALAR_IS_FLOAT
            if (!fscanf(test, "%f", & (instances[i][j])))
#else
            if (!fscanf(test, "%lf", & (instances[i][j])))
#endif
                OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");

        // Read label
        if (!fscanf(test, "%d", &label[i]))
            OnErrorExit(ERR_READ, "ReadNextBlockOfInstancesAndLabels");
    }

    return -1;
}

// Read a limited number of patterns
void ReadLimitedDataAndLabels(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, int limit)
{
    int s;
    FILE *f;
    double outputs;
    char line[100];
    bool one_label = false;

    // Opens the file and read number of elements
    if ((f = fopen(file, "rt")) == NULL) {
        printf("Dataset %s not found\n", file);
        OnErrorExit(ERR_FILE, "ReadLimitedDataAndLabels");
    }

    // Check sparse label
    if (fgets(line, 100, f) == NULL)
        OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");
    if (!strncmp(line, "[SPARSE]", 8)) {
        fclose(f);
        ReadLimitedSparseDataAndLabels(data, labels, instances, inputs, classes, file, limit);
        return;
    } else if (!strncmp(line, "[ONE LABEL]", 11)) {
        // One label in[0, K-1] for K classes
        one_label = true;
    } else
        rewind(f);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");

    // Not enough instances
    if (*instances < limit)
        limit = *instances;

    // Memory allocation
    MALLOC(*data, scalar *, limit, "ReadLimitedDataAndLabels");
    for (int i = 0; i < limit; i++)
        MALLOC((*data)[i], scalar, *inputs, "ReadData: (*data)[i]");
    MALLOC(*labels, int, limit, "ReadLimitedDataAndLabels");

    // Random selection of instances
    bool *selected;
    CALLOC(selected, bool, *instances, "ReadLimitedDataAndLabels: selected");

    if (*instances > limit) {
        for (s = 0; s < limit;) {
            int next = RANDOM(0, *instances);
            if (!selected[next]) {
                selected[next] = true;
                s++;
            }
        }
    } else {
        for (int i = 0; i < *instances; i++)
            selected[i] = true;
    }

    // Read data
    s = 0;
    for (int i = 0; i < *instances && s < limit; i++) {
        for (int j = 0; j < *inputs; j++)
#ifdef SCALAR_IS_FLOAT
            if (!fscanf(f, "%f", & ((*data)[s][j])))
#else
            if (!fscanf(f, "%lf", & ((*data)[s][j])))
#endif
                OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");

        // Read labels
        if (one_label) {
            if (!fscanf(f, "%d", &(*labels)[s]))
                OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");
        } else
            for (int j = 0; j < *classes; j++) {
                if (!fscanf(f, "%lf", &outputs))
                    OnErrorExit(ERR_READ, "ReadLimitedDataAndLabels");
                if (outputs == 1.0)
                    (*labels)[s] = j;
            }

        if (selected[i])
            s++;
    }

    fclose(f);
    FREE(selected);

    *instances = limit;
}
