/**************************************************************************
*   Copyright (C) 2016 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions for spare files
 */
#include <stdio.h>
#include <malloc.h>
#include <stdbool.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include "common.h"
#include "macros.h"

/*
 * Read a data file. The structure of the file must be:
 *
 * $ Number of lines
 * $ Number of inputs
 * $ Number of classes
 * One pattern per line, variables in pairs feature:value ending in class label
 */

#define MAX_LINE_LENGTH 1000000

scalar ReadSparseDataAndLabelsInSparseFormat(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, int ***index, bool show_progress)
{
    FILE *f;
    char *line;
    int feature, label;
    scalar value, *next;
    scalar size = 0.0;    // Size of the reserved memory in GB

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadSparseDataAndLabelsInSparseFormat: line");

    // Opens the file and read number of elements.
    if ((f = fopen(file, "rt")) == NULL)
        OnErrorExit(ERR_FILE, "ReadSparseDataAndLabelsInSparseFormat");

    // Read sparse label
    if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabelsInSparseFormat: Problems reading file in first line");

    // If a number after "[SPARSE]" is the index of the first feature, otherwise assume 1
    int shift = 1;
    if (strlen(line) > 9)
        shift = atoi(line+8);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabelsInSparseFormat");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabelsInSparseFormat");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabelsInSparseFormat");

    // Memory allocation
    MALLOC(*data, scalar *, *instances, "ReadSparseDataAndLabelsInSparseFormat: *data");
    MALLOC(*index, int *, *instances, "ReadSparseDataAndLabelsInSparseFormat: *index");
    MALLOC(*labels, int, *instances, "ReadSparseDataAndLabelsInSparseFormat: *labels");

    size += *instances * (sizeof(scalar *) + sizeof(int *));

    MALLOC(next, scalar, *inputs, "ReadData: next");

    // Read data
    for (int i = 0; i < *instances; i++) {
        // Read line
        if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
            OnErrorExit(ERR_READ, "ReadSparseDataAndLabelsInSparseFormat: Problems reading file");

        for (int j = 0; j < *inputs; j++)
            next[j] = 0.0;

        // Read pairs feature:value separated by spaces or tabs until reaching label
        int offset = 0;
        while (GetNextPair(line, &offset, &feature, &value, &label)) {
            if (feature-shift >= *inputs || feature-shift < 0)
                OnErrorExit(ERR_READ, "Bad read feature");
            next[feature-shift] = value;
        }

        // Label
        (*labels)[i] = label;

        // Copy to data
        int c = 0;
        for (int k = 0; k < *inputs; k++)
            if (fabs(next[k]) > ZERO)
                c++;

        MALLOC((*data)[i], scalar, c, "ReadSparseDataAndLabelsInSparseFormat: (*data)[j]");
        MALLOC((*index)[i], int, c+1, "ReadSparseDataAndLabelsInSparseFormat: (*index)[j]");

        size += c*sizeof(scalar) + (c+1)*sizeof(int);

        c = 0;
        for (int k = 0; k < *inputs; k++)
            if (fabs(next[k]) > ZERO) {
                (*data)[i][c] = next[k];
                (*index)[i][c] = k+1;
                c++;
            }

        (*index)[i][c] = -1;

        SHOWPROGRESS(show_progress, "Read instances: %*d/%d\r", WIDTH(*instances), i+1, *instances);
    }

    SHOWPROGRESS(show_progress, "%c", '\n');

    fclose(f);

    FREE(next);
    FREE(line);

    return size/pow(2.0, 30.0);
}

scalar ReadDataAndLabelsInSparseFormat(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, int ***index, bool show_progress)
{
    FILE *f;
    double outputs;
    scalar *next = NULL;
    char *line;
    bool one_label = false;
    scalar size = 0.0;    // Size of the reserved memory in GB

    // If binary file exist read it
    char file_name[MAX_FILE_NAME];
    sprintf(file_name, "%s.bin", file);
    if (!access(file_name, R_OK))
        return ReadBinaryDataAndLabels(data, labels, instances, inputs, classes, file_name, index, show_progress, true);

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadDataAndLabelsInSparseFormat: line");

    // Opens the file and read number of elements
    if ((f = fopen(file, "rt")) == NULL) {
        printf("Dataset %s not found\n", file);
        OnErrorExit(ERR_FILE, "ReadDataAndLabelsInSparseFormat");
    }

    // Check sparse label
    if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
        OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");
    if (!strncmp(line, "[SPARSE]", 8)) {
        fclose(f);
        FREE(line);
        return ReadSparseDataAndLabelsInSparseFormat(data, labels, instances, inputs, classes, file, index, show_progress);
    } else if (!strncmp(line, "[ONE LABEL]", 11)) {
        // One label in[0, K-1] for K classes
        one_label = true;
    } else
        rewind(f);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");

    // Memory allocation
    MALLOC(*data, scalar *, *instances, "ReadDataAndLabelsInSparseFormat: *data");
    MALLOC(*index, int *, *instances, "ReadDataAndLabelsInSparseFormat: *index");
    MALLOC(*labels, int, *instances, "ReadDataAndLabelsInSparseFormat: *labels");

    size += *instances*(sizeof(int *)+sizeof(scalar *));

    MALLOC(next, scalar, *inputs, "ReadDataAndLabelsInSparseFormat: next");

    // Read data
    for (int i = 0; i < *instances; i++) {
        for (int j = 0; j < *inputs; j++)
#ifdef SCALAR_IS_FLOAT
            if (!fscanf(f, "%f", &(next[j])))
#else
            if (!fscanf(f, "%lf", &(next[j])))
#endif
                OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");

        // Read labels
        if (one_label) {
            if (!fscanf(f, "%d", &(*labels)[i]))
                OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");
        } else
            for (int j = 0; j < *classes; j++) {
                if (!fscanf(f, "%lf", &outputs))
                    OnErrorExit(ERR_READ, "ReadDataAndLabelsInSparseFormat");
                if (outputs == 1.0)
                    (*labels)[i] = j;
            }

        // Copy to data
        int c = 0;
        for (int k = 0; k < *inputs; k++)
            if (fabs(next[k]) > ZERO)
                c++;

        MALLOC((*data)[i], scalar, c, "ReadDataAndLabelsInSparseFormat: (*data)[i]");
        MALLOC((*index)[i], int, c+1, "ReadDataAndLabelsInSparseFormat: (*index)[i]")

        size += c*sizeof(scalar) + (c+1)*sizeof(int);

        c = 0;
        for (int k = 0; k < *inputs; k++)
            if (fabs(next[k]) > ZERO) {
                (*data)[i][c] = next[k];
                (*index)[i][c] = k+1;
                c++;
            }

        (*index)[i][c] = -1;

        SHOWPROGRESS(show_progress, "Read instances: %*d/%d\r", WIDTH(*instances), i+1, *instances);
    }

    SHOWPROGRESS(show_progress, "%c", '\n');

    fclose(f);

    FREE(next);
    FREE(line);

    return size/pow(2.0, 30.0);
}
