/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "type.h"

#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>

void InvertMatrix(scalar **m, scalar **inverse, int f)
{
    gsl_matrix *im, *lu;
    gsl_permutation *p;

    int i, j, signum;

    lu = gsl_matrix_alloc(f, f);
    im = gsl_matrix_alloc(f, f);
    p = gsl_permutation_alloc(f);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            gsl_matrix_set(lu, i, j, m[i][j]);

    /* LU decomposition. */
    gsl_linalg_LU_decomp(lu, p, &signum);
    /* Invert matrix. */
    gsl_linalg_LU_invert(lu, p, im);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            inverse[i][j] = gsl_matrix_get(im, i, j);

    // Free memory
    gsl_matrix_free(lu);
    gsl_matrix_free(im);
    gsl_permutation_free(p);
}


/* Determinant of an square matrix. */
scalar MatrixDeterminant(scalar **m, int f)
{
    gsl_matrix *lu;
    gsl_permutation *p;

    int i, j, signum;

    lu = gsl_matrix_alloc(f, f);
    p = gsl_permutation_alloc(f);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            gsl_matrix_set(lu, i, j, m[i][j]);

    /* LU decomposition. */
    gsl_linalg_LU_decomp(lu, p, &signum);

    /* Determinant. */
    scalar determinant = gsl_linalg_LU_det(lu, signum);

    // Free memory
    gsl_matrix_free(lu);
    gsl_permutation_free(p);

    return determinant;
}

void MaskedMatrixVectorProduct(scalar **m, scalar *v, scalar *res, int n, int *S)
{
    for (int i = 0; i < n; i++) {
        res[i] = 0.0;
        if (S[i])
            for (int j = 0; j < n; j++)
                res[i] += m[i][j] * v[j] * S[j];
    }
}

scalar MaskedDotProduct(scalar *x, scalar *y, int n, int *S)
{
    scalar dot = 0.0;

    for (int i = 0; i < n; i++)
        dot += x[i] * y[i] * S[i];

    return dot;
}

void PrintMatrix(scalar **m, int f, int c, FILE *file)
{
    for (int i = 0; i < f; i++) {
        for (int j = 0; j < c; j++)
            fprintf(file, "%7.4lf ", m[i][j]);
        fprintf(file, "\n");
    }
}
