/***************************************************************************
*   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "macros.h"

/* Accuracy measures for multiclass datasets from the confusion matrix

Confusion matrix must be:

                Predicted
           -------------------
           |
Actual     |
           |


Mesures in: Marina Sokolova and Guy Lapalme, "systematic analysis of performance measures for classification tasks", text, Information Processing and Management 45 (2009) 427–437
*/


void PrintMulticlassMeasures(int **M, int classes, int instances, FILE *out_file, char *text)
{
    scalar tp[classes], tn[classes], fp[classes], fn[classes], TP = 0.0, kappa, theta_1 = 0.0, theta_2 = 0.0;
    scalar average_accuracy, error_rate, precision_micro, recall_micro, f1_score_micro, precision_macro, recall_macro, f1_score_macro, accuracy, recall_macro_g_mean;

    // Get TP, TN, FP and FN for every class
    for (int i = 0; i < classes; i++) {
        TP += (tp[i] = M[i][i]);

        fp[i] = fn[i] = -M[i][i];
        for (int j = 0; j < classes; j++) {
            fp[i] += M[i][j];
            fn[i] += M[j][i];
        }

        tn[i] = instances - tp[i] - fp[i] - fn[i];
    }

    // Accuracy
    accuracy = TP/instances;

    // Average accuracy
    average_accuracy = 0.0;
    for (int i = 0; i < classes; i++)
        average_accuracy += (tp[i]+tn[i])/instances;
    average_accuracy /= classes;

    // Error error
    error_rate = 0.0;
    for (int i = 0; i < classes; i++)
        error_rate += (fp[i]+fn[i])/instances;
    error_rate /= classes;

    // Micro-averaged precision
    precision_micro = 0.0;
    for (int i = 0; i < classes; i++)
        precision_micro += tp[i]+fp[i];
    precision_micro = TP/precision_micro;

    // Micro-averaged recall
    recall_micro = 0.0;
    for (int i = 0; i < classes; i++)
        recall_micro += tp[i]+fn[i];
    recall_micro = TP/recall_micro;

    // Micro-averaged F1-score
    f1_score_micro = (2.0*precision_micro*recall_micro)/(precision_micro+recall_micro);

    // Macro-averaged precision
    precision_macro = 0.0;
    for (int i = 0; i < classes; i++)
        precision_macro += (tp[i]+fp[i] > 0) ? tp[i]/(tp[i]+fp[i]) : 0.0;
    precision_macro /= classes;

    // Macro-averaged recall
    recall_macro = 0.0;
    recall_macro_g_mean = 1.0;
    for (int i = 0; i < classes; i++) {
        recall_macro += (tp[i]+fn[i] > 0) ? tp[i]/(tp[i]+fn[i]) : 0.0;
        recall_macro_g_mean *= (tp[i]+fn[i] > 0) ? tp[i]/(tp[i]+fn[i]) : 0.0;
    }
    recall_macro /= classes;
    recall_macro_g_mean = pow(recall_macro_g_mean, 1.0/classes);

    // Macro-averaged F1-score
    f1_score_macro = (2.0*precision_macro*recall_macro)/(precision_macro+recall_macro);

    // Kappa
    // Sum of the diagonal
    for (int j = 0; j < classes; j++)
        theta_1 += M[j][j];

    if (theta_1 == instances)
        kappa = 1.0;
    else {
        theta_1 *= instances;

        for (int i = 0; i < classes; i++) {
            // Rows/columns total counts
            scalar row_i = 0, column_i = 0;
            for (int j = 0; j < classes; j++) {
                row_i += M[i][j];
                column_i += M[j][i];
            }

            theta_2 += row_i * column_i;
        }

        kappa = (theta_1 - theta_2)/(((scalar) instances)*instances-theta_2);
    }

    fprintf(out_file, "@%s Accuracy: %lf\n", text, accuracy);
    fprintf(out_file, "@%s Average accuracy: %lf\n", text, average_accuracy);
    fprintf(out_file, "@%s Error rate: %lf\n", text, error_rate);
    fprintf(out_file, "@%s Precision_\u03BC: %lf\n", text, precision_micro);
    fprintf(out_file, "@%s Recall_\u03BC: %lf\n", text, recall_micro);
    fprintf(out_file, "@%s F1-score_\u03BC: %lf\n", text, f1_score_micro);
    fprintf(out_file, "@%s Precision_M: %lf\n", text, precision_macro);
    fprintf(out_file, "@%s Recall_M: %lf\n", text, recall_macro);
    fprintf(out_file, "@%s Recall_M G-mean: %lf\n", text, recall_macro_g_mean);
    fprintf(out_file, "@%s F1-score_M: %lf\n", text, f1_score_macro);
    fprintf(out_file, "@%s Kappa: %lf\n", text, kappa);

    // Print confusion matrix
    fprintf(out_file, "%s contingency table\n", text);
    for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");
    fprintf(out_file, "       | Predicted\n");
    fprintf(out_file, "Actual |");
    for (int i = 0; i < classes; i++)
        fprintf(out_file, "   %*d", WIDTH(instances), i+1);
    fprintf(out_file, "\n");
    for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");
    for (int i = 0; i < classes; i++) {
        fprintf(out_file, "%3d    |", i+1);
        for (int j = 0; j < classes; j++)
            fprintf(out_file, "   %*d", WIDTH(instances), (int) M[i][j]);
        fprintf(out_file, "\n");
    }
    for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");
    fflush(out_file);
}
