/***************************************************************************
*   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions for spare files
 */
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include "common.h"
#include "macros.h"

/*
 * Read a data file. The structure of the file must be:
 *
 * $ Number of lines
 * $ Number of inputs
 * $ Number of classes
 * One pattern per line, variables in pairs feature:value ending in class label
 */

#define MAX_LINE_LENGTH 1000000

// If huge dataset, labels are read, but actual data is not
void ReadSparseDataAndLabels(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, bool huge)
{
    FILE *f;
    char *line;
    int feature, label;
    scalar value;

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadSparseDataAndLabels: line");

    // Opens the file and read number of elements.
    if ((f = fopen(file, "rt")) == NULL)
        OnErrorExit(1, "ReadSparseDataAndLabels");

    // Read sparse label
    if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
        OnErrorExit(1, "ReadSparseDataAndLabels: Problems reading file in first line");

    // If a number after "[SPARSE]" is the index of the first feature, otherwise assume 1
    int shift = 1;
    if (strlen(line) > 9)
        shift = atoi(line+8);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabels");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabels");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadSparseDataAndLabels");

    // Memory allocation
    if (!huge) {
        MALLOC(*data, scalar *, *instances, "ReadSparseDataAndLabels: *data");

        // Fill with zero
        for (int i = 0; i < *instances; i++)
            CALLOC((*data) [i], scalar, *inputs, "ReadData: (*data)[i]");
    }

    MALLOC(*labels, int, *instances, "ReadSparseDataAndLabels: *labels");

    // Read data
    for (int i = 0; i < *instances; i++) {
        // Read line
        if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
            OnErrorExit(ERR_READ, "ReadSparseDataAndLabels: Problems reading file");

        // Read pairs feature:value separated by spaces or tabs until reaching label
        int offset = 0;
        if (huge)      // Just skip to label
            while (GetNextPair(line, &offset, &feature, &value, &label));
        else
            while (GetNextPair(line, &offset, &feature, &value, &label)) {
                if (feature-shift >= *inputs || feature-shift < 0) {
                    printf("Feature: %d\n", feature-shift);
                    OnErrorExit(ERR_READ, "Bad read feature");
                }
                (*data)[i][feature-shift] = value;
            }

        // Label
        (*labels)[i] = label;
    }

    fclose(f);
    FREE(line);
}

// Returns: TRUE for pair feature:value FALSE for class label
bool GetNextPair(char *string, int *offset, int *feature, scalar *value, int *label)
{
    static char *next = NULL;

    if (!next)
        MALLOC(next, char, MAX_LINE_LENGTH, "ReadSparseDataAndLabels: line");

    // Copy to next the next pair and update offset
    int i = 0;
    while (string[*offset] != '\0' && string[*offset] != ' ' && string[*offset] != '\t') {
        next[i] = string[*offset];
        (*offset)++;
        i++;
    }
    next[i] = '\0';

    // Remove repeated separators and update offset for next pair
    while (string[*offset] == ' ' || string[*offset] == '\t')
        (*offset)++;

    // Check if there is a ":" meaning a pair
    char *pair = strchr(next, ':');

    if (pair == NULL) {
        *label = atoi(next);
        return false;
    } else {
        pair[0] = '\0';
        *feature = atoi(next);
        *value = atof(pair+1);
        return true;
    }
}

void ReadLimitedSparseDataAndLabels(scalar ***data, int **labels, int *instances, int *inputs, int *classes, char *file, int limit)
{
    FILE *f;
    char *line;
    int feature, label, s;
    scalar value;

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadSparseDataAndLabels: line");

    // Opens the file and read number of elements.
    if ((f = fopen(file, "rt")) == NULL)
        OnErrorExit(1, "ReadLimitedSparseDataAndLabels");

    // Read sparse label
    if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
        OnErrorExit(1, "ReadLimitedSparseDataAndLabels: Problems reading file in first line");

    // If a number after "[SPARSE]" is the index of the first feature, otherwise assume 1
    int shift = 1;
    if (strlen(line) > 9)
        shift = atoi(line+8);

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadLimitedSparseDataAndLabels");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadLimitedSparseDataAndLabels");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadLimitedSparseDataAndLabels");

    // Not enough instances
    if (*instances < limit)
        limit = *instances;

    // Memory allocation
    MALLOC(*data, scalar *, limit, "ReadLimitedDataAndLabels");
    for (int i = 0; i < limit; i++)
        CALLOC((*data)[i], scalar, *inputs, "ReadData: (*data)[i]");
    MALLOC(*labels, int, limit, "ReadLimitedDataAndLabels");

    // Random selection of instances
    bool *selected;
    CALLOC(selected, bool, *instances, "ReadLimitedDataAndLabels: selected");

    if (*instances > limit) {
        for (s = 0; s < limit;) {
            int next = RANDOM(0, *instances);
            if (!selected[next]) {
                selected[next] = true;
                s++;
            }
        }
    } else {
        for (int i = 0; i < *instances; i++)
            selected[i] = true;
    }

    // Read data
    s = 0;
    for (int i = 0; i < *instances && s < limit; i++) {
        // Read line
        if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
            OnErrorExit(1, "ReadLimitedSparseDataAndLabels: Problems reading file");

        // Read pairs feature:value separated by spaces or tabs until reaching label
        int offset = 0;
        while (GetNextPair(line, &offset, &feature, &value, &label)) {
            if (feature-shift >= *inputs || feature-shift < 0)
                OnErrorExit(ERR_READ, "Bad read feature");
            (*data)[s][feature-shift] = value;
        }

        // Label
        (*labels)[s] = label;

        if (selected[i])
            s++;
        else
            // Clean instance
            for (int j = 0; j < *inputs; j++)
                (*data)[s][j] = 0.0;
    }

    fclose(f);
    FREE(selected);
    FREE(line);

    *instances = limit;
}
