/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// Obtains AUC from ROC from an output vector

scalar AUROCFromOutputs(scalar *outputs, int n_positives, int n_negatives, bool write_roc_data, char *filename, bool sorted)
{
    scalar qtrap(scalar **, scalar, scalar, int);
    int CompareDouble(const void *, const void *);
    int CompareDoubleInverted(const void *, const void *);

    int n_points = 200;
    FILE *f = NULL;
    scalar **points, *pos_outputs, *neg_outputs, threshold[n_points];

    ALLOCA(points, scalar *, n_points, "AUROCFromOutputs: points");
    for (int i = 0; i < n_points; i++)
        ALLOCA(points[i], scalar, 2, "AUROCFromOutputs: points[i]");

    pos_outputs = outputs;
    neg_outputs = outputs + n_positives;

    // Initial point
    points[0][0] = points[0][1] = 0.0;

    // Sort outputs
    if (!sorted) {
        qsort(pos_outputs, n_positives, sizeof(scalar), CompareDouble);
        qsort(neg_outputs, n_negatives, sizeof(scalar), CompareDouble);
    }

    // Get threshold for sensitivity and specificity and sort
    scalar sn = 0.01;
    for (int point = 1; point < n_points/2; point++, sn += 0.01)
        threshold[point] = pos_outputs[(int)(n_positives * (1.0 - sn))];
    scalar sp = 0.99;
    for (int point = n_points/2; point < n_points; point++, sp -= 0.01)
        threshold[point] = neg_outputs[(int)(n_negatives * sp)];

    qsort(threshold+1, n_points-2, sizeof(scalar), CompareDoubleInverted);

    // Write ROC data?
    if (write_roc_data) {
        if ((f = fopen(filename, "w")) == NULL) {
            printf("Filename: \"%s\"\n", filename);
            OnErrorExit(1, "Could not open ROC data file");
        }
        fprintf(f, "#1-Sp\tSn\tThreshold\tSp\tSn\tG\tTP\tFN\tTN\tFP\n");
        fprintf(f, "0.000000\t0.000000\t----\t1.0000\t0.0000\t0.0000\t%5d\t%4d\t%8d\t%7d\n", 0, n_positives, n_negatives, 0);
    }

    for (int point = 1; point < n_points-1; point++) {
        scalar specificity, sensitivity, thr, TN, TP, FN, FP;

        thr = threshold[point];

        // Obtain sensitivity
        // Positive sequences
        TP = 0.0;
        for (int i = n_positives - 1; pos_outputs[i] > thr && i >= 0; i--)
            TP++;
        FN = n_positives - TP;

        // Obtain specificity
        // Negative sequences
        TN = 0.0;
        for (int i = 0; neg_outputs[i] <= thr && i < n_negatives; i++)
            TN++;
        FP = n_negatives - TN;

        sensitivity = TP/(TP + FN);
        specificity = TN/(TN + FP);

        points[point][0] = 1.0 - specificity;
        points[point][1] = sensitivity;

        if (write_roc_data)
            fprintf(f, "%8.6lf\t%8.6lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%5d\t%4d\t%8d\t%7d\n", points[point][0], points[point][1], thr, specificity, sensitivity,
                    pow(specificity*sensitivity, 0.5), (int) TP, (int) FN, (int) TN, (int) FP);
    }

    // Final point
    points[n_points-1][0] = points[n_points-1][1] = 1.0;

    // Write ROC data?
    if (write_roc_data) {
        fprintf(f, "1.000000\t1.000000\t----\t0.0000\t1.0000\t0.0000\t%5d\t%4d\t%8d\t%7d\n", n_positives, 0, 0, n_negatives);
        fclose(f);
    }

    // Return auROC
    scalar auc = qtrap(points, 0.0, 1.0, n_points);

    return auc;
}

// auPRC: Area under precision/recall curve

#define N_POINTS 101
scalar AUPRCFromOutputs(scalar *outputs, int n_positives, int n_negatives, bool write_prc_data, char *filename, bool sorted)
{
    scalar qtrap(scalar **, scalar, scalar, int);
    int CompareDouble(const void *, const void *);

    scalar **points, *pos_outputs, *neg_outputs;

    pos_outputs = outputs;
    neg_outputs = outputs + n_positives;

    // Sort outputs
    if (!sorted) {
        qsort(pos_outputs, n_positives, sizeof(scalar), CompareDouble);
        qsort(neg_outputs, n_negatives, sizeof(scalar), CompareDouble);
    }

    // Get number of possible thresholds
    int n_thresholds = 1;
    scalar *thresholds;
    ALLOCA(thresholds, scalar, n_positives+1, "AUPRCFromOutputs: thresholds");
    thresholds[0] = pos_outputs[n_positives-1]; // TP = 0
    for (int i = n_positives-2; i >= 0; i--) {
        if (pos_outputs[i] < thresholds[n_thresholds-1]) {
            thresholds[n_thresholds] = pos_outputs[i];
            n_thresholds++;
        }
    }
    thresholds[n_thresholds] = pos_outputs[0] - 0.001;
    n_thresholds++;

    // Get TP and FP values for all thresholds
    int *TP, *FP;
    ALLOCA(TP, int, n_positives, "AUPRCFromOutputs: TP");
    ALLOCA(FP, int, n_positives, "AUPRCFromOutputs: FP");
    for (int i = 0; i < n_thresholds; i++) {
        // Obtain TP
        TP[i] = 0;
        for (int j = n_positives - 1; pos_outputs[j] > thresholds[i] && j >= 0; j--)
            TP[i]++;

        // Obtain FP
        FP[i] = 0;
        for (int j = n_negatives - 1; neg_outputs[j] > thresholds[i] && j >= 0; j--)
            FP[i]++;
    }

    // As many points as number of positives + 1
    int n_points = 0;
    ALLOCA(points, scalar *, n_positives+1, "AURPCFromOutputs: points");
    for (int i = 0; i < n_positives+1; i++)
        ALLOCA(points[i], scalar, 2, "AURPCFromOutputs: points[i]");

    for (int i = 1; i < n_thresholds-1 && i < n_positives; i++) {
        if (n_points == 0) {
            // First point (0,1) if largest values is a positive instance and (0,0) otherwise
            if (pos_outputs[n_positives-1] > neg_outputs[n_negatives-1]) {
                points[n_points][0] = 0.0;
                points[n_points][1] = 1.0;
            } else
                points[n_points][0] = points[0][1] = 0.0;
        } else {
            // Get Recall/Precision for current point
            points[n_points][0] = TP[i]/((scalar) n_positives);
            points[n_points][1] = TP[i]/((scalar) TP[i]+FP[i]);
        }
        n_points++;

        // Follow J. Davis & M. Goadrich (2006). "The Relationship Between Precision-Recall and ROC Curves"
        // Proceedings of the Twenty-Third International Conference on Machine Learning (ICML'06), Pittsburgh, PA.
        // Interpolate?
        for (int j = 1; TP[i] + j < TP[i+1] && n_points < n_positives; j++) {
            scalar interpolated_value;
            points[n_points][0] = (TP[i]+j)/((scalar) n_positives);
            interpolated_value = TP[i]+j+FP[i] + j*((FP[i+1]-FP[i])/((scalar) (TP[i+1]-TP[i])));
            points[n_points][1] = (TP[i]+j)/interpolated_value;
            n_points++;
        }
    }
    points[n_points][0] = TP[n_thresholds-1]/((scalar) n_positives);
    points[n_points][1] = TP[n_thresholds-1]/((scalar) TP[n_thresholds-1]+FP[n_thresholds-1]);

    // Write PRC data?
    if (write_prc_data) {
        FILE *f;

        if ((f = fopen(filename, "w")) == NULL)
            OnErrorExit(1, "Could not open PRC data file");
        fprintf(f, "#Recall\tPrecision\n");
        for (int i = 0; i < n_points; i++)
            fprintf(f, "%6.4lf\t%6.4lf\n", points[i][0], points[i][1]);
        fclose(f);
    }

    if (n_points <= 1)
        return 0.0;

    // Return AUC
    return qtrap(points, points[0][0], points[n_points-1][0], n_points);
}

int CompareDoubleInverted(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (scalar *) x - * (scalar *) y) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}
