/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <string.h>

#define MAX(a, b)                       (((a)<(b)) ? (b) : (a))
#define MALLOC(pt, type, number, str)   {if (number) { if (((pt) = (type *) malloc (MAX(sizeof (type) *(number), 1))) == NULL) exit (1); } else pt = NULL;}

// Implements numerical integration using trapezoid rule
// Reads a plain text file. Integrates each pair of columns.
// Left column is X, right column is Y

int main(int argc, char *argv[])
{
    double qtrap(double **, double, double, int);
    void ReadDataForIntegration(char *, double ***, int, int *);

    double ***data;
    int points, pairs;

    // Check parameters
    if (argc < 3) {
        printf("auroc points-file pairs\n");
        exit(1);
    }

    // Get number of columns
    pairs = atoi(argv[2]);
    MALLOC(data, double **, pairs, "main: data");

    // Read data
    ReadDataForIntegration(argv[1], data, pairs, &points);

    // Do integration
    for (int i = 0; i < pairs; i++) {
        double auc = qtrap(data[i], 0.0, 1.0, points);

        printf("AUC pair %d = %lf\n", i+1, auc);
    }

    return 0;
}


// Interpolate value
double f(double x, double **data, int n)
{
    double x1, x2, y1, y2, y;

    // If outside limits return 0
    if (x < data[0][0] && x > data[n-1][0])
        return 0.0;

    // Otherwise linear interpolation
    // Get points to interpolate
    int i = 0;
    while (x > data[i][0])
        i++;

    // Check especial cases
    if (i == 0)
        return data[0][1];

    x1 = data[i-1][0];
    y1 = data[i-1][1];

    // Check especial cases
    if (x == x1)
        y = y1;
    else {
        x2 = data[i][0];
        y2 = data[i][1];

        if (x==x2)
            y=y2;
        // If x1=x2, return y1
        else if (x1==x2)
            y = y1;
        else {
            // Equation is: y = y1 + [(y2-y1)/(x2-x1)]·(x-x1) if x1 != x2
            y = y1 + ((y2-y1) / (x2-x1)) * (x-x1);
        }
    }

    return y;
}

double trapzd(double **data, double a, double b, int n, int nvalues)
{
    double  x, tnm, sum, del;
    static double s;
    int it, j;
    if (n == 1) {
        return (s = 0.5* (b-a) * (f(a, data, nvalues) + f(b, data, nvalues)));
    } else {
        for (it = 1,j = 1; j < n-1; j++)
            it <<= 1;
        tnm = it;
        del = (b - a) / tnm;   /*This is the spacing of points to be added. */
        x = a + 0.5 * del;
        for (sum = 0.0, j = 1; j <= it; j++, x += del)
            sum += f(x, data, nvalues);
        s = 0.5 * (s + (b-a) * sum/tnm);     /*This replaces s by its refined value.*/
        return s;
    }
}

void ReadDataForIntegration(char *file, double ***data, int functions, int *n)
{
    FILE *f;
    char line[1000];

    if ((f = fopen(file, "r")) == NULL)
        exit(1);

    // Count points
    *n = -1;
    do {
        if (fgets(line, 1000, f) == NULL && !feof(f))
            exit(1);

        // Ignore comments
        if (line[0] != '#')
            (*n)++;
    } while (!feof(f));

    rewind(f);

    // Allocate memory
    for (int i = 0; i < functions; i++) {
        MALLOC(data[i], double *, *n, "ReadDataForIntegration: data[i]");
        for (int j = 0; j < *n; j++)
            MALLOC(data[i][j], double, 2, "ReadDataForIntegration: data[i][j]");
    }

    for (int i = 0; i < *n; i++) {
        if (fgets(line, 1000, f) == NULL && !feof(f))
            exit(1);

        // Ignore comments
        if (line[0] == '#') {
            i--;
            continue;
        }

        // Split string and get values
        char *pch = strtok(line, " \t,");
        if (pch == NULL) {
            printf("Error reading the file\n");
            exit(1);
        }
        for (int j = 0; j < functions; j++) {
            data[j][i][0] = atof(pch);
            if ((pch = strtok(NULL, " \t,")) == NULL) {
                printf("Error reading the file\n");
                exit(1);
            }
            data[j][i][1] = atof(pch);
        }
    }

    fclose(f);
}

#define EPS 1.0e-5
#define JMAX 20
double qtrap(double **data, double a, double b, int n)
{
    double trapzd(double **, double, double, int, int);

    int j;
    double s,olds;
    olds = -1.0e30;
    for (j=1; j<=JMAX; j++) {
        s=trapzd(data, a, b, j, n);
        if (j > 5)
            if (fabs(s-olds) < EPS*fabs(olds) || (s == 0.0 && olds == 0.0))
                return s;
        olds=s;
    }

    return 0.0;
}
