/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
// #include "common.h"

// Implements learning from bites method:
// N. Chawla, L. O. Hall, K. W. Bowyer and W. P. Kegelmeyer, "Learning Ensembles from Bites: A Scalable and Accurate Approach", J. Mach. Learn. Res.,
// vol. 5, 421-451, 2004.

t_Ensemble *TrainLearningFromBites(const t_Parameters *parm, char **p_seqs, char **n_seqs, int length, int n_poss, int n_negs, int number_of_bites, int bite_size)
{
    t_Ensemble *ensemble;
    int *list_pos, *list_neg;
    char **positives, **negatives;

    // Vectors for random sampling
    MALLOC(list_neg, int, n_negs, "TrainLearningFromBites: list_neg");
    for (int j = 0; j < n_negs; j++)
        list_neg[j] = j;
    MALLOC(list_pos, int, n_poss, "TrainLearningFromBites: list_pos");
    for (int j = 0; j < n_poss; j++)
        list_pos[j] = j;

    // Data for local training
    MALLOC(positives, char *, bite_size/2, "TrainLearningFromBites: positives");
    MALLOC(negatives, char *, bite_size/2, "TrainLearningFromBites: negatives");

    // Allocate memory for the classifiers
    MALLOC(ensemble, t_Ensemble, 1, "TrainLearningFromBites: ensemble");
    MALLOC(ensemble->alpha, scalar, number_of_bites, "TrainLearningFromBites: ensemble->alpha");
    switch (parm->base_learner) {
        case C45:
            MALLOC(ensemble->trees, Tree, number_of_bites, "TrainLearningFromBites: ensemble->trees");
            ensemble->lsvm = NULL;
            break;
        case SVM:
            MALLOC(ensemble->lsvm, struct svm_model *, number_of_bites, "TrainLearningFromBites: ensemble->lsvm");
            ensemble->trees = NULL;
            break;
        default:
            OnErrorExit(1, "Bad base learner");
    }

    ensemble->threshold = 0.0;
    ensemble->s = number_of_bites;

    // Cross-validate parameters?
    /* if (parm->cv_parameters)
        CrossValidatedParameters(positives, bite_size/2, length, negatives, bite_size/2);*/

    // Construct the ensemble for the specified number of bites
    for (int bite = 0; bite < number_of_bites; bite++) {
        ensemble->alpha[bite] = 1.0;

        // Random undersampling
        GetRandomisedVector(list_neg, n_negs);
        GetRandomisedVector(list_pos, n_poss);

        for (int i = 0; i < bite_size/2; i++) {
            positives[i] = p_seqs[list_pos[i]];
            negatives[i] = n_seqs[list_neg[i]];
        }

        if (parm->base_learner == C45)
            // Train and set threshold
            ensemble->trees[bite] = TrainC45Classifier(parm, positives, bite_size/2, length, negatives, bite_size/2, false, NULL);
        else
            TrainSVMClassifier(parm, &(ensemble->lsvm[bite]), positives, bite_size/2, length, negatives, bite_size/2, false, NULL, NULL, NULL, NULL);

        if (parm->verbose >= 0)
            printf("Learned bite %d\n", bite+1);
    }

    FREE(list_neg);
    FREE(list_pos);
    FREE(positives);
    FREE(negatives);

    return ensemble;
}
