/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"

Tree TrainC45Classifier(const t_Parameters *parm, char **pseqs, int n_pseqs, int length, char **nseqs, int n_nseqs, bool releasedata, scalar *threshold)
{
    int CompareDouble(const void *, const void *);

    int best;
    Tree tree;
    extern Tree *Raw, *Pruned;

    // Global variables
    MaxClass = 1;
    MaxItem = n_pseqs + n_nseqs - 1;
    MaxAtt = length - 1;

    // Allocate space for items
    Description desc;
    MALLOC(Item, Description, n_pseqs + n_nseqs + 1, "main: Item");
    desc = (Description) calloc(MaxAtt + 2, sizeof(AttValue));

    MaxItem = -1;
    MaxAtt = length - 1;

    // Create data for tree taking into account sampling
    char name[2];
    name[1] = 0;
    for (int j = 0; j < n_pseqs; j++) {
        MaxItem++;
        Item[MaxItem] = (Description) calloc(MaxAtt + 2, sizeof(AttValue));
        for (int i = 0; i < length; i++) {
            name[0] = pseqs[j][i];
            int Dv = Which(name, AttValName[i], 1, MaxAttVal[i]);
            DVal(Item[MaxItem], i) = Dv;    // Order of discrete value in the series
        }
        Class(Item[MaxItem]) = POSITIVE;
    }

    for (int j = 0; j < n_nseqs; j++) {
        MaxItem++;
        Item[MaxItem] = (Description) calloc(MaxAtt + 2, sizeof(AttValue));
        for (int i = 0; i < length; i++) {
            name[0] = nseqs[j][i];
            int Dv = Which(name, AttValName[i], 1, MaxAttVal[i]);
            DVal(Item[MaxItem], i) = Dv;    // Order of discrete value in the series
        }
        Class(Item[MaxItem]) = NEGATIVE;

        if (releasedata)   // Release training data to save memory
            FREE(nseqs[j]);
    }

    if (releasedata) {
        FREE(nseqs);
    }

    // Create tree
    if (TRIALS == 1) {
        best = 0;
        OneTree(parm->prune_tree, parm->out_file, parm->c45_weights);
    } else
        best = BestTree(parm->out_file, parm->c45_weights);

    // Copy tree to classifier
    if (parm->prune_tree)
        tree = Pruned[best];
    else
        tree = Raw[best];

    if (parm->soften_thresholds)
        SoftenThresh(tree, parm->out_file, parm->c45_weights);

    ReleaseTreeData();

    if (parm->prune_tree) {
        for (int t = 0; t < TRIALS; t++) {
            ReleaseTree(Raw[t]);
            if (t != best)
                ReleaseTree(Pruned[t]);
        }
        free(Raw);
        free(Pruned);
    } else {
        for (int t = 0; t < TRIALS; t++)
            if (t != best)
                ReleaseTree(Raw[t]);
        free(Raw);
    }

    // If a sensitivity level is required the threshold is adjusted, otherwise it is 0.0
    if (parm->sensitivity > 0.0 && threshold != NULL) {
        // Get threshold to obtain desired sensitivity = TP/(TP+FN)
        // Get all outputs and ...
        scalar *sc;
        ALLOCA(sc, scalar, n_pseqs, "TrainC45Classifier: sc");
        for (int i = 0; i < n_pseqs; i++) {
            sc[i] = C45Output(tree, pseqs[i], length);
            if (releasedata)   // Release training data to save memory
                FREE(pseqs[i]);
        }
        // ...sort them
        qsort(sc, n_pseqs, sizeof(scalar), CompareDouble);

        // Get threshold for desired sensitivity
        *threshold = sc[(int)(n_pseqs * (1.0 - parm->sensitivity)) ];
    } else if (threshold != NULL)
        *threshold = 0.0;

    if (threshold != NULL)
        fprintf(parm->out_file, "C4.5 threshold: %lf\n", *threshold);

    if (releasedata) {
        FREE(pseqs);
    }

    // Free memory for tree learning
    FREE(desc);
    for (int i = 0; i <= MaxItem; i++)
        FREE(Item[i]);
    FREE(Item);

    return tree;
}

// Return probability of having a positive class
scalar C45Output(Tree tree, char *sequence, int length)
{
    static Description desc = NULL;
    char name[2];
    name[1] = 0;
    extern float *ClassSum;

    if (desc == NULL)
        desc = (Description) calloc(length + 2, sizeof(AttValue));

    MaxAtt = length - 1;

    // Create description
    for (int i = 0; i < length; i++) {
        name[0] = sequence[i];
        int Dv = Which(name, AttValName[i], 1, MaxAttVal[i]);
        DVal(desc, i) = Dv;
    }

    // Get output
    Category(desc, tree);

    // Return class
    return 2.0 * (ClassSum[POSITIVE] - 0.5);        // Output in [-1, 1]
}
