/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
// #include "common.h"

// Implements BalanceCascadeEnsemble method from:
// X-Y. Liu, J. Wu and Z.-H. Zhou, "Exploratory undersampling for class-imbalance learning", IEEE Transactions on Systems, Man, and Cybernetics - Part B: Cybernetics, vol. 39, no. 2, April 2009, pp. 539-548.

// Due to computational cost, C4.5 is chosen as base learner
// Uses resampling by default for AdaBoost

t_MultiEnsemble *TrainBalanceCascadeEnsemble(const t_Parameters *parm, char **p_seqs, char **n_seqs, int length, int n_poss, int n_negs, int T, int s_i)
{
    t_MultiEnsemble *multiensemble;
    int *list;
    char **seqs;
    scalar *sc, f, *probs, betta, e_i, *w;    // Weights for Adaboost
    bool *ok;

    // Vector for random sampling
    MALLOC(list, int, n_negs, "TrainBalanceCascadeEnsemble: list");
    for (int j = 0; j < n_negs; j++)
        list[j] = j;

    // Data for local training of trees and boosting
    MALLOC(seqs, char *, n_poss, "TrainBalanceCascadeEnsemble: seqs");
    MALLOC(w, scalar, 2 * n_poss, "TrainBalanceCascadeEnsemble: w");
    MALLOC(ok, bool, 2 * n_poss, "TrainBalanceCascadeEnsemble: ok");
    MALLOC(probs, scalar, n_poss, "TrainBalanceCascadeEnsemble: probs");

    // As many trees as classifiers
    MALLOC(multiensemble, t_MultiEnsemble, 1, "TrainBalanceCascadeEnsemble: multiensemble");
    MALLOC(multiensemble->ensemble, t_Ensemble, T, "TrainBalanceCascadeEnsemble: multiensemble->ensemble");
    for (int i = 0; i < T; i++) {
        multiensemble->ensemble[i].s = s_i;
        multiensemble->ensemble[i].threshold = 0.0;
        switch (parm->base_learner) {
            case C45:
                MALLOC(multiensemble->ensemble[i].trees, Tree, s_i, "TrainEasyEnsemble: multiensemble->ensemble[i].trees");
                break;
            case SVM:
                MALLOC(multiensemble->ensemble[i].lsvm, struct svm_model *, s_i, "TrainEasyEnsemble: multiensemble->ensemble[i].lsvm");
                break;
            default:
                OnErrorExit(1, "Bad base learner");
        }
        MALLOC(multiensemble->ensemble[i].alpha, scalar, s_i, "TrainBalanceCascadeEnsemble: multiensemble->alpha");
    }

    // For obtaining desired threshold
    MALLOC(sc, scalar, n_negs, "TrainBalanceCascadeEnsemble: sc");

    // Construct T AdaBoost ensembles of size s_i randomly undersampling
    f = pow(((scalar) n_poss) / n_negs, 1.0/ (T - 1.0));              // Desired FP ratio
    fprintf(parm->out_file, "Desired FP ratio: %lf\n", f);
    multiensemble->T = 0;
    for (int i = 0; i < T; i++) {
        // Random undersampling
        GetRandomisedVector(list, n_negs);

        // Initial weights
        for (int j = 0; j < 2*n_poss; j++)
            w[j] = 1.0;

        // Random chosen instances
        for (int j = 0; j < MIN(n_poss, n_negs); j++)
            seqs[j] = n_seqs[list[j]];

        // Construct ensemble
        for (int j = 0; j < s_i; j++) {
            // Train and set threshold
            if (parm->base_learner == C45)
                // Train and set threshold
                multiensemble->ensemble[i].trees[j] = TrainC45Classifier(parm, p_seqs, n_poss, length, seqs, MIN(n_poss, n_negs), false, NULL);
            else
                TrainSVMClassifier(parm, &(multiensemble->ensemble[i].lsvm[j]), p_seqs, n_poss, length, seqs,  MIN(n_poss, n_negs), false, NULL, NULL, NULL, NULL);

            // Test and update weights
            // Positive samples
            e_i = 0.0;
            if (parm->base_learner == C45) {
                for (int k = 0; k < n_poss; k++)
                    if (C45Output(multiensemble->ensemble[i].trees[j], p_seqs[k], length) > 0.0)
                        // Ok
                        ok[k] = true;
                    else {
                        // Misclassified
                        ok[k] = false;
                        e_i += w[k];
                    }

                // Negative samples
                for (int k = 0; k < MIN(n_poss, n_negs); k++)
                    if (C45Output(multiensemble->ensemble[i].trees[j], n_seqs[list[k]], length) < 0.0)
                        // Ok
                        ok[k + n_poss] = true;
                    else {
                        // Misclassified
                        ok[k + n_poss] = false;
                        e_i += w[k + n_poss];
                    }
            } else if (parm->base_learner == SVM) {
                for (int k = 0; k < n_poss; k++)
                    if (SVMOutput(multiensemble->ensemble[i].lsvm[j], p_seqs[k], length, NULL, parm->use_salzberg_method, NULL) > 0.0)
                        // Ok
                        ok[k] = true;
                    else {
                        // Misclassified
                        ok[k] = false;
                        e_i += w[k];
                    }

                // Negative samples
                for (int k = 0; k < MIN(n_poss, n_negs); k++)
                    if (SVMOutput(multiensemble->ensemble[i].lsvm[j], n_seqs[list[k]], length, NULL, parm->use_salzberg_method, NULL) < 0.0)
                        // Ok
                        ok[k + n_poss] = true;
                    else {
                        // Misclassified
                        ok[k + n_poss] = false;
                        e_i += w[k + n_poss];
                    }
            }

            // Total error
            e_i /= n_poss + MIN(n_poss, n_negs);

            if (parm->verbose > 1) {
                if (parm->base_learner == C45)
                    fprintf(parm->out_file, "Classifier %d/%d of ensemble %d/%d error: %lf\t size: %d\n", j+1, s_i, i+1, T, e_i, TreeSize(multiensemble->ensemble[i].trees[j]));
                else if (parm->base_learner == SVM)
                    fprintf(parm->out_file, "Classifier %d/%d of ensemble %d/%d error: %lf\t suport vectors: %d\n", j+1, s_i, i+1, T, e_i, SVMSize(multiensemble->ensemble[i].lsvm[j]));
            }

            // If above 50% random resampling
            if (e_i > 0.5) {
                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = 0.0;

                betta = 1.0; // Remove classifier assigning 0 weight
                for (int k = 0; k < 2 * n_poss; k++)
                    w[k] = 1.0;
            } else if (e_i == 0.0) {   // If perfect also random resampling
                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = 10.0;

                betta = 1.0e-10;
                for (int k = 0; k < 2 * n_poss; k++)
                    w[k] = 1.0;
            } else {
                betta = e_i / (1.0 - e_i);

                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = log(1.0 / betta);

                for (int k = 0; k < 2 * n_poss; k++)
                    if (ok[k])
                        w[k] /= 2.0 * (1.0 - e_i);
                    else
                        w[k] /= 2.0 * e_i;
            }

            // Resample for next step
            probs[0] = w[0];
            for (int k = 1; k < MIN(n_poss, n_negs); k++)
                probs[k] = probs[k - 1] + w[k];

            for (int k = 0; k < MIN(n_poss, n_negs); k++) {
                int pattern = RouletteSelection(probs, n_poss);

                seqs[k] = n_seqs[list[pattern]];
            }

            // Normalize w[] to sum n
            scalar sum = 0.0;
            for (int k = 0; k < n_poss + MIN(n_poss, n_negs); k++)
                sum += w[k];
            for (int k = 0; k < n_poss + MIN(n_poss, n_negs); k++)
                w[k] *= (2.0*n_poss) /sum;

            if (parm->verbose > 2) {
                fprintf(parm->out_file, "Positive instances boosting weights: ");
                for (int k = 0; k < n_poss; k++)
                    fprintf(parm->out_file, "%6.3lf ", w[k]);
                fprintf(parm->out_file, "\nNegative instances boosting weights: ");
                for (int k = n_poss; k < n_poss+ MIN(n_poss, n_negs); k++)
                    fprintf(parm->out_file, "%6.3lf ", w[k]);
                fprintf(parm->out_file, "\n");
            }
        }

        // Add ensemble
        multiensemble->T++;

        // Obtain theta_i to achieve a false positive rate of f
        // Get all outputs and ...
        for (int k = 0; k < n_negs; k++)
            sc[k] = EnsembleOutput(&(multiensemble->ensemble[i]), n_seqs[list[k]], length, parm->base_learner, parm->use_salzberg_method);

        // ...sort them
        qsort(sc, n_negs, sizeof(scalar), CompareDouble);

        // Get threshold for desired false positive rate
        multiensemble->ensemble[i].threshold = sc[(int)(n_negs * (1.0 - f)) ];

        // Remove correctly classified instances from negatives samples
        // Include in list only misclassified instances and adjust n_negs accordingly
        int c = 0;
        for (int k = 0; k < n_negs; k++)
            if (EnsembleOutput(&(multiensemble->ensemble[i]), n_seqs[list[k]], length, parm->base_learner, parm->use_salzberg_method) > 0.0) {
                // Misclassified
                list[c] = list[k];
                c++;
            }
        n_negs = c;
    }

    FREE(list);
    FREE(seqs);
    FREE(w);
    FREE(ok);
    FREE(probs);
    FREE(sc);

    return multiensemble;
}
