/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <unistd.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// Functions for obtaining best percentage of features to keep using cross-validation
void CrossValidateFeatureSelection(t_Parameters *parm, char **pseqs, int n_pseqs, int length, char **nseqs, int n_nseqs, t_Feat *base_info, scalar step)
{
    scalar best_so_far, sensitivity, specificity, TP, TN, FP, FN, G = 0.0, best_percentage = 0.0;
    char **train_pos, **train_neg, **test_pos, **test_neg;
    t_RecognitionMethod method;

    // Do validation runs
    best_so_far = -1.0;

    // If we are dealing with and ensemble use base learner
    if (parm->r_method == BalanceCascadeEnsemble || parm->r_method == EasyEnsembleEnsemble)
        method = parm->base_learner;
    else
        method = parm->r_method;

    // Space for temporal data
    char **positives, **negatives;
    int n_features;

    MALLOC(positives, char *, n_pseqs, "CrossValidateFeatureSelection: positives");
    MALLOC(negatives, char *, n_nseqs, "CrossValidateFeatureSelection: negatives");
    MALLOC(train_pos, char *, n_pseqs, "CrossValidateFeatureSelection: train_pos");
    MALLOC(train_neg, char *, n_nseqs, "CrossValidateFeatureSelection: train_neg");
    MALLOC(test_pos, char *, n_pseqs, "CrossValidateFeatureSelection: test_pos");
    MALLOC(test_neg, char *, n_nseqs, "CrossValidateFeatureSelection: test_neg");
    for (int i = 0; i < n_pseqs; i++)
        MALLOC(positives[i], char, length + 1, "CrossValidateFeatureSelection: positives[i]");
    for (int i = 0; i < n_nseqs; i++)
        MALLOC(negatives[i], char, length + 1, "CrossValidateFeatureSelection: negatives[i]");

    // Train classifier
    switch (method) {
        case SVM:
            // MODIFIED 4/November/2010
            // G-measure is used to parameter setting
            for (scalar percent = parm->fs_min_percent; percent <= 1.0; percent += step) {
                struct svm_model *lsvm = NULL;

                // Select features
                // Positive sequences
                n_features = length * percent;
                for (int i = 0; i < n_pseqs; i++) {
                    for (int j = 0; j < n_features; j++)
                        positives[i][j] = pseqs[i][base_info[j].index];

                    positives[i][n_features] = '\0';
                }

                // Negative sequences
                for (int i = 0; i < n_nseqs; i++) {
                    for (int j = 0; j < n_features; j++)
                        negatives[i][j] = negatives[i][base_info[j].index];

                    negatives[i][n_features] = '\0';
                }

                int p_test_size, n_test_size;
                p_test_size = n_pseqs/parm->cv_folds;
                n_test_size = n_nseqs/parm->cv_folds;

                TN = TP = FP = FN = 0.0;

                for (int fold = 0; fold < parm->cv_folds; fold++) {
                    // Create train and test sets
                    int i_pos = p_test_size * fold;
                    for (int j = 0; j < p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        test_pos[j] = positives[i_pos];
                    for (int j = 0; j < n_pseqs - p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        train_pos[j] = positives[i_pos];
                    int i_neg = n_test_size * fold;
                    for (int j = 0; j < n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        test_neg[j] = negatives[i_neg];
                    for (int j = 0; j < n_nseqs - n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        train_neg[j] = negatives[i_neg];

                    // Train and evaluate classifier
                    // Learn model
                    if (parm->cv_parameters)
                        CrossValidatedParameters(parm, train_pos, n_pseqs - p_test_size, length, train_neg, n_nseqs - p_test_size);

                    TrainSVMClassifier(parm, &lsvm, train_pos, n_pseqs - p_test_size, n_features, train_neg, n_nseqs - n_test_size, false, NULL, NULL, NULL, NULL);

                    // Positive sequences
                    for (int j = 0; j < p_test_size; j++)
                        if (SVMOutput(lsvm, test_pos[j], n_features, NULL, parm->use_salzberg_method, NULL) > 0.0)
                            TP++;
                        else
                            FN++;

                    // Negative sequences
                    for (int j = 0; j < n_test_size; j++)
                        if (SVMOutput(lsvm, test_neg[j], n_features, NULL, parm->use_salzberg_method, NULL) <= 0.0)
                            TN++;
                        else
                            FP++;
                    svm_free_and_destroy_model(&lsvm);
                }

                // Better than current best?
                sensitivity = TP / (TP + FN);
                specificity = TN / (TN + FP);
                G = pow(sensitivity * specificity, 0.5);

                if (G >= best_so_far) {
                    best_so_far = G;
                    best_percentage = percent;
                }

                if (parm->verbose > 1)
                    fprintf(parm->out_file, "Percentage = %lf\tG: %lf\n", percent, G);
            }

            parm->fs_percent_to_keep = best_percentage;

            // Accuracy in percentage
            fprintf(parm->out_file, "Best parameters (accuracy: %lf): Kernel = %d\tC = %lf\tGamma: %lf\n", G, parm->kernel, parm->c, parm->gamma);
            break;

        case NoOne:
        case kNN:
        case C45:
        case STOP:
        case EasyEnsembleEnsemble:
        case BalanceCascadeEnsemble:
        case PWM:
        case NeuralNet:
        case LearningFromBites:
            // No validation yet
            OnErrorExit(1, "No cross-validation for these learners");
            break;
    }

    for (int i = 0; i < n_pseqs; i++)
        FREE(positives[i]);
    for (int i = 0; i < n_nseqs; i++)
        FREE(negatives[i]);
    FREE(positives);
    FREE(negatives);
    FREE(train_pos);
    FREE(train_neg);
    FREE(test_pos);
    FREE(train_neg);
}
