/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <unistd.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// Functions for obtaining training parameters using cross-validation

#define SVM_PARAM_SETS  21

// TODO: Include validation of string kernels


void CrossValidatedParameters(t_Parameters *parm, char **pseqs, int n_pseqs, int length, char **nseqs, int n_nseqs)
{
    scalar best_so_far, sensitivity, specificity, TP, TN, FP, FN, G;
    char **train_pos, **train_neg, **test_pos, **test_neg;
    int best_index = 0;
    t_Recognizer R;
    t_RecognitionMethod method;

    // Sets of parameters
    scalar svm_p[SVM_PARAM_SETS][3] = {{2.0, 0.0001, 0.1}, {2.0, 0.0001, 1.0}, {2.0, 0.0001, 10.0}, {2.0, 0.001, 0.1}, {2.0, 0.001, 1.0}, {2.0, 0.001, 10.0},
                                       {2.0, 0.01, 0.1}, {2.0, 0.01, 1.0}, {2.0, 0.01, 10.0}, {2.0, 0.1, 0.1}, {2.0, 0.1, 1.0}, {2.0, .1, 10.0},
                                       {2.0, 1.0, 0.1}, {2.0, 1.0, 1.0}, {2.0, 1.0, 10.0}, {2.0, 10.0, 0.1}, {2.0, 10.0, 1.0}, {2.0, 10.0, 10.0},
                                       {0.0, 0.0, 0.1}, {0.0, 0.0, 1.0}, {0.0, 0.0, 10.0}};

    // Do validation runs
    best_so_far = -1.0;

    // If we are dealing with and ensemble use base learner
    if (parm->r_method == BalanceCascadeEnsemble || parm->r_method == EasyEnsembleEnsemble)
        method = parm->base_learner;
    else
        method = parm->r_method;

    MALLOC(train_pos, char *, n_pseqs, "CrossValidateFeatureSelection: train_pos");
    MALLOC(train_neg, char *, n_nseqs, "CrossValidateFeatureSelection: train_neg");
    MALLOC(test_pos, char *, n_pseqs, "CrossValidateFeatureSelection: test_pos");
    MALLOC(test_neg, char *, n_nseqs, "CrossValidateFeatureSelection: test_neg");

    // Train classifier
    switch (method) {
        case SVM:
            // MODIFIED 4/November/2010
            // G-measure is used to parameter setting
            for (int i = 0; i < SVM_PARAM_SETS; i++) {
                struct svm_model *lsvm = NULL;

                parm->kernel = (int) svm_p[i][0];
                parm->gamma = svm_p[i][1];
                parm->c = svm_p[i][2];

                // Train and evaluate classifier
                int p_test_size, n_test_size;
                p_test_size = n_pseqs / parm->cv_folds;
                n_test_size = n_nseqs / parm->cv_folds;

                TN = TP = FP = FN = 0.0;

                for (int fold = 0; fold < parm->cv_folds; fold++) {
                    // Create train and test sets
                    int i_pos = p_test_size * fold;
                    for (int j = 0; j < p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        test_pos[j] = pseqs[i_pos];
                    for (int j = 0; j < n_pseqs - p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        train_pos[j] = pseqs[i_pos];
                    int i_neg = n_test_size * fold;
                    for (int j = 0; j < n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        test_neg[j] = nseqs[i_neg];
                    for (int j = 0; j < n_nseqs - n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        train_neg[j] = nseqs[i_neg];

                    // Train and evaluate classifier
                    // Learn model
                    TrainSVMClassifier((const t_Parameters *) parm, &lsvm, train_pos, n_pseqs - p_test_size, length, train_neg, n_nseqs - n_test_size, false, NULL, NULL, NULL, NULL);

                    // Positive sequences
                    for (int j = 0; j < p_test_size; j++)
                        if (SVMOutput(lsvm, test_pos[j], length, NULL, parm->use_salzberg_method, NULL) > 0.0)
                            TP++;
                        else
                            FN++;

                    // Negative sequences
                    for (int j = 0; j < n_test_size; j++)
                        if (SVMOutput(lsvm, test_neg[j], length, NULL, parm->use_salzberg_method, NULL) <= 0.0)
                            TN++;
                        else
                            FP++;
                    svm_free_and_destroy_model(&lsvm);
                }

                // Better than current best?
                sensitivity = TP / (TP + FN);
                specificity = TN / (TN + FP);
                G = pow(sensitivity * specificity, 0.5);
                if (G > best_so_far) {
                    best_so_far = G;
                    best_index = i;
                } else if (G == best_so_far) {
                    // In case of a tie choose higher parm->c
                    if (svm_p[best_index][2] < svm_p[i][2])
                        best_index = i;
                }

                if (parm->verbose > 1)
                    fprintf(parm->out_file, "Kernel = %d\tC = %lf\tgamma = %lf\tAccuracy: %lf\n", parm->kernel, parm->c, parm->gamma, G);
            }

            // Accuracy in percentage
            parm->kernel = (int) svm_p[best_index][0];
            parm->gamma = svm_p[best_index][1];
            parm->c = svm_p[best_index][2];
            fprintf(parm->out_file, "Best parameters (accuracy: %lf): Kernel = %d\tC = %lf\tGamma: %lf\n", best_so_far, parm->kernel, parm->c, parm->gamma);
            break;

        case kNN:
            for (int i = parm->min_k; i <= parm->max_k; i++) {
                R.k = i;

                // Train and evaluate classifier
                int p_test_size, n_test_size;
                p_test_size = n_pseqs / parm->cv_folds;
                n_test_size = n_nseqs / parm->cv_folds;

                // Evaluate classifier
                TN = TP = FP = FN = 0.0;

                for (int fold = 0; fold < parm->cv_folds; fold++) {
                    // Create train and test sets
                    int i_pos = p_test_size * fold;
                    for (int j = 0; j < p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        test_pos[j] = pseqs[i_pos];
                    for (int j = 0; j < n_pseqs - p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                        train_pos[j] = pseqs[i_pos];
                    int i_neg = n_test_size * fold;
                    for (int j = 0; j < n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        test_neg[j] = nseqs[i_neg];
                    for (int j = 0; j < n_nseqs - n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                        train_neg[j] = nseqs[i_neg];

                    R.n_poss = n_pseqs - p_test_size;
                    R.n_negs = n_nseqs - n_test_size;
                    R.poss = train_pos;
                    R.negs = train_neg;

                    // Positive sequences
                    for (int j = 0; j < p_test_size; j++)
                        if (ClassifySequencekNN(&R, test_pos[j], length, parm->knn_weight_by_distance) > 0.0)
                            TP++;
                        else
                            FN++;

                    // Negative sequences
                    for (int j = 0; j < n_test_size; j++)
                        if (ClassifySequencekNN(&R, test_neg[j], length, parm->knn_weight_by_distance) <= 0.0)
                            TN++;
                        else
                            FP++;
                }

                // Better than current best?
                sensitivity = TP / (TP + FN);
                specificity = TN / (TN + FP);
                G = pow(sensitivity * specificity, 0.5);
                if (G > best_so_far) {
                    best_so_far = G;
                    best_index = i;
                }

                if (parm->verbose > 1)
                    fprintf(parm->out_file, "parm->k = %d\tAccuracy: %lf\n", i, G);
            }

            // Accuracy in percentage
            parm->k = best_index;
            fprintf(parm->out_file, "Best parameters (accuracy: %lf): parm->k = %d\n", best_so_far, parm->k);
            break;

        case NoOne:
        case C45:
        case STOP:
        case EasyEnsembleEnsemble:
        case BalanceCascadeEnsemble:
        case PWM:
        case NeuralNet:
        case LearningFromBites:
            // No validation yet
            OnErrorExit(1, "No cross-validation for these learners");
            break;
    }

    FREE(train_pos);
    FREE(train_neg);
    FREE(test_pos);
    FREE(train_neg);
}
