/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
// #include "common.h"

// Implements EasyEnsembleEnsemble method from:
// X-Y. Liu, J. Wu and Z.-H. Zhou, "Exploratory undersampling for class-imbalance learning", IEEE Transactions on Systems, Man, and Cybernetics - Part B: Cybernetics, vol. 39, no. 2, April 2009, pp. 539-548.

// Due to computational cost, C4.5 is chosen as base learner
// Uses resampling by default for AdaBoost

// #define DEFAULT_THRESHOLD  0.5

// MODIFICATION 11/NOVEMBER/2010
// Added SVM as base classifier together with C4.5

t_MultiEnsemble *TrainEasyEnsemble(const t_Parameters *parm, char **p_seqs, char **n_seqs, int length, int n_poss, int n_negs, int T, int s_i)
{
    int RouletteSelection(scalar *, int);

    t_MultiEnsemble *multiensemble;
    int *list;
    char **seqs;
    scalar *probs, betta, e_i, *w;  // Weights for adaboost
    bool *ok;

    // Vector for random sampling
    MALLOC(list, int, n_negs, "TrainEasyEnsemble: list");
    for (int j = 0; j < n_negs; j++)
        list[j] = j;

    // Data for local training of trees and boosting
    MALLOC(seqs, char *, n_poss, "TrainEasyEnsemble: seqs");
    MALLOC(w, scalar, 2 * n_poss, "TrainEasyEnsemble: w");
    MALLOC(ok, bool, 2 * n_poss, "TrainEasyEnsemble: ok");
    MALLOC(probs, scalar, n_poss, "TrainEasyEnsemble: probs");

    // As many trees as classifiers
    MALLOC(multiensemble, t_MultiEnsemble, 1, "TrainEasyEnsemble: multiensemble");
    MALLOC(multiensemble->ensemble, t_Ensemble, T, "TrainEasyEnsemble: multiensemble->ensemble");
    multiensemble->T = T;
    for (int i = 0; i < T; i++) {
        multiensemble->ensemble[i].s = s_i;
        multiensemble->ensemble[i].threshold = 0.0;
        MALLOC(multiensemble->ensemble[i].alpha, scalar, s_i, "TrainEasyEnsemble: multiensemble->ensemble[i].alpha");
        switch (parm->base_learner) {
            case C45:
                MALLOC(multiensemble->ensemble[i].trees, Tree, s_i, "TrainEasyEnsemble: multiensemble->ensemble[i].trees");
                break;
            case SVM:
                MALLOC(multiensemble->ensemble[i].lsvm, struct svm_model *, s_i, "TrainEasyEnsemble: multiensemble->ensemble[i].lsvm");
                break;
            default:
                OnErrorExit(1, "Bad base learner");
        }
    }

    // Construct T AdaBoost ensembles of size s_i randomly undersampling
    for (int i = 0; i < T; i++) {

        // Random undersampling
        GetRandomisedVector(list, n_negs);

        // Initial weights
        for (int j = 0; j < 2*n_poss; j++)
            w[j] = 1.0;

        // Random chosen instances
        for (int j = 0; j < n_poss; j++)
            seqs[j] = n_seqs[list[j]];

        // Construct ensemble
        for (int j = 0; j < s_i; j++) {
            if (parm->base_learner == C45)
                // Train and set threshold
                multiensemble->ensemble[i].trees[j] = TrainC45Classifier(parm, p_seqs, n_poss, length, seqs, n_poss, false, NULL);
            else {
                if (parm->cv_parameters)
                    CrossValidatedParameters((t_Parameters *) parm, p_seqs, n_poss, length, seqs, n_poss);
                TrainSVMClassifier(parm, &(multiensemble->ensemble[i].lsvm[j]), p_seqs, n_poss, length, seqs, n_poss, false, NULL, NULL, NULL, NULL);
            }

            // Test and update weights
            // Positive samples
            e_i = 0.0;
            if (parm->base_learner == C45) {
                for (int k = 0; k < n_poss; k++)
                    if (C45Output(multiensemble->ensemble[i].trees[j], p_seqs[k], length) > 0.0)
                        // Ok
                        ok[k] = true;
                    else {
                        // Misclassified
                        ok[k] = false;
                        e_i += w[k];
                    }

                // Negative samples
                for (int k = 0; k < n_poss; k++)
                    if (C45Output(multiensemble->ensemble[i].trees[j], n_seqs[list[k]], length) < 0.0)
                        // Ok
                        ok[k + n_poss] = true;
                    else {
                        // Misclassified
                        ok[k + n_poss] = false;
                        e_i += w[k + n_poss];
                    }
            } else if (parm->base_learner == SVM) {
                for (int k = 0; k < n_poss; k++)
                    if (SVMOutput(multiensemble->ensemble[i].lsvm[j], p_seqs[k], length, NULL, parm->use_salzberg_method, NULL) > 0.0)
                        // Ok
                        ok[k] = true;
                    else {
                        // Misclassified
                        ok[k] = false;
                        e_i += w[k];
                    }

                // Negative samples
                for (int k = 0; k < n_poss; k++)
                    if (SVMOutput(multiensemble->ensemble[i].lsvm[j],n_seqs[list[k]], length, NULL, parm->use_salzberg_method, NULL) < 0.0)
                        // Ok
                        ok[k + n_poss] = true;
                    else {
                        // Misclassified
                        ok[k + n_poss] = false;
                        e_i += w[k + n_poss];
                    }
            }
            // Total error
            e_i /= 2.0 * n_poss;

            if (parm->verbose > 1) {
                if (parm->base_learner == C45)
                    fprintf(parm->out_file, "Classifier %d/%d of ensemble %d/%d error: %lf\t size: %d\n", j+1, s_i, i+1, T, e_i, TreeSize(multiensemble->ensemble[i].trees[j]));
                else if (parm->base_learner == SVM)
                    fprintf(parm->out_file, "Classifier %d/%d of ensemble %d/%d error: %lf\t support vectors: %d\n", j+1, s_i, i+1, T, e_i, SVMSize(multiensemble->ensemble[i].lsvm[j]));
            }

            // If above 50% random resampling
            if (e_i > 0.5) {
                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = 0.0;

                betta = 1.0; // Remove classifier assigning 0 weight
                for (int k = 0; k < 2 * n_poss; k++)
                    w[k] = 1.0;

            } else if (e_i == 0.0) {   // If perfect also random resampling
                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = 10.0;

                betta = 1.0e-10;
                for (int k = 0; k < 2 * n_poss; k++)
                    w[k] = 1.0;
            } else {
                betta = e_i / (1.0 - e_i);

                // Weight classifier
                multiensemble->ensemble[i].alpha[j] = log(1.0 / betta);

                for (int k = 0; k < 2 * n_poss; k++)
                    if (ok[k])
                        w[k] /= 2.0 * (1.0 - e_i);
                    else
                        w[k] /= 2.0 * e_i;
            }

            // Resample for next step
            probs[0] = w[0];
            for (int k = 1; k < n_poss; k++)
                probs[k] = probs[k - 1] + w[k];

            for (int k = 0; k < n_poss; k++) {
                int pattern = RouletteSelection(probs, n_poss);

                seqs[k] = n_seqs[list[pattern]];
            }

            // Normalize w[] to sum n
            scalar sum = 0.0;
            for (int k = 0; k < 2*n_poss; k++)
                sum += w[k];
            for (int k = 0; k < 2*n_poss; k++)
                w[k] *= (2.0*n_poss) /sum;

            if (parm->verbose > 2) {
                fprintf(parm->out_file, "Positive instances boosting weights: ");
                for (int k = 0; k < n_poss; k++)
                    fprintf(parm->out_file, "%6.3lf ", w[k]);
                fprintf(parm->out_file, "\nNegative instances boosting weights: ");
                for (int k = n_poss; k < 2*n_poss; k++)
                    fprintf(parm->out_file, "%6.3lf ", w[k]);
                fprintf(parm->out_file, "\n");
            }
        }

        // Get ensemble threshold if needed
        if (parm->sensitivity > 0.0) {
            int CompareDouble(const void *, const void *);

            // Threshold for desired sensitivity
            // Get threshold to obtain desired sensitivity = TP/(TP+FN)
            // Get all outputs and ...
            scalar *sc;
            ALLOCA(sc, scalar, n_poss, "TrainEasyEnsemble: sc");
            for (int k = 0; k < n_poss; k++)
                sc[k] = EnsembleOutput(& (multiensemble->ensemble[i]), p_seqs[k], length, parm->base_learner, parm->use_salzberg_method);

            // ...sort them
            qsort(sc, n_poss, sizeof(scalar), CompareDouble);

            // Get threshold for desired sensitivity
            multiensemble->ensemble[i].threshold = sc[(int)(n_poss * (1.0 - parm->sensitivity)) ];
        }
    }

    FREE(list);
    FREE(seqs);
    FREE(w);
    FREE(ok);
    FREE(probs);

    return multiensemble;
}

scalar EnsembleOutput(t_Ensemble *ensemble, char *seq, int length, t_RecognitionMethod base_learner, bool use_salzberg_method)
{
    scalar output = 0.0;

    // Output of every classifier
    if (base_learner == C45)
        for (int i = 0; i < ensemble->s; i++)
            output += ensemble->alpha[i] * C45Output(ensemble->trees[i], seq, length);
    else if (base_learner == SVM)
        for (int i = 0; i < ensemble->s; i++)
            output += ensemble->alpha[i] * SVMOutput(ensemble->lsvm[i], seq, length, NULL, use_salzberg_method, NULL);

    output -= ensemble->threshold;

    return output;
}

scalar MultiEnsembleOutput(t_MultiEnsemble *multiensemble, char *seq, int length, t_RecognitionMethod base_learner, bool use_salzberg_method)
{
    scalar output = 0.0;

    // Output of every classifier
    for (int i = 0; i < multiensemble->T; i++)
        output += EnsembleOutput(& (multiensemble->ensemble[i]), seq, length, base_learner, use_salzberg_method);

    return output/multiensemble->T;
}


void SetMultiEnsembleThreshold(const t_Parameters *parm, t_Recognizer *R, char **positives, int n_poss, int length)
{
    int CompareDouble(const void *, const void *);

    // Get threshold to obtain desired sensitivity = TP/(TP+FN)
    // Get all differences and ...
    scalar *sc;
    ALLOCA(sc, scalar, n_poss, "SetMultiEnsembleThreshold: sc");
    for (int i = 0; i < n_poss; i++)
        if (parm->r_method == EasyEnsembleEnsemble)
            sc[i] = MultiEnsembleOutput(R->easy, positives[i], length, parm->base_learner, parm->use_salzberg_method);
        else
            sc[i] = MultiEnsembleOutput(R->cascade, positives[i], length, parm->base_learner, parm->use_salzberg_method);

    // ...sort them
    qsort(sc, n_poss, sizeof(scalar), CompareDouble);

    // Output outputs for all positive instances
    if (parm->verbose > 2) {
        fprintf(parm->out_file, "Training positive instances ensemble outputs\n");
        for (int i = 0; i < n_poss; i++)
            fprintf(parm->out_file, "%6.3lf ", sc[i]);
        fprintf(parm->out_file, "\n");
    }

    // Get threshold for desired sensitivity
    R->threshold = sc[(int)(n_poss * (1.0 - parm->sensitivity)) ];

    fprintf(parm->out_file, "Ensemble threshold: %lf\n", R->threshold);
}

int RouletteSelection(scalar *prob, int n)
{
    int being;
    scalar uniform;

    // An item is selected using a random value
    uniform = RANDOM(0, prob[n - 1]);
    being = 0;
    while (uniform > prob[being])
        being++;

    return (being);
}
