/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// Functions that deal with FASTA files

void ReadFastaFile(char *f_name, char ***data, int *sequences, int *length, scalar sampling_ratio, bool undersampling, int n_positives, bool just_count, int *selected_list,
                   int selected_instances, bool allow_different_lengths, int offset, int global_max_length, scalar global_sampling_ratio, bool allow_only_5_letters,
                   bool do_not_check_sequences, char *letters, int *global_length, char ***labels, bool compress)
{
    char *s, *seq;
    FILE *f;
    int max_length, current = 0;
    bool on_the_list;

    // Allocate memory
    MALLOC(s, char, global_max_length+1, "ReadFastaFile: s");
    MALLOC(seq, char, global_max_length+1, "ReadFastaFile: seq");

    // Count sequences
    *sequences = 0;
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_FILE, "Sequence file not found");

    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");
    while (!feof(f)) {
        if (s[0] == '>')   // New sequence
            (*sequences)++;
        if (fgets(s, global_max_length, f) == NULL && !feof(f))
            OnErrorExit(ERR_FILE, "Error reading FASTA file");
    }

    // If undersampling adjust sampling ratio
    if (undersampling)
        sampling_ratio = global_sampling_ratio * (((scalar) n_positives)/(*sequences));

    // Get sequence length
    rewind(f);
    // Read first sequence
    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");
    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");
    max_length = 0;
    while (s[0] != '>' && !feof(f)) {
        // Remove new-line at the end
        REMOVE_NEWLINE(s);
        max_length += strlen(s);

        if (fgets(s, global_max_length, f) == NULL && !feof(f))
            OnErrorExit(ERR_FILE, "Error reading FASTA file");
    }

    // Check valid parameters
    if (!allow_different_lengths && (offset + *global_length > max_length)) {
	fprintf(stderr, "Offset: %d, length: %d, maximum length: %d\n", offset, *global_length, max_length);
        OnErrorExit(ERR_FILE, "Sequence too short");
    } else {
        if (*global_length == 0)
            *global_length = *length = max_length - offset;
        else
            *length = *global_length;
    }

    // Allocate memory
    if (!just_count) {
        MALLOC(*data, char *, *sequences, "ReadFastaFile: *data");

        // If requested allocate memory for labels
        if (labels != NULL)
            MALLOC(*labels, char *, *sequences, "ReadFastaFile: *labels");
    }

    // Read every sequence
    rewind(f);
    // Read first "> label"
    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");

    // Store label?
    if (labels != NULL) {
        REMOVE_NEWLINE(s);
        MALLOC((*labels)[0], char, strlen(s)+1, "ReadFastaFile: (*labels)[0]");
        strcpy((*labels[0]), s);
    }

    // If selected list adjust sampling ration
    if (selected_list)
        sampling_ratio *= (*sequences)/((scalar) selected_instances);

    // If allow different lengths, length will be the maximum length
    if (allow_different_lengths)
        *length = 0;

    if (labels && (sampling_ratio < 1.0 || selected_list))
        OnErrorExit(ERR_FILE, "Labels incompatible with sampling and selected list");

    for (int i = 0; i < *sequences;) {
        int cursor = 0, current_length = 0;

        // Read sequence line by line
        if (fgets(s, global_max_length, f) == NULL)
            OnErrorExit(ERR_FILE, "Error reading FASTA file");

        while (s[0] != '>' && !feof(f)) {
            // Remove new-line at the end
            REMOVE_NEWLINE(s);

            int l = strlen(s);

            if (!allow_different_lengths) {
                if (offset < cursor+l && current_length < *length) {
                    for (int j = MAX(offset - cursor, 0); j < l && current_length < *length; j++, current_length++) {
                        seq[cursor-offset+j] = toupper(s[j]);
                        // Convert to DNA to avoid double checks
                        if (seq[cursor-offset+j] == 'U')
                            seq[cursor-offset+j] = 'T';
                    }
                }
            } else {
                if (offset < cursor+l && current_length < global_max_length) {
                    for (int j = MAX(offset - cursor, 0); j < l && current_length < global_max_length; j++, current_length++) {
                        seq[cursor-offset+j] = toupper(s[j]);
                        // Convert to DNA to avoid double checks
                        if (seq[cursor-offset+j] == 'U')
                            seq[cursor-offset+j] = 'T';
                    }
                }
            }

            cursor += l;

            if (fgets(s, global_max_length, f) == NULL && !feof(f))
                OnErrorExit(ERR_FILE, "Error reading FASTA file");
        }

        // Store label?
        if (labels != NULL && !feof(f)) {
            REMOVE_NEWLINE(s);
            MALLOC((*labels)[i+1], char, strlen(s)+1, "ReadFastaFile: (*labels)[i]");
            strcpy((*labels)[i+1], s);
        }

        seq[current_length] = '\0';

        if (allow_different_lengths) {
            if (current_length > *length)
                *length = current_length;
        }

        // Allow more than four letters?
        if (allow_only_5_letters && !do_not_check_sequences) {
            bool more_than_five = false;
            for (int j = 0; j < strlen(seq) && !more_than_five; j++)
                if (seq[j] != 'A' && seq[j] != 'T' && seq[j] != 'G' && seq[j] != 'C' && seq[j] != 'U' && seq[j] != 'N')
                    seq[j] = 'N';
        } else if (!do_not_check_sequences) {
            for (int l = 0; l < strlen(seq); l++) {
                bool ok = false;
                for (int j = 0; j < strlen(letters); j++)
                    if (seq[l] == letters[j]) {
                        ok = true;
                        break;
                    }

                if (!ok)
                    OnErrorExit(ERR_FILE, "Sequence contains unknown letters");
            }
        }

        // Not in the selected list?
        if (!selected_list)
            on_the_list = true;
        else {
            if (selected_list[current] == i+1) {
                on_the_list = true;
                current++;
            } else
                on_the_list = false;
        }

        // If sampling test whether to forget this sequence
        if (just_count)
            i++;
        else {
            if (RANDOM(0.0, 1.0) > sampling_ratio || !on_the_list)
                (*sequences)--;
            else {
                MALLOC((*data)[i], char, strlen(seq)+1, "ReadFastaFile: (*data)[i]");
                strcpy((*data)[i], seq);

                // Compress?
                if (compress) {
                    char original_sequence[MAX_SEQUENCE_LENGTH], uncompressed_sequence[MAX_SEQUENCE_LENGTH];

                    strcpy(original_sequence, (*data)[i]);
                    CompressSequence(&(*data)[i]);

                    // Check compression Ok, if not raise error
                    UncompressSequence((*data)[i], uncompressed_sequence, *length);
                    if (strcmp(original_sequence, uncompressed_sequence)) {
                        printf("Original: %s\nCompressed: %s\n", original_sequence, uncompressed_sequence);
                        OnErrorExit(ERR_FILE, "Error in sequence compression");
                    }
                }

                i++;
            }
        }
    }

    fclose(f);

    // If undersampling or random sampling
    if (!just_count)
        REALLOC(*data, char *, *sequences, "ReadFastaFile: *data");

    FREE(s);
    FREE(seq);
}

int CompareInteger(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (int *) x - * (int *) y) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Read the raw sequence of the file
void ReadRawFastaFile(char *f_name, char **sequence, int *length, int global_max_length)
{
    char s[global_max_length];
    FILE *f;
    int cur_length;

    // Get length
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_FILE, "Sequence file not found");

    *length = 0;
    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");
    if (s[0] == '>')   // Skip ID
        if (fgets(s, global_max_length, f) == NULL)
            OnErrorExit(ERR_FILE, "Error reading FASTA file");
    while (!feof(f)) {
        // Remove new line character
        if (s[strlen(s)-1] == '\n')
            s[strlen(s)-1] = '\0';
        *length += strlen(s);
        if (fgets(s, global_max_length, f) == NULL && !feof(f))
            OnErrorExit(ERR_FILE, "Error reading FASTA file");
    }

    // Allocate memory
    MALLOC(*sequence, char, *length+1, "ReadRawFastaFile: *length");

    // Read raw sequence
    rewind(f);
    // Read first "> label"
    if (fgets(s, global_max_length, f) == NULL)
        OnErrorExit(ERR_FILE, "Error reading FASTA file");
    if (s[0] == '>')   // Skip ID
        if (fgets(s, global_max_length, f) == NULL)
            OnErrorExit(ERR_FILE, "Error reading FASTA file");

    cur_length = 0;
    while (!feof(f)) {
        // Remove new line character
        if (s[strlen(s)-1] == '\n')
            s[strlen(s)-1] = '\0';
        strcpy(*sequence+cur_length, s);
        cur_length += strlen(s);
        if (fgets(s, global_max_length, f) == NULL && !feof(f))
            OnErrorExit(ERR_FILE, "Error reading FASTA file");
    }

    fclose(f);
}
