/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Functions for GA-based feature selection
#include "macros.h"
#include "type.h"

// Structure for ordering the size
typedef struct {
    int being;
    scalar fitness;
} RANK;

void GAFeatureSelection(const t_Parameters *parm, int *S, char **positives, int n_positives, char **negatives, int n_negatives, int features, int size, int generations, scalar probability,
                        scalar mutation, scalar bit_mutation, struct spectrum_kernel_info *spectrum)
{
    int **population, actual_size;
    scalar *fitness, *G, *reduction;

    // Allocate and create initial population
    ALLOCA(population, int *, 2*size + 1, "GAFeatureSelection: population");
    for (int i = 0; i < size; i++) {
        ALLOCA(population[i], int, features, "GAFeatureSelection: population[i]");
        for (int j = 0; j < features; j++)
            population[i][j] = RANDOM(0.0, 1.0) < probability ? 1 : 0;
    }
    for (int i = size; i < 2*size + 1; i++)
        ALLOCA(population[i], int, features, "GAFeatureSelection: population[i]");

    // Initial evaluation
    ALLOCA(fitness, scalar, 2*size + 1, "GAFeatureSelection: fitness");
    ALLOCA(G, scalar, 2*size + 1, "GAFeatureSelection: G");
    ALLOCA(reduction, scalar, 2*size + 1, "GAFeatureSelection: reduction");
    for (int i = 0; i < size; i++)
        fitness[i] = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, population[i], parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[i], &reduction[i], spectrum);

    // For selection without replacement
    int *list;
    ALLOCA(list, int, 2*size, "GAFeatureSelection: list");
    for (int i = 0; i < 2 * size; i++)
        list[i] = i;

    // Evolution
    int min_dist = features / 4;
    for (int g = 0; g < generations; g++) {
        // CHC next generation
        actual_size = size;

        // Perform HUX crossover
        for (int i = 0; i < size; i += 2) {
            // Random selection without replacement
            int order = RANDOM(i, size);
            int father = list[order];
            list[order] = list[i];
            list[i] = father;
            order = RANDOM(i + 1, size);
            int mother = list[order];
            list[order] = list[i + 1];
            list[i + 1] = mother;

            // If no minimum distance skip
            if (HammingDistanceInteger(population[mother], population[father], features)  < min_dist)
                continue;

            HuxCrossover(population[mother], population[father], population[size + i], population[size + i + 1], features);

            // Two new size in the population.
            actual_size += 2;
        }

        // If no crossover decrement threshold and retry.
        if (actual_size == size) {
            min_dist--;
            continue;
        }

        // Introduce random mutation if required
        if (mutation > ZERO && bit_mutation > ZERO)
            for (int i = 1; i < actual_size; i++)   // Do not mutate best one
                if (RANDOM(0.0, 1.0) < mutation) {
                    if (RandomMutation(population[i], features, bit_mutation) && i < size) {      // Reevaluate if necessary
                        fitness[i] = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, population[i], parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[i], &reduction[i],
                                                        spectrum);
                    }
                }

        // Evaluate new individuals
        for (int k = size; k < actual_size; k++)
            fitness[k] = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, population[k], parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[k], &reduction[k], spectrum);

        // Local search
        if (parm->gafs_lsmutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->gafs_lsmutation)
                    fitness[i] = LSMutation(parm, population[i], features, positives, n_positives, negatives, n_negatives, parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[i], &reduction[i], spectrum);

        // Always do local search with the best
        int best = OrderedSelection(fitness, 0, actual_size);
        for (int i = 0; i < features; i++)
            population[actual_size][i] = population[best][i];
        fitness[actual_size] = LSMutation(parm, population[actual_size], features, positives, n_positives, negatives, n_negatives, parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[actual_size],
                                          &reduction[actual_size], spectrum);
        actual_size++;

        // Obtain new population selecting best size.
        for (int i = 0; i < size; i++) {
            int order = OrderedSelection(fitness + i, 0, actual_size - i) + i;
            int *aux_pt = population[order];
            population[order] = population[i];
            population[i] = aux_pt;
            scalar aux_f = fitness[order];
            fitness[order] = fitness[i];
            fitness[i] = aux_f;
            aux_f = G[order];
            G[order] = G[i];
            G[i] = aux_f;
            aux_f = reduction[order];
            reduction[order] = reduction[i];
            reduction[i] = aux_f;
        }

        // If minimum distance 0 re-initialize
        if (min_dist == 0) {
            // Reinitialize populations excepting best one
            int best = OrderedSelection(fitness, 0, size);
            for (int i = 0; i < size; i++)
                if (i != best) {
                    for (int k = 0; k < features; k++)
                        population[i][k] = RANDOM(0.0, 1.0) < probability ? 1 : 0;
                    fitness[i] = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, population[i], parm->r_method, parm->cv_folds, parm->gafs_alpha, &G[i], &reduction[i], spectrum);
                }
        }

        if (parm->verbose > 0) {
            fprintf(parm->out_file, "Generation:%3d. Best fitness: %6.4lf\tG: %6.4lf\tReduction: %6.4lf\t", g+1, fitness[0], G[0], reduction[0]);
            for (int i = 0; i < features; i++)
                fprintf(parm->out_file, "%d ", population[0][i]);
            fprintf(parm->out_file, "\n");
            fflush(parm->out_file);
        }
    }

    // Copy best to result
    for (int i = 0; i < features; i++)
        S[i] = population[0][i];
}

// Returns Hamming distance between integers
int HammingDistanceInteger(int *x, int *y, int l)
{
    int distance = 0.0;

    for (int i = 0; i < l; i++)
        if (x[i] != y[i])
            distance++;

    return distance;
}

int OrderedSelection(scalar *fitness, int n, int s)
{
    int Compare(const void *, const void *);

    int i, max_index;
    RANK ranking[s];

    // If the best just search the maximum
    if (n == 0) {
        max_index = 0;
        for (i = 1; i < s; i++)
            if (fitness[i] > fitness[max_index])
                max_index = i;

        return max_index;
    } else {
        // Ordered by fitness
        for (i = 0; i < s; i++) {
            ranking[i].being = i;
            ranking[i].fitness = fitness[i];
        }

        qsort(ranking, s, sizeof(RANK), Compare);

        // n-th individual
        return (ranking[n].being);
    }
}

// Auxiliary function for qsort.
int Compare(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((RANK *) y)->fitness - ((RANK *) x)->fitness) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}


void HuxCrossover(int *father, int *mother, int *off1, int *off2, int n)
{
    bool turn = true;

    for (int i = 0; i < n; i++)
        if (father[i] == mother[i])
            off1[i] = off2[i] = mother[i];
        else {
            if (turn) {
                off1[i] = father[i];
                off2[i] = mother[i];
                turn = false;
            } else {
                off1[i] = mother[i];
                off2[i] = father[i];
                turn = true;
            }
        }
}

bool RandomMutation(int *individual, int size, scalar bit_mutation)
{
    bool modified = false;

    for (int j = 0; j < size; j++)
        if (RANDOM(0.0, 1.0) < bit_mutation) {
            modified = true;

            individual[j] = (individual[j] + 1) % 2;
        }

    return modified;
}

// Functions for evaluating the individual using cross-validation
scalar EvaluateIndividual(const t_Parameters *parm, char **pseqs, int n_pseqs, int features, char **nseqs, int n_nseqs, int *S, t_RecognitionMethod method, int folds, scalar alpha, scalar *G, 
                          scalar *reduction, struct spectrum_kernel_info *spectrum)
{
    scalar sensitivity, specificity, TP, TN, FP, FN;
    char **train_pos, **train_neg, **test_pos, **test_neg;
    struct svm_model *lsvm = NULL;
    int selected = 0;

    // Reduction
    for (int i = 0; i < features; i++)
        if (S[i])
            selected++;

    *reduction = 1.0 - ((scalar) selected) / features;

    // If we are dealing with and ensemble use base learner
    // Space for temporal data
    char **positives, **negatives;

    MALLOC(positives, char *, n_pseqs, "CrossValidateFeatureSelection: positives");
    MALLOC(negatives, char *, n_nseqs, "CrossValidateFeatureSelection: negatives");
    MALLOC(train_pos, char *, n_pseqs, "CrossValidateFeatureSelection: train_pos");
    MALLOC(train_neg, char *, n_nseqs, "CrossValidateFeatureSelection: train_neg");
    MALLOC(test_pos, char *, n_pseqs, "CrossValidateFeatureSelection: test_pos");
    MALLOC(test_neg, char *, n_nseqs, "CrossValidateFeatureSelection: test_neg");
    for (int i = 0; i < n_pseqs; i++)
        MALLOC(positives[i], char, features + 1, "CrossValidateFeatureSelection: positives[i]");
    for (int i = 0; i < n_nseqs; i++)
        MALLOC(negatives[i], char, features + 1, "CrossValidateFeatureSelection: negatives[i]");

    // Train classifier
    switch (method) {
        case SVM:
            // Select features
            // Positive sequences
            for (int i = 0; i < n_pseqs; i++) {
                int index = 0;
                for (int j = 0; j < features; j++)
                    if (S[j]) {
                        positives[i][index] = pseqs[i][j];
                        index++;
                    }

                positives[i][index] = '\0';
            }

            // Negative sequences
            for (int i = 0; i < n_nseqs; i++) {
                int index = 0;
                for (int j = 0; j < features; j++)
                    if (S[j]) {
                        negatives[i][index] = nseqs[i][j];
                        index++;
                    }

                negatives[i][index] = '\0';
            }

            int p_test_size, n_test_size;
            p_test_size = n_pseqs / folds;
            n_test_size = n_nseqs / folds;

            TN = TP = FP = FN = 0.0;

            for (int fold = 0; fold < folds; fold++) {
                // Create train and test sets
                int i_pos = p_test_size * fold;
                for (int j = 0; j < p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                    test_pos[j] = positives[i_pos];
                for (int j = 0; j < n_pseqs - p_test_size; j++, i_pos = (i_pos + 1) % n_pseqs)
                    train_pos[j] = positives[i_pos];
                int i_neg = n_test_size * fold;
                for (int j = 0; j < n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                    test_neg[j] = negatives[i_neg];
                for (int j = 0; j < n_nseqs - n_test_size; j++, i_neg = (i_neg + 1) % n_nseqs)
                    train_neg[j] = negatives[i_neg];

                TrainSVMClassifier(parm, &lsvm, train_pos, n_pseqs - p_test_size, selected, train_neg, n_nseqs - n_test_size, false, NULL, NULL, NULL, spectrum);

                // Positive sequences
                for (int j = 0; j < p_test_size; j++)
                    if (SVMOutput(lsvm, test_pos[j], selected, NULL, parm->use_salzberg_method, &spectrum[j]) > 0.0)
                        TP++;
                    else
                        FN++;

                // Negative sequences
                for (int j = 0; j < n_test_size; j++)
                    if (SVMOutput(lsvm, test_neg[j], selected, NULL, parm->use_salzberg_method, &spectrum[j+p_test_size]) <= 0.0)
                        TN++;
                    else
                        FP++;

                svm_free_and_destroy_model(&lsvm);
            }

            // Get G
            sensitivity = TP / (TP + FN);
            specificity = TN / (TN + FP);
            *G = pow(sensitivity * specificity, 0.5);

            break;

        case NoOne:
        case kNN:
        case C45:
        case STOP:
        case EasyEnsembleEnsemble:
        case BalanceCascadeEnsemble:
        case PWM:
        case NeuralNet:
        case LearningFromBites:
            OnErrorExit(1, "Not yet implemented");
            break;
    }

    for (int i = 0; i < n_pseqs; i++)
        FREE(positives[i]);
    for (int i = 0; i < n_nseqs; i++)
        FREE(negatives[i]);
    FREE(positives);
    FREE(negatives);
    FREE(train_pos);
    FREE(train_neg);
    FREE(test_pos);
    FREE(train_neg);

    return alpha * *G + (1.0 - alpha) * *reduction ;
}

// Local search
scalar LSMutation(const t_Parameters *parm, int *S, int features, char **positives, int n_positives, char **negatives, int n_negatives, t_RecognitionMethod method, int folds, scalar alpha, scalar *G, 
                  scalar *reduction, struct spectrum_kernel_info *spectrum)
{
    scalar best = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, S, method, folds, alpha, G, reduction, spectrum);

    // Flip all bits and evaluate if improvement
    for (int i = 0; i < features; i++) {
        S[i] = (S[i] + 1) % 2;

        scalar current = EvaluateIndividual(parm, positives, n_positives, features, negatives, n_negatives, S, method, folds, alpha, G, reduction, spectrum);

        // Undo if worst
        if (current < best)
            S[i] = (S[i] + 1) % 2;
        else
            best = current;

    }

    return best;
}
