/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// k-NN functions for stacking method

int CrossValidatedkNNRealParameters(const t_Parameters *parm, scalar **data, int instances, int inputs, int *labels)
{
    int *test_label, *train_label, best_k = 0;
    scalar TN, TP, FP, FN, best_so_far, **train, **test;

    // Do validation runs
    best_so_far = -1.0;

    // Auxiliary data
    MALLOC(train, scalar *, instances, "CrossValidatedSVMRealParameters: train");
    MALLOC(test, scalar *, instances, "CrossValidatedSVMRealParameters: test");
    MALLOC(train_label, int, instances, "CrossValidatedSVMRealParameters: train_label");
    MALLOC(test_label, int, instances, "CrossValidatedSVMRealParameters: test_label");

    for (int k = parm->min_k; k <= parm->max_k; k++) {
        // Train and evaluate classifier
        int test_size = instances / parm->cv_folds;

        // Evaluate classifier
        TN = TP = FP = FN = 0.0;

        for (int fold = 0; fold < parm->cv_folds; fold++) {
            // Create train and test sets
            // Create train and test sets
            int index = test_size * fold;
            for (int j = 0; j < test_size; j++, index = (index + 1) % instances) {
                test[j] = data[index];
                test_label[j] = labels[index];
            }
            for (int j = 0; j < instances - test_size; j++, index = (index + 1) % instances) {
                train[j] = data[index];
                train_label[j] = labels[index];
            }

            // Positive sequences
            for (int j = 0; j < test_size; j++)
                if (test_label[j] == POSITIVE) {
                    // Positive instances
                    if (kNearestNeighbour(test[j], train, instances - test_size, inputs, k, train_label, parm->knn_weight_by_distance) > 0.0)
                        TP++;
                    else
                        FN++;
                } else {
                    // Negative instances
                    if (kNearestNeighbour(test[j], train, instances - test_size, inputs, k, train_label, parm->knn_weight_by_distance) < 0.0)
                        TN++;
                    else
                        FP++;
                }
        }

        // Better than current best?
        scalar sensitivity = TP / (TP + FN);
        scalar specificity = TN / (TN + FP);
        scalar G = pow(sensitivity * specificity, 0.5);
        if (G > best_so_far) {
            best_so_far = G;
            best_k = k;
        }

        if (parm->verbose > 1)
            fprintf(parm->out_file, "parm->k = %d\tG: %lf\n", k, G);
    }

    // Accuracy in percentage
    fprintf(parm->out_file, "Stacking k-NN best parameters (accuracy: %lf): parm->k = %d\n", best_so_far, best_k);

    FREE(train);
    FREE(test);
    FREE(train_label);
    FREE(test_label);

    return best_k;
}

scalar kNearestNeighbour(scalar *instance, scalar **protos, int n_protos, int dim, int k, int *label, bool knn_weight_by_distance)
{
    scalar value, dmin, dist[k], votes[] = {0, 0};
    int nearest[k];

    // Select k first as k initial neighbors
    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < n_protos && dmin > 0.0; i++) {
        value = EuclideanDistance(instance, protos[i], dim);

        // Test if is one of the nearest
        if (value < dmin) {
            // Look for its place
            int j = 0;
            while (j < k && value > dist[j])
                j++;

            if (j == k)
                break;

            // Sets in j-th place
            for (int l = k - 1; l > j; l--) {
                nearest[l] = nearest[l - 1];
                dist[l] = dist[l - 1];
            }
            nearest[j] = i;
            dist[j] = value;
            dmin = dist[k - 1];
        }

    }

    if (knn_weight_by_distance) {
        scalar sum = 0.0;
        for (int i = 0; i < k; i++)
            sum += exp(-dist[i]);
        if (sum < ZERO)
            sum = ZERO;

        for (int i = 0; i < k; i++)
            if (label[nearest[i]] == POSITIVE)
                votes[POSITIVE] += exp(-dist[i]) / sum;
            else
                votes[NEGATIVE] += exp(-dist[i]) / sum;

    } else {
        for (int i = 0; i < k; i++)
            if (label[nearest[i]] == POSITIVE)
                votes[POSITIVE] ++;
            else
                votes[NEGATIVE] ++;
    }

    // Get maximum votes
    return votes[POSITIVE] - votes[NEGATIVE];
}
