/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "type.h"
#include "macros.h"

void SetkNNThreshold(const t_Parameters *parm, t_Recognizer *R, FILE *out_file)
{
    // Get threshold to obtain desired sensitivity = TP/(TP+FN)
    // Get all differences and ...
    scalar *sc;
    ALLOCA(sc, scalar, R->n_poss, "SetkNNThreshold: sc");
    for (int i = 0; i < R->n_poss; i++)
        sc[i] = ClassifySequencekNN(R, R->poss[i], R->pos_l, parm->knn_weight_by_distance);

    // ...sort them
    qsort(sc, R->n_poss, sizeof(scalar), CompareDouble);

    // Get threshold for desired sensitivity
    R->threshold = sc[(int)(R->n_poss * (1.0 - parm->sensitivity)) ];

    fprintf(out_file, "k-NN threshold: %lf\n", R->threshold);
}

// Returns relative Hamming distance if it is below dmin, otherwise returns 1
scalar HammingDistance(char *seq1, char *seq2, int l/*, int dmin*/)
{
    scalar distance = 0.0;

    for (int i = 0; i < l /*&& distance < dmin*/; i++)
        if (seq1[i] != seq2[i])
            distance++;

    return distance/l;
}

scalar ClassifySequencekNN(t_Recognizer *rec, char *seq, int s_length, bool knn_weight_by_distance)
{
    scalar distance[rec->k], dmin, dist;
    t_SequenceClass nearest[rec->k];
    scalar votes[2] = {0.0, 0.0};

    // Select k first as k initial neighbors
    dmin = 2.0;
    for (int i = 0; i < rec->k; i++)
        distance[i] = 2.0;

    // Get k-nearest neighbors using Hamming distance
    for (int i = 0; i < rec->n_poss; i++) {
        dist = HammingDistance(seq, rec->poss[i], s_length);

        // Test if is one of the nearest
        if (dist < dmin) {
            // Look for its place
            int j = 0;
            while (dist > distance[j])
                j++;

            // Sets in j-th place
            for (int l = rec->k - 1; l > j; l--) {
                nearest[l] = nearest[l - 1];
                distance[l] = distance[l - 1];
            }
            nearest[j] = POSITIVE;
            distance[j] = dist;
            dmin = distance[rec->k - 1];
        }
    }

    for (int i = 0; i < rec->n_negs; i++) {
        dist = HammingDistance(seq, rec->negs[i], s_length);

        // Test if is one of the nearest
        if (dist < dmin) {
            // Look for its place
            int j = 0;
            while (dist > distance[j])
                j++;

            // Sets in j-th place
            for (int l = rec->k - 1; l > j; l--) {
                nearest[l] = nearest[l - 1];
                distance[l] = distance[l - 1];
            }
            nearest[j] = NEGATIVE;
            distance[j] = dist;
            dmin = distance[rec->k - 1];
        }
    }

    if (knn_weight_by_distance) {
        scalar sum = 0.0;
        for (int i = 0; i < rec->k; i++)
            sum += exp(-distance[i]);
        if (sum < ZERO)
            sum = ZERO;

        for (int i = 0; i < rec->k; i++)
            if (nearest[i] == POSITIVE)
                votes[0] += exp(-distance[i]) / sum;
            else
                votes[1] += exp(-distance[i]) / sum;

    } else {
        for (int i = 0; i < rec->k; i++)
            if (nearest[i] == POSITIVE)
                votes[0] ++;
            else
                votes[1] ++;
    }

    return (votes[0] - votes[1]);
}
