/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "type.h"
#include "type.h"
#include "macros.h"
#include "defns.i"
#include "types.i"
#include "extern.i"

void SaveModel(const t_Parameters *parm, t_Recognizer *R, char *model_file)
{
    char model_filename[1000];
    FILE *m_file;

    switch (parm->r_method) {
        case SVM:
            sprintf(model_filename, "%s.svm", model_file);

            // Save SVM
            svm_save_model(model_filename, R->lsvm);

            // Save model data
            if ((m_file = fopen(model_file, "w")) == NULL)
                OnErrorExit(1, "Unable to open model file for writing");

            // Write info
            fprintf(m_file, "Classifier: %s\n", "SVM");
            fprintf(m_file, "Type: %d\n", parm->r_method);
            fprintf(m_file, "Sequence length: %d\n", R->length);
            fprintf(m_file, "Offset from site: %d\n", parm->offset - parm->offset_to_site);
            fprintf(m_file, "Threshold: %lf\n", R->threshold);
            fprintf(m_file, "SVM model file name: %s\n", model_filename);

            fclose(m_file);
            break;

        case kNN:
            // Save model data
            if ((m_file = fopen(model_file, "w")) == NULL)
                OnErrorExit(1, "Unable to open model file for writing");

            fprintf(m_file, "Classifier: %s\n", "kNN");
            fprintf(m_file, "Threshold: %lf\n", R->threshold);
            fprintf(m_file, "k: %d\n", R->k);
            fprintf(m_file, "Positives: %d\n", R->n_poss);
            fprintf(m_file, "Negatives: %d\n", R->n_negs);
            fprintf(m_file, "Length: %d\n", R->pos_l);  // Same as R->neg_l
            fprintf(m_file, "Offset from site: %d\n", parm->offset - parm->offset_to_site);

            // Positive sequences
            for (int i = 0; i < R->n_poss; i++)
                fprintf(m_file, "%s\n", R->poss[i]);

            // Negative sequences
            for (int i = 0; i < R->n_negs; i++)
                fprintf(m_file, "%s\n", R->negs[i]);

            fclose(m_file);
            break;

        case STOP:
            // Save model data
            if ((m_file = fopen(model_file, "w")) == NULL)
                OnErrorExit(1, "Unable to open model file for writing");

            fprintf(m_file, "Classifier: %s\n", "STOP");
            fprintf(m_file, "Reverse: %c\n", parm->reverse_stop_codon ? 'T' : 'F');
            fprintf(m_file, "Threshold: %lf\n", R->threshold);
            fprintf(m_file, "Positions: %d\n", R->n_positions);
            fprintf(m_file, "Sequence length: %d\n", R->length);
            fprintf(m_file, "Offset from site: %d\n", parm->offset - parm->offset_to_site);

            // Scores
            for (int i = 0; i < R->n_positions; i++)
                fprintf(m_file, "%lf\n", R->scores[i]);

            fclose(m_file);
            break;

        case PWM:
            // Save model data
            if ((m_file = fopen(model_file, "w")) == NULL)
                OnErrorExit(1, "Unable to open model file for writing");

            fprintf(m_file, "Classifier: %s\n", "PWM");
            fprintf(m_file, "Threshold: %lf\n", R->threshold);
            fprintf(m_file, "Sequence length: %d\n", R->length);
            fprintf(m_file, "Offset from site: %d\n", parm->offset - parm->offset_to_site);

            // Scores
            for (int j = 0; j < 18; j++)
                for (int i = 0; i < R->length; i++)
                    fprintf(m_file, "%lf\n", R->M[j][i]);

            fclose(m_file);
            break;

        case C45:
            // Save tree
            SaveTree(R->tree, ".c45", model_file, parm->out_file);

            // Save model data
            if ((m_file = fopen(model_file, "w")) == NULL)
                OnErrorExit(1, "Unable to open model file for writing");

            // Write info
            fprintf(m_file, "Classifier: %s\n", "C4.5");
            fprintf(m_file, "Sequence length: %d\n", R->length);
            fprintf(m_file, "Offset from site: %d\n", parm->offset - parm->offset_to_site);
            fprintf(m_file, "Threshold: %lf\n", R->threshold);
            fprintf(m_file, "C4.5 model file name: %s\n", model_file);

            fclose(m_file);
            break;

        default:
            OnErrorExit(1, "Bad model for saving");
    }
}

void LoadModel(t_Parameters *parm, t_Recognizer *R, char *model_file)
{
    char model_filename[1000], name[100], c;
    FILE *m_file;
    int dummy;

    // Load model type and check it matches
    // Load model data
    if ((m_file = fopen(model_file, "r")) == NULL) {
        char message[1000];
        sprintf(message, "Unable to open model file \"%s\"for reading", model_file);
        OnErrorExit(1, message);
    }
    if (!fscanf(m_file, "Classifier: %s\n", name))
        OnErrorExit(1, "Error reading model file");

    if (parm->r_method == SVM && strncmp(name, "SVM", 3))
        OnErrorExit(1, "Model is not SVM");

    if (parm->r_method == kNN && strncmp(name, "kNN", 3))
        OnErrorExit(1, "Model is not kNN");

    if (parm->r_method == STOP && strncmp(name, "STOP", 4))
        OnErrorExit(1, "Model is not STOP");

    if (parm->r_method == PWM && strncmp(name, "PWM", 3))
        OnErrorExit(1, "Model is not PWM");

    if (parm->r_method == C45 && strncmp(name, "C4.5", 4))
        OnErrorExit(1, "Model is not C4.5");

    switch (parm->r_method) {
        case SVM:
            sprintf(model_filename, "%s.svm", model_file);

            // Read info
            if (!fscanf(m_file, "Type: %d\n", (int *) &dummy))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Sequence length: %d\n", &(R->length)))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Offset from site: %d\n", &parm->offset_to_site))
                OnErrorExit(1, "Error reading model file");
            parm->offset_to_site = parm->offset - parm->offset_to_site;
            if (!fscanf(m_file, "Threshold: "SCALAR_FORMAT"\n", &(R->threshold)))
                OnErrorExit(1, "Error reading model file");

            // Load SVM model
            if ((R->lsvm = svm_load_model(model_filename)) == NULL)
                OnErrorExit(1, "Error loading SVM model");
            break;

        case kNN:
            // Read info
            if (!fscanf(m_file, "Threshold: "SCALAR_FORMAT"\n", &R->threshold))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "k: %d\n", &R->k))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Positives: %d\n", &R->n_poss))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Negatives: %d\n", &R->n_negs))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Length: %d\n", &R->pos_l))
                OnErrorExit(1, "Error reading model file");
            R->neg_l = R->pos_l;
            if (!fscanf(m_file, "Offset from site: %d\n", &parm->offset_to_site))
                OnErrorExit(1, "Error reading model file");
            parm->offset_to_site = parm->offset - parm->offset_to_site;

            // Positive sequences
            MALLOC(R->poss, char *, R->n_poss, "LoadModel: R->poss");
            for (int i = 0; i < R->n_poss; i++) {
                MALLOC(R->poss[i], char, R->pos_l+1, "LoadModel: R->poss[i]");
                if (!fscanf(m_file, "%s\n", R->poss[i]))
                    OnErrorExit(1, "Error reading model file");
            }

            // Negative sequences
            MALLOC(R->negs, char *, R->n_negs, "LoadModel: R->negs");
            for (int i = 0; i < R->n_negs; i++) {
                MALLOC(R->negs[i], char, R->neg_l+1, "LoadModel: R->negs[i]");
                if (!fscanf(m_file, "%s\n", R->negs[i]))
                    OnErrorExit(1, "Error reading model file");
            }
            break;

        case STOP:
            // Read info
            if (!fscanf(m_file, "Reverse: %c\n", &c))
                OnErrorExit(1, "Error reading model file");
            parm->reverse_stop_codon = (c == 'T') ? true : false;
            if (!fscanf(m_file, "Threshold: "SCALAR_FORMAT"\n", &R->threshold))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Positions: %d\n", &R->n_positions))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Sequence length: %d\n", &(R->length)))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Offset from site: %d\n", &parm->offset_to_site))
                OnErrorExit(1, "Error reading model file");
            parm->offset_to_site = parm->offset - parm->offset_to_site;

            // Scores
            MALLOC(R->scores, scalar, R->n_positions, "LoadModel: R->scores");
            for (int i = 0; i < R->n_positions; i++)
                if (!fscanf(m_file, SCALAR_FORMAT"\n", &R->scores[i]))
                    OnErrorExit(1, "Error reading model file");
            break;

        case PWM:
            // Read model data
            if (!fscanf(m_file, "Threshold: "SCALAR_FORMAT"\n", &R->threshold))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Sequence length: %d\n", &(R->length)))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Offset from site: %d\n", &parm->offset_to_site))
                OnErrorExit(1, "Error reading model file");
            parm->offset_to_site = parm->offset - parm->offset_to_site;

            // Allocate memory
            for (int i = 0; i < 18; i++)
                MALLOC(R->M[i], scalar, R->length, "LoadModel: R->M[i]");

            // Scores
            for (int j = 0; j < 18; j++)
                for (int i = 0; i < R->length; i++)
                    if (!fscanf(m_file, SCALAR_FORMAT"\n", &R->M[j][i]))
                        OnErrorExit(1, "Error reading model file");
            break;

        case C45:
            // Read info
            if (!fscanf(m_file, "Sequence length: %d\n", &(R->length)))
                OnErrorExit(1, "Error reading model file");
            if (!fscanf(m_file, "Offset from site: %d\n", &parm->offset_to_site))
                OnErrorExit(1, "Error reading model file");
            parm->offset_to_site = parm->offset - parm->offset_to_site;
            if (!fscanf(m_file, "Threshold: "SCALAR_FORMAT"\n", &(R->threshold)))
                OnErrorExit(1, "Error reading model file");

            // Load C4.5 model
            // Create dummy names file
            extern String FileName;
            char *name_template = tempnam(".", NULL);
            sprintf(model_filename, "%s.names", name_template);
            FILE *fp = fopen(model_filename, "w");
            fprintf(fp, "0, 1.    | classes\n");
            for (int i = 0; i < R->length; i++)
                fprintf(fp, "x%d: A,T,U,G,C,R,Y,K,M,S,W,B,D,H,V,N,X,-\n", i);
            fclose(fp);

            //  Initialize  values for C4.5
            ALLOCA(FileName, char, 100, "TrainC45Classifier: FileName");
            sprintf(FileName, "%s", name_template);
            GetNames();

            if ((R->tree = GetTree(".c45", model_file)) == NULL)
                OnErrorExit(1, "Error loading C4.5 model");
            remove(model_filename);
            if (parm->verbose > 1)
                printf("Read %d tree nodes\n", TreeSize(R->tree));
            break;

        default:
            OnErrorExit(1, "Cannot read model");
    }

    fclose(m_file);
}
