/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <pvm3.h>
#include <stdbool.h>
#include <wait.h>
#include <stddef.h>
#include <bits/types.h>
#include <features.h>
#include "type.h"
#include "macros.h"
#include "common.h"

#define VERBOSE_INTERVAL        1000000
#define VERBOSE_OUTPUT(str, integer)   printf("%s: %d\n", str, integer)

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal;
int MaxAtt;
ItemNo MaxItem;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY = 0, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float CF = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

// Outputs of a SVM model already trained with more than 2 classes and using 1-against-1 method

// POSITIVE CLASS MUST ALWAYS BE CLASS 1

/* Need a description file of the form:
Positive sequences file
Negative sequences file
[Test positive sequences file]    // Only for constructive/destructive models
[Test negative sequences file]
Models output file
[Test models output file]    // Only for constructive/destructive models
Number of models
Description of class 1
Parameters file for 1 against all model
1 against all model file
Description of class 2
Parameters file for 1 against 2
Threshold
1 against 2 model file
...
Description of class k
Parameters file for 1 against k
1 against k model file
Threshold
// Negatives info in data file. Types = {E, I, IG}
Offset to type 2 (introns)
Offset to type 3 (intergenic space)
// List of combinations for saving ROC info. Might be omitted.
[AUC]
1, 2, 3, 4, ...
// List of combinations for saving FP files. Might be omitted.
[FP]
1, 2, 3, 4, ...
// List of models to use. It none use all
[USE]
1, 2, 3, 4, ...
*/

#define MAX_LENGTH      10000

typedef struct {
    char description[MAX_LENGTH];
    char parameters_file[MAX_LENGTH];
    char model_file[MAX_LENGTH];
    scalar threshold;
} t_class_description;

// Combination and stepwise methods
typedef enum {
    SumOfOutputs = 0, Majority = 1, Maximum = 2
} t_CombinationMethod;

typedef enum {
    SFS = 0, SBS = 1, LRS = 2, SFFS = 3, SFBS = 4, RSFFS = 5, RSFBS = 6
} t_StepwiseMethod;

int n_combination_methods = 3, n_stepwise_methods = 7;
char *combination_method_label[] = {"Sum", "Majority", "Maximum"};
char *stepwise_method_label[] = {"SFS", "SBS", "LRS", "SFFS", "SFBS", "RSFFS", "RSFBS"};

/*
METHODS:

    Sequential forward selection (SFS)
    Sequential backward selection (SBS)
    Plus-L Minus-R selection (LRS)
    Sequential Floating Forward Selection (SFFS)
    Sequential Floating Backward Selection (SFBS)
    Random Sequential Floating Forward Selection (RSFFS)
    Random Sequential Floating Backward Selection (RSFBS)
*/

int main(int argc, char *argv[])
{
    extern char letters[];

    int n_positives, n_negatives, length, *tids, n_models, first_model = 0, offset[2], ex, in, ig, TP, TN, FP, FN;
    char **positives = NULL, **negatives = NULL, file_name[MAX_LENGTH], pos_file_name[MAX_LENGTH], neg_file_name[MAX_LENGTH], line[MAX_LENGTH], models_output_file[MAX_LENGTH],
           test_models_output_file[MAX_LENGTH], test_pos_file_name[MAX_LENGTH], test_neg_file_name[MAX_LENGTH], uncompressed_sequence[MAX_SEQUENCE_LENGTH];
    t_Recognizer R;
    FILE *description;
    scalar **outputs = NULL, sp, sn, *one_model_outputs = NULL;
    t_class_description *desc;
    bool verbose = false;
    t_Parameters parm;
    struct spectrum_kernel_info spectrum;

    // Test enough parameters
    if (argc >= 3) {
        // File for output messages.
        if ((parm.out_file = fopen(argv[2], "w")) == NULL)
            OnErrorExit(1, "main: Cannot open output file\n");
    } else {
        fprintf(stdout, "models-output description-file output [number-of-slaves]\n");
        fprintf(stdout, "Options: -x (exit after saving the outputs) -s (stepwise) [AUROC, AUPRC, G] -v (verbose) -b (best combination of n models) [AUROC, AUPRC, G] -c (compress)\n");
        exit(1);
    }

#ifdef SCALAR_IS_FLOAT
    printf("WARNING: With scalar as float AUC results might not be reliable\n");
#endif

    // Options
    bool exit_after_outputs = false, stepwise_methods = false, use_auroc = false, use_auprc = false, best_combination = false, compress_sequences = false;
    int opt, shift = 0;
    while ((opt = getopt(argc, argv, "xs:vb:c")) != EOF)
        switch (opt) {
            case 'x':
                exit_after_outputs = true;
                shift++;
                break;

            case 'v':
                verbose = true;
                shift++;
                break;

            case 'c':
                printf("Compress sequences for space saving\n");
                compress_sequences = true;
                shift++;
                break;

            case 's':
                stepwise_methods = true;
                printf("Stepwise methods ");
                shift++;

                // What to save
                if (optarg == NULL)
                    OnErrorExit(1, "I need a performance measure: AUROC, AUPRC or G(-mean)");

                if (!strncasecmp("AUROC", optarg, 5)) {
                    use_auroc = true;
                    printf("using AUROC as performance measure\n");
                } else if (!strncasecmp("AUPRC", optarg, 5)) {
                    use_auprc = true;
                    printf("using AUPRC as performance measure\n");
                } else if (!strncasecmp("G", optarg, 1)) {
                    printf("using G-mean as performance measure\n");
                } else
                    OnErrorExit(1, "Bad argument for constructive/destructive methods option");
                shift++;
                break;

            case 'b':
                printf("Evaluate best combination of all selected models ");
                best_combination = true;
                shift++;

                // What to save
                if (optarg == NULL)
                    OnErrorExit(1, "I need a performance measure: AUROC, AUPRC or G-mean");

                if (!strncasecmp("AUROC", optarg, 5)) {
                    use_auroc = true;
                    printf("using AUROC as performance measure\n");
                } else if (!strncasecmp("AUPRC", optarg, 5)) {
                    use_auprc = true;
                    printf("using AUPRC as performance measure\n");
                } else if (!strncasecmp("G", optarg, 1)) {
                    printf("using G-mean as performance measure\n");
                } else
                    OnErrorExit(1, "Bad argument for constructive/destructive methods option");
                shift++;
                break;

            default:
                OnErrorExit(1, "Bad option");
        }

    argv += shift;

    printf("Ignoring MAXIMUM combination due to worse performance\n");

    // Common values for all datasets
    // Threshold for classification
    char *home = getenv("HOME");
    // Number of slaves in case of a parallel version
    int n_slaves = 0;
    if (argc >= 4)
        n_slaves = atoi(argv[3]);

    // Read description file
    if ((description = fopen(argv[1], "rt")) == NULL)
        OnErrorExit(ERR_FILE, "main: Cannot open description file");

    if (fgets(pos_file_name, MAX_LENGTH, description) == NULL)
        OnErrorExit(ERR_FILE, "Error reading description file");
    REMOVE_NEWLINE(pos_file_name);
    if (fgets(neg_file_name, MAX_LENGTH, description) == NULL)
        OnErrorExit(ERR_FILE, "Error reading description file");
    REMOVE_NEWLINE(neg_file_name);

    // Test outputs file?
    if (stepwise_methods || best_combination) {
        if (fgets(test_pos_file_name, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(test_pos_file_name);
        if (fgets(test_neg_file_name, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(test_neg_file_name);
    }

    // File for the outputs of the model
    if (fgets(models_output_file, MAX_LENGTH, description) == NULL)
        OnErrorExit(ERR_FILE, "Error reading description file");
    REMOVE_NEWLINE(models_output_file);

    // Test outputs file?
    if (stepwise_methods || best_combination) {
        // File for the outputs of the model
        if (fgets(test_models_output_file, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(test_models_output_file);
    }

    // Read number of classes and allocate needed info
    if (fgets(line, MAX_LENGTH, description) == NULL)
        OnErrorExit(ERR_FILE, "Error reading description file");
    n_models = atoi(line);
    MALLOC(desc, t_class_description, n_models, "main: desc");
    for (int i = 0; i < n_models; i++) {
        if (fgets(desc[i].description, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(desc[i].description);
        if (fgets(desc[i].parameters_file, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(desc[i].parameters_file);
        if (fgets(desc[i].model_file, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        REMOVE_NEWLINE(desc[i].model_file);
        if (fgets(line, MAX_LENGTH, description) == NULL)
            OnErrorExit(ERR_FILE, "Error reading description file");
        desc[i].threshold = atof(line);
    }

    // Read order of negative instances and offsets
    if (!fscanf(description, "%d", &offset[0]))
        OnErrorExit(ERR_FILE, "Error reading description file");
    if (!fscanf(description, "%d\n", &offset[1]))
        OnErrorExit(ERR_FILE, "Error reading description file");

    // Combinations for ROC info
    int n_roc = 0, *roc_models = NULL;

    if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
        OnErrorExit(ERR_FILE, "Error reading description file");

    // Read [AUC] token?
    if (!strncmp(line, "[AUC]", 5)) {
        // Allocate by hundreds
        MALLOC(roc_models, int, 100, "main: roc_models");

        if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
            OnErrorExit(ERR_FILE, "Error reading description file");

        while (!feof(description) && strncmp(line, "[FP]", 4) && strncmp(line, "[USE]", 5)) {
            int next;
            char *token;
            roc_models[n_roc] = 0;

            token = strtok(line, ", ;\t");
            while (token) {
                next = atoi(token);
                roc_models[n_roc] += (int) pow(2.0, next-1);
                token = strtok(NULL, ", ;\t");
            }

            n_roc++;

            // Another hundred?
            if (!(n_roc % 100))
                REALLOC(roc_models, int, n_roc + 100, "main: roc_models");

            // Next?
            if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
                OnErrorExit(ERR_FILE, "Error reading description file");
        }
    }

    // Combinations for FP file
    int n_fp = 0, *fp_models = NULL;

    // Read [FP] token?
    if (!feof(description) && !strncmp(line, "[FP]", 4)) {
        MALLOC(fp_models, int, (int) pow(2.0, n_models), "main: fp_models");

        if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
            OnErrorExit(ERR_FILE, "Error reading description file");

        while (!feof(description) && strncmp(line, "[USE]", 5)) {
            int next;
            char *token;
            fp_models[n_fp] = 0;

            token = strtok(line, ", ;\t");
            while (token) {
                next = atoi(token);
                fp_models[n_fp] += (int) pow(2.0, next-1);
                token = strtok(NULL, ", ;\t");
            }

            n_fp++;

            // Next?
            if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
                OnErrorExit(ERR_FILE, "Error reading description file");
        }
    }

    // Read [USE] token?
    bool *use_model;
    MALLOC(use_model, bool, n_models, "main: use_model");

    if (!feof(description) && !strncmp(line, "[USE]", 4)) {
        if (fgets(line, MAX_LENGTH, description) == NULL && !feof(description))
            OnErrorExit(ERR_FILE, "Error reading description file");

        for (int i = 0; i < n_models; i++)
            use_model[i] = false;

        int next;
        char *token;

        token = strtok(line, ", ;\t");
        while (token) {
            next = atoi(token);
            use_model[next-1] = true;
            token = strtok(NULL, ", ;\t");
        }

        printf("Model(s) to use: ");
        for (int i = 0; i < n_models; i++)
            if (use_model[i])
                printf("%d ", i+1);
        printf("\n");
    } else // Use all
        for (int i = 0; i < n_models; i++)
            use_model[i] = true;

    fclose(description);

    // Show info
    fprintf(parm.out_file, "Positive sequences testing file: %s\n", pos_file_name);
    fprintf(parm.out_file, "Negative sequences testing file: %s\n", neg_file_name);
    for (int i = 0; i < n_models; i++)
        fprintf(parm.out_file, "Model %d: %s. Model file: %s. Parameters file: %s. Threshold: %lf\n", i+1, desc[i].description, desc[i].model_file, desc[i].parameters_file, desc[i].threshold);

    // For measuring expended time
    time_t init_time = time(NULL);

    // If models output exist read previous models
    int n_lines = -1;
    if (!access(models_output_file, R_OK)) {
        // Open and count lines
        FILE *modelf;
        if ((modelf = fopen(models_output_file, "rt")) == NULL)
            OnErrorExit(1, "Cannot open models output file for reading");
        char line[MAX_LENGTH], *next_output;
        do {
            if (fgets(line, MAX_LENGTH, modelf) == NULL && !feof(modelf))
                OnErrorExit(ERR_FILE, "Error reading model outputs file");
            n_lines++;
        } while (!feof(modelf));

        if (!exit_after_outputs) {
            MALLOC(outputs, scalar *, n_lines, "main: outputs");
            for (int i = 0; i < n_lines; i++)
                MALLOC(outputs[i], scalar, n_models, "main: outputs[i]");
        }

        // Read first line and count previous models
        rewind(modelf);
        if (fgets(line, MAX_LENGTH, modelf) == NULL)
            OnErrorExit(ERR_FILE, "Error reading model outputs file");
        int n_generated_models = 0;
        next_output = strtok(line, " \t");
        while (next_output != NULL && strncmp(next_output, "\n", 1)) {
            n_generated_models++;
            next_output = strtok(NULL, " \t");
        }

        printf("%d model(s) already generated for %d instances\n", n_generated_models, n_lines);

        // Excess of models? Then, reallocate memory
        if (!exit_after_outputs && n_generated_models > n_models)
            for (int i = 0; i < n_lines; i++)
                REALLOC(outputs[i], scalar, n_generated_models, "main: outputs[i]");

        printf("%d model(s) need to be generated for %d instances\n", MAX(0, n_models - n_generated_models), n_lines);

        // Read outputs
        if (!exit_after_outputs) {
            rewind(modelf);
            for (int i = 0; i < n_lines; i++)
                for (int j = 0; j < n_generated_models; j++)
                    if (!fscanf(modelf, SCALAR_FORMAT, &outputs[i][j]))
                        OnErrorExit(ERR_FILE, "Error reading model outputs file");
        }

        fclose(modelf);

        // Ignore models
        if (n_generated_models > n_models)
            n_generated_models = n_models;

        // First model
        first_model = n_generated_models;

        // If all models done get number of positive and negative instances
        if (first_model == n_models) {
            LoadGlobalParameters(&parm, desc[0].parameters_file);

            // Count instances
            sprintf(file_name, "%s", pos_file_name);
            if (access(file_name, R_OK))
                sprintf(file_name, "%s/%s/%s", home, parm.datad, pos_file_name);
            // No file, no positive sequences
            if (access(file_name, R_OK))
                n_positives = 0;
            else
                ReadFastaFile(file_name, NULL, &n_positives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                              parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

            fprintf(parm.out_file, "Positive sequences: %d\n", n_positives);

            sprintf(file_name, "%s", neg_file_name);
            if (access(file_name, R_OK))
                sprintf(file_name, "%s/%s/%s", home, parm.datad, neg_file_name);
            // No file, no negative sequences
            if (access(file_name, R_OK))
                n_negatives = 0;
            else
                ReadFastaFile(file_name, NULL, &n_negatives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                              parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

            fprintf(parm.out_file, "Negative sequences: %d\n", n_negatives);

            if (n_lines != n_positives + n_negatives) {
                printf("Generated outputs: %d\tRead from file: %d\n", n_lines, n_positives+n_negatives);
                OnErrorExit(1, "Generated outputs and read data files do not match");
            }
        }
    }

    // List of sequences regardless their type
    char **sequences = NULL;
    int n_sequences = 0;

    // REPEAT FOR EACH PAIRED MODEL
    int previous_offset = -1, previous_length = -1;
    for (int model = first_model; model < n_models; model++) {
        VERBOSE_OUTPUT("MODEL", model+1);

        LoadGlobalParameters(&parm, desc[model].parameters_file);

        if (previous_offset != parm.offset || previous_length != parm.length) {
            // Free data?
            if (previous_length > 0) {
                // Free training data
                for (int i = 0; i < n_positives; i++)
                    FREE(positives[i]);
                if (n_positives)
                    FREE(positives);
                for (int i = 0; i < n_negatives; i++)
                    FREE(negatives[i]);
                if (n_negatives)
                    FREE(negatives);
                if (sequences)
                    FREE(sequences);
            }

            previous_offset = parm.offset;
            previous_length = parm.length;

            // Read training data first time or when the sequence window changes
            sprintf(file_name, "%s", pos_file_name);
            if (access(file_name, R_OK))
                sprintf(file_name, "%s/%s/%s", home, parm.datad, pos_file_name);
            // No file, no positive sequences
            if (access(file_name, R_OK))
                n_positives = 0;
            else
                ReadFastaFile(file_name, &positives, &n_positives, &length, 1.0, false, 0, false, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                              parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);
            fprintf(parm.out_file, "Positive sequences: %d\n", n_positives);

            sprintf(file_name, "%s", neg_file_name);
            if (access(file_name, R_OK))
                sprintf(file_name, "%s/%s/%s", home, parm.datad, neg_file_name);
            // No file, no negative sequences
            if (access(file_name, R_OK))
                n_negatives = 0;
            else
                ReadFastaFile(file_name, &negatives, &n_negatives, &length, 1.0, false, 0, false, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                              parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);
            fprintf(parm.out_file, "Negative sequences: %d\n", n_negatives);

            // If first time reserve memory for outputs
            if (!exit_after_outputs && outputs == NULL) {
                MALLOC(outputs, scalar *, n_positives+n_negatives, "main: outputs");
                for (int i = 0; i < n_positives+n_negatives; i++)
                    CALLOC(outputs[i], scalar, n_models, "main: outputs[i]");
            }

            // Check with previously generated models if such is the case
            if (n_lines > 0 && n_lines != n_positives + n_negatives) {
                printf("Generated outputs: %d\tRead from file: %d\n", n_lines, n_positives+n_negatives);
                OnErrorExit(1, "Generated outputs and read data files do not match");
            }
        }

        printf("Read data files\n");

        // Outputs for one model only, first iteration
        if (!one_model_outputs)
            MALLOC(one_model_outputs, scalar, n_positives+n_negatives, "main: one_model_outputs");

        // Sequential or parallel version?
        if (n_slaves <= 0) {
            // Load model
            LoadModel(&parm, &R, desc[model].model_file);

            printf("Read model file %s\n", desc[model].model_file);

            if (!compress_sequences) {
                for (int i = 0; i < n_positives; i++) {
                    spectrum.n_k_mers = -1;
                    spectrum.list = spectrum.count = NULL;
                    spectrum.k_xx = -1.0;
                    one_model_outputs[i] = ClassifierOutput((const t_Parameters *) &parm, &R, positives[i], length, 0, NULL, &spectrum);
                    if (!(i%VERBOSE_INTERVAL))
                        VERBOSE_OUTPUT("Positive instance", i);
                }

                for (int i = 0; i < n_negatives; i++) {
                    spectrum.n_k_mers = -1;
                    spectrum.list = spectrum.count = NULL;
                    spectrum.k_xx = -1.0;
                    one_model_outputs[i+n_positives] = ClassifierOutput((const t_Parameters *) &parm, &R, negatives[i], length, 0, NULL, &spectrum);
                    if (!(i%VERBOSE_INTERVAL))
                        VERBOSE_OUTPUT("Negative instance", i);
                }
            } else {
                for (int i = 0; i < n_positives; i++) {
                    UncompressSequence(positives[i], uncompressed_sequence, length);
                    spectrum.n_k_mers = -1;
                    spectrum.list = spectrum.count = NULL;
                    spectrum.k_xx = -1.0;
                    one_model_outputs[i] = ClassifierOutput((const t_Parameters *) &parm, &R, uncompressed_sequence, length, 0, NULL, &spectrum);
                    if (!(i%VERBOSE_INTERVAL))
                        VERBOSE_OUTPUT("Positive instance", i);
                }

                for (int i = 0; i < n_negatives; i++) {
                    UncompressSequence(negatives[i], uncompressed_sequence, length);
                    spectrum.n_k_mers = -1;
                    spectrum.list = spectrum.count = NULL;
                    spectrum.k_xx = -1.0;
                    one_model_outputs[i+n_positives] = ClassifierOutput((const t_Parameters *) &parm, &R, uncompressed_sequence, length, 0, NULL, &spectrum);
                    if (!(i%VERBOSE_INTERVAL))
                        VERBOSE_OUTPUT("Negative instance", i);
                }
            }

            if (!exit_after_outputs)
                for (int i = 0; i < n_positives+n_negatives; i++)
                    outputs[i][model] = one_model_outputs[i];
        } else {
            if (sequences == NULL) {
                MALLOC(sequences, char *, n_positives+n_negatives, "main: sequences");
                for (int i = 0; i < n_positives; i++)
                    sequences[i] = positives[i];
                for (int i = n_positives; i < n_positives+n_negatives; i++)
                    sequences[i] = negatives[i - n_positives];
                n_sequences = n_positives + n_negatives;
            }

            char *args[4];
            MALLOC(args[0], char, strlen(desc[model].parameters_file) + 1, "main");
            MALLOC(args[1], char, strlen(desc[model].model_file) + 1, "main");
            strcpy(args[0], desc[model].parameters_file);      // Parameters file name
            strcpy(args[1], desc[model].model_file);    // Model file name
            args[2] = getcwd(NULL, 0);                  // Current directory for slaves
            args[3] = NULL;

            MALLOC(tids, int, n_slaves, "main: tids");
            // Avoid running processes in master node
            int numt = pvm_spawn("model-output-slave", args, PvmTaskHost | PvmHostCompl, "skynet.uco.es", n_slaves, tids);
            // int numt = pvm_spawn("model-output-slave", args, PvmTaskHost | PvmTaskDebug | PvmHostCompl, "", n_slaves, tids);

            if (numt != n_slaves)
                OnErrorExit(1, "Processes not spawned correctly");

            // Start guardian and send info
            int guardian = InitGuardian(numt, tids, "model-output-slave", "/dev/null");

            int seqs_sent, sent = 0, received = 0, first_instance;

            // Send to slaves a first round for all of them
            for (int j = 0; j < n_slaves && sent < n_sequences; j++) {
                // Sequences to sent
                seqs_sent = MIN(SEQUENCES_BLOCK, n_sequences-sent);
                pvm_initsend(PvmDataDefault);
                pvm_pkint(&seqs_sent, 1, 1);
                pvm_pkint(&sent, 1, 1); // Index of first pattern
                if (compress_sequences)
                    for (int k = 0; k < seqs_sent; k++, sent++) {
                        UncompressSequence(sequences[sent], uncompressed_sequence, length);
                        pvm_pkstr(uncompressed_sequence);
                    } else
                    for (int k = 0; k < seqs_sent; k++, sent++)
                        pvm_pkstr(sequences[sent]);
                pvm_send(tids[j], SENT_BY_MASTER);
            }

            while (received < n_sequences) {
                // Receive classification from any slave
                int bufid = pvm_recv(-1, SENT_BY_SLAVE);

                // If still more instances send to this slave
                if (sent < n_sequences) {
                    // Get TID of sending slave
                    int bytes, mstag, tid;
                    pvm_bufinfo(bufid, &bytes, &mstag, &tid);

                    // Sequences to sent
                    seqs_sent = MIN(SEQUENCES_BLOCK, n_sequences-sent);
                    pvm_initsend(PvmDataDefault);
                    pvm_pkint(&seqs_sent, 1, 1);
                    pvm_pkint(&sent, 1, 1); // Index of first pattern
                    if (compress_sequences)
                        for (int k = 0; k < seqs_sent; k++, sent++) {
                            UncompressSequence(sequences[sent], uncompressed_sequence, length);
                            pvm_pkstr(uncompressed_sequence);
                        } else
                        for (int k = 0; k < seqs_sent; k++, sent++)
                            pvm_pkstr(sequences[sent]);
                    pvm_send(tid, SENT_BY_MASTER);
                }

                pvm_upkint(&seqs_sent, 1, 1);
                pvm_upkint(&first_instance, 1, 1);
                for (int k = 0; k < seqs_sent; k++, received++) {
                    pvm_upkscalar(&one_model_outputs[first_instance+k], 1, 1);

                    if (first_instance+k < n_positives) {
                        if (!((first_instance+k)%VERBOSE_INTERVAL))
                            VERBOSE_OUTPUT("Positive instance", first_instance+k);
                    } else {
                        if (!((first_instance+k)%VERBOSE_INTERVAL))
                            VERBOSE_OUTPUT("Negative instance", first_instance+k-n_positives);
                    }
                }
            }

            if (!exit_after_outputs)
                for (int i = 0; i < n_positives+n_negatives; i++)
                    outputs[i][model] = one_model_outputs[i];

            fflush(parm.out_file);

            // Kill slaves
            pvm_kill(guardian);
            for (int i = 0; i < n_slaves; i++)
                pvm_kill(tids[i]);

            FREE(args[0]);
            FREE(args[1]);
            FREE(tids);
        }

        // Save outputs so far
        FILE *save = NULL;

        if (exit_after_outputs) {
            FILE *temp_file;
            scalar yet_generated[model+1];

            if ((temp_file = fopen("tempfile", "wt")) == NULL)
                OnErrorExit(1, "Cannot open temporary file for outputs");
            if (model) // Do not open for first model, it does not exist
                if ((save = fopen(models_output_file, "rt")) == NULL)
                    OnErrorExit(1, "Cannot open models output file for writing");

            for (int i = 0; i < n_positives+n_negatives; i++) {
                // Read generated outputs
                for (int j = 0; j < model; j++)
                    if (!fscanf(save, SCALAR_FORMAT, &yet_generated[j]))
                        OnErrorExit(ERR_FILE, "Error reading model outputs file");

                // Save old ones and add new one
                for (int j = 0; j < model; j++)
                    fprintf(temp_file, "%7.4lf ", yet_generated[j]);
                fprintf(temp_file, "%7.4lf\n", one_model_outputs[i]);
            }

            fclose(temp_file);
            if (save)
                fclose(save);

            unlink(models_output_file);
            rename("tempfile", models_output_file);
        } else {
            if ((save = fopen(models_output_file, "wt")) == NULL)
                OnErrorExit(1, "Cannot open models output file for writing");
            for (int i = 0; i < n_positives+n_negatives; i++) {
                for (int j = 0; j <= model; j++)
                    fprintf(save, "%7.4lf ", outputs[i][j]);
                fprintf(save, "\n");
            }
            fclose(save);
        }
    }

    // Free training data
    if (positives != NULL) {
        for (int i = 0; i < n_positives; i++)
            FREE(positives[i]);
        FREE(positives);
    }
    if (negatives != NULL) {
        for (int i = 0; i < n_negatives; i++)
            FREE(negatives[i]);
        FREE(negatives);
    }
    if (one_model_outputs != NULL)
        FREE(one_model_outputs);

    // Exit after getting the outputs?
    if (exit_after_outputs) {
        printf("Model outputs generated and saved, exiting...\n");
        exit(1);
    }

    /**********************************************/
    /* Obtain classification by different methods */
    /**********************************************/
    bool selected_models[n_models];

    fprintf(parm.out_file, "IMPORTANT!!! False positive distribution assumes the order in the negative file is: Exon, Intron, Intergenic Space\n");

    bool *positive_predicted_seq, write_roc_prc_data, write_fp_data;
    scalar *output_summary;
    MALLOC(positive_predicted_seq, bool, n_positives+n_negatives, "main: positive_predicted_seq");
    MALLOC(output_summary, scalar, n_positives+n_negatives, "main: output_summary");

    if (stepwise_methods) {
        int models_to_use = 0, list[n_models];
        for (int i = 0; i < n_models; i++)
            if (use_model[i]) {
                list[models_to_use] = i;
                models_to_use++;
            }

        // Best constructive/destructive model for each combination method
        bool best[n_stepwise_methods][n_combination_methods][n_models];
        for (int stepwise_method = SFS; stepwise_method <= RSFBS; stepwise_method++) {
            int l, r;

            // Set appropriate values for l and r
            switch (stepwise_method) {
                case SFS:
                    l = 1;
                    r = 0;
                    break;

                case SBS:
                    l = 0;
                    r = 1;
                    break;

                case LRS:
                    l = parm.plus_l;
                    r = parm.minus_r;
                    break;

                case SFFS:
                case SFBS:
                case RSFFS:
                case RSFBS:
                    l = r = 1;
                    break;
            }

            for (int combination_method = SumOfOutputs; combination_method < Maximum; combination_method++) {
                bool finish = false, finish_add = false, finish_remove = false, models[n_models], current_best[n_models];
                scalar current_performance = 0.0, previous_performance = 0.0;
                int attempts = 0;

                if (verbose)
                    printf("%8s %s\n", stepwise_method_label[stepwise_method], combination_method_label[combination_method]);

                // Only adding or removing?
                if (l <= 0)
                    finish_add = true;
                else if (r <= 0)
                    finish_remove = true;

                for (int i = 0; i < n_models; i++)
                    best[stepwise_method][combination_method][i] = false;

                if (l < r || stepwise_method == SFBS)
                    for (int i = 0; i < models_to_use; i++)
                        best[stepwise_method][combination_method][list[i]] = true;

                // RSFFS start with a random selection and then perform SFFS
                // RSFBS start with a random selection and then perform SFBS
                if (stepwise_method == RSFFS || stepwise_method == RSFBS)
                    for (int i = 0; i < models_to_use; i++)
                        best[stepwise_method][combination_method][list[i]] = (RANDOM(0.0, 1.0) < 0.5) ? true : false;

                while (!finish && attempts < parm.stepwise_max_attempts) {
                    bool added_models, removed_models;
                    int list_added[l+1], list_removed[r+1];

try_add:
                    attempts++;

                    // Add l models
                    added_models = false;
                    for (int add = 0; add < l; add++) {
                        scalar best_so_far_performance = 0.0;

                        for (int i = 0; i < n_models; i++)
                            models[i] = best[stepwise_method][combination_method][i];

                        // If all models added exit
                        int c = 0;
                        for (int i = 0; i < n_models; i++)
                            if (models[i])
                                c++;
                        if (c == models_to_use)
                            break;

                        // Add each non-present model an evaluate
                        int tested = 0;
                        for (int m = 0; m < models_to_use; m++) {
                            // Add i-th model if not present
                            if (!models[list[m]]) {
                                models[list[m]] = true;
                                tested++;
                            } else
                                continue;

                            // Evaluate
                            TP = TN = FP = FN = 0;

                            switch (combination_method) {
                                case SumOfOutputs:
                                    // SUM OF OUTPUTS
                                    // Positive sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        scalar sum = 0.0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j])
                                                sum += outputs[i][j] - desc[j].threshold;

                                        positive_predicted_seq[i] = (sum >= 0.0);
                                        output_summary[i] = sum;
                                    }
                                    break;

                                case Majority:
                                    // MAJORITY VOTING OUTPUT
                                    // Sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        int positive_votes = 0, negative_votes = 0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j]) {
                                                if (outputs[i][j] >= desc[j].threshold)
                                                    positive_votes++;
                                                else
                                                    negative_votes++;
                                            }

                                        positive_predicted_seq[i] = (positive_votes >= negative_votes);
                                        output_summary[i] = positive_votes;
                                    }
                                    break;

                                case Maximum:
                                    // MAXIMUM OUTPUT
                                    // Positive sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        scalar max_output = 0.0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j])
                                                if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                                    max_output = outputs[i][j] - desc[j].threshold;

                                        positive_predicted_seq[i] = (max_output >= 0.0);
                                        output_summary[i] = max_output;
                                    }
                                    break;
                            }

                            // auROC, auPRC or G?
                            scalar performance = 0.0;
                            if (use_auroc)
                                performance = AUROCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, false);
                            else if (use_auprc)
                                performance = AUPRCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, false);
                            else {
                                // Positive sequences
                                for (int i = 0; i < n_positives; i++)
                                    if (positive_predicted_seq[i])
                                        TP++;
                                    else
                                        FN++;

                                // Negative sequences
                                for (int i = n_positives; i < n_positives+n_negatives; i++)
                                    if (positive_predicted_seq[i])
                                        FP++;
                                    else
                                        TN++;

                                sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
                                sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);

                                performance = sqrt(sn*sp);
                            }

                            // Best?
                            if (performance > best_so_far_performance) {
                                best_so_far_performance = performance;
                                for (int i = 0; i < n_models; i++)
                                    current_best[i] = models[i];

                                current_performance = best_so_far_performance;
                            }

                            models[list[m]] = false;
                        }

                        // Add best if any added
                        if (tested) {
                            for (int i = 0; i < n_models; i++)
                                if (best[stepwise_method][combination_method][i] != current_best[i]) {
                                    best[stepwise_method][combination_method][i] = true;
                                    list_added[add] = i;
                                    added_models = true;

                                    if (verbose)
                                        printf("%s (%s): + model %3d with performance %lf\n", stepwise_method_label[stepwise_method], combination_method_label[combination_method], i+1,
                                               current_performance);
                                }
                        } else
                            list_added[add] = -1;
                    }

                    if (added_models) {
                        // Forward step must improve to add
                        if (current_performance > previous_performance) {
                            previous_performance = current_performance;
                            finish_add = false;

                            // If SFFS keep trying to remove models
                            if (stepwise_method == SFBS || stepwise_method == RSFBS)
                                goto try_add;
                        } else {
                            // Remove last added models
                            for (int i = 0; i < l; i++)
                                if (list_added[i] >= 0) {
                                    best[stepwise_method][combination_method][list_added[i]] = false;
                                    if (verbose)
                                        printf("%s (%s): - model %3d\n", stepwise_method_label[stepwise_method], combination_method_label[combination_method], list_added[i]+1);
                                }

                            finish_add = true;
                        }
                    } else
                        finish_add = true;

try_removing:
                    attempts++;

                    // Remove models
                    removed_models = false;
                    for (int remove = 0; remove < r; remove++) {
                        scalar best_so_far_performance = 0.0;

                        for (int i = 0; i < n_models; i++)
                            models[i] = best[stepwise_method][combination_method][i];

                        // If only one model exit
                        int c = 0;
                        for (int i = 0; i < n_models; i++)
                            if (models[i])
                                c++;
                        if (c == 1)
                            break;

                        // Delete each present model an evaluate
                        int tested = 0;
                        for (int m = 0; m < models_to_use; m++) {
                            // Remove i-th model if present
                            if (models[list[m]]) {
                                models[list[m]] = false;

                                tested++;
                            } else
                                continue;

                            // Evaluate
                            TP = TN = FP = FN = 0;

                            switch (combination_method) {
                                case SumOfOutputs:
                                    // SUM OF OUTPUTS
                                    // Positive sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        scalar sum = 0.0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j])
                                                sum += outputs[i][j] - desc[j].threshold;

                                        positive_predicted_seq[i] = (sum >= 0.0);
                                        output_summary[i] = sum;
                                    }
                                    break;

                                case Majority:
                                    // MAJORITY VOTING OUTPUT
                                    // Sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        int positive_votes = 0, negative_votes = 0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j]) {
                                                if (outputs[i][j] >= desc[j].threshold)
                                                    positive_votes++;
                                                else
                                                    negative_votes++;
                                            }

                                        positive_predicted_seq[i] = (positive_votes >= negative_votes);
                                        output_summary[i] = positive_votes;
                                    }
                                    break;

                                case Maximum:
                                    // MAXIMUM OUTPUT
                                    // Positive sequences
                                    for (int i = 0; i < n_positives+n_negatives; i++) {
                                        scalar max_output = 0.0;
                                        for (int j = 0; j < n_models; j++)
                                            if (models[j])
                                                if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                                    max_output = outputs[i][j] - desc[j].threshold;

                                        positive_predicted_seq[i] = (max_output >= 0.0);
                                        output_summary[i] = max_output;
                                    }
                                    break;
                            }

                            // auROC, auPRC or G?
                            scalar performance = 0.0;
                            if (use_auroc)
                                performance = AUROCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, false);
                            else if (use_auprc)
                                performance = AUPRCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, false);
                            else {
                                // Positive sequences
                                for (int i = 0; i < n_positives; i++)
                                    if (positive_predicted_seq[i])
                                        TP++;
                                    else
                                        FN++;

                                // Negative sequences
                                for (int i = n_positives; i < n_positives+n_negatives; i++)
                                    if (positive_predicted_seq[i])
                                        FP++;
                                    else
                                        TN++;

                                sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
                                sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);

                                performance = sqrt(sn*sp);
                            }

                            // Best?
                            if (performance > best_so_far_performance) {
                                best_so_far_performance = performance;
                                for (int i = 0; i < n_models; i++)
                                    current_best[i] = models[i];

                                current_performance = best_so_far_performance;
                            }

                            models[list[m]] = true;
                        }

                        // Remove worst if any removed
                        if (tested) {
                            for (int i = 0; i < n_models; i++)
                                if (best[stepwise_method][combination_method][i] != current_best[i]) {
                                    best[stepwise_method][combination_method][i] = false;
                                    list_removed[remove] = i;
                                    removed_models = true;

                                    if (verbose)
                                        printf("%s (%s): - model %3d with performance %lf\n", stepwise_method_label[stepwise_method], combination_method_label[combination_method], i+1,
                                               current_performance);
                                }
                        } else
                            list_removed[remove] = -1;
                    }

                    if (removed_models) {
                        // Backward step must deteriorate to keep
                        if (current_performance >= previous_performance) {
                            previous_performance = current_performance;
                            finish_remove = false;

                            // If SFFS keep trying to remove models
                            if (stepwise_method == SFFS || stepwise_method == RSFBS)
                                goto try_removing;
                        } else {
                            // Add last removed models
                            for (int i = 0; i < r; i++)
                                if (list_removed[i] >= 0) {
                                    best[stepwise_method][combination_method][list_removed[i]] = true;
                                    if (verbose)
                                        printf("%s (%s): + model %3d\n", stepwise_method_label[stepwise_method], combination_method_label[combination_method], list_removed[i]+1);
                                }

                            finish_remove = true;
                        }
                    } else
                        finish_remove = true;

                    // Finish?
                    if (r <= 0 && finish_add)
                        finish = true;
                    else if (l <= 0 && finish_remove)
                        finish = true;
                    else
                        finish = finish_add & finish_remove;
                }
            }
        }

        // Free training data
        for (int i = 0; i < n_positives + n_negatives; i++)
            FREE(outputs[i]);
        FREE(outputs);

        // Results with test model
        // Read outputs
        int n_lines = -1;
        if (!access(test_models_output_file, R_OK)) {
            // Open and count lines
            FILE *modelf;
            if ((modelf = fopen(test_models_output_file, "rt")) == NULL)
                OnErrorExit(1, "Cannot open models output file for reading");
            char line[MAX_LENGTH], *next_output;
            do {
                if (fgets(line, MAX_LENGTH, modelf) == NULL && !feof(modelf))
                    OnErrorExit(ERR_FILE, "Error reading model outputs file");
                n_lines++;
            } while (!feof(modelf));

            MALLOC(outputs, scalar *, n_lines, "main: outputs");
            for (int i = 0; i < n_lines; i++)
                MALLOC(outputs[i], scalar, n_models, "main: outputs[i]");

            // Read first line and count previous models
            rewind(modelf);
            if (fgets(line, MAX_LENGTH, modelf) == NULL)
                OnErrorExit(ERR_FILE, "Error reading model outputs file");
            int n_generated_models = 0;
            next_output = strtok(line, " \t");
            while (next_output != NULL && strncmp(next_output, "\n", 1)) {
                n_generated_models++;
                next_output = strtok(NULL, " \t");
            }

            if (n_generated_models != n_models)
                OnErrorExit(1, "Not enough columns in test data outputs");

            // Read outputs
            rewind(modelf);
            for (int i = 0; i < n_lines; i++)
                for (int j = 0; j < n_generated_models; j++)
                    if (!fscanf(modelf, SCALAR_FORMAT, &outputs[i][j]))
                        OnErrorExit(ERR_FILE, "Error reading model outputs file");

            fclose(modelf);
        } else
            OnErrorExit(1, "No file for outputs for test data");

        // Count instances of test files and check
        LoadGlobalParameters(&parm, desc[0].parameters_file);

        // Count instances
        sprintf(file_name, "%s", test_pos_file_name);
        if (access(file_name, R_OK))
            sprintf(file_name, "%s/%s/%s", home, parm.datad, test_pos_file_name);
        // No file, no positive sequences
        if (access(file_name, R_OK))
            n_positives = 0;
        else
            ReadFastaFile(file_name, NULL, &n_positives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                          parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

        fprintf(parm.out_file, "Test positive sequences: %d\n", n_positives);

        sprintf(file_name, "%s", test_neg_file_name);
        if (access(file_name, R_OK))
            sprintf(file_name, "%s/%s/%s", home, parm.datad, test_neg_file_name);
        // No file, no negative sequences
        if (access(file_name, R_OK))
            n_negatives = 0;
        else
            ReadFastaFile(file_name, NULL, &n_negatives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                          parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

        fprintf(parm.out_file, "Test negative sequences: %d\n", n_negatives);

        if (n_lines != n_positives + n_negatives) {
            printf("Generated outputs: %d\tRead from file: %d\n", n_lines, n_positives+n_negatives);
            OnErrorExit(1, "Generated outputs and read data files do not match");
        }

        REALLOC(positive_predicted_seq, bool, n_positives+n_negatives, "main: positive_predicted_seq");
        REALLOC(output_summary, scalar, n_positives+n_negatives, "main: output_summary");

        for (int stepwise_method = SFS; stepwise_method <= RSFBS; stepwise_method++)
            for (int combination_method = SumOfOutputs; combination_method < Maximum; combination_method++) {
                int used_models = 0;

                // Evaluate best models
                switch (combination_method) {
                    case SumOfOutputs:
                        // SUM OF OUTPUTS
                        // Positive sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            scalar sum = 0.0;
                            for (int j = 0; j < n_models; j++)
                                if (best[stepwise_method][combination_method][j])
                                    sum += outputs[i][j] - desc[j].threshold;

                            positive_predicted_seq[i] = (sum >= 0.0);
                            output_summary[i] = sum;
                        }
                        break;

                    case Majority:
                        // MAJORITY VOTING OUTPUT
                        // Count models
                        for (int i = 0; i < n_models; i++)
                            if (best[stepwise_method][combination_method][i])
                                used_models++;

                        // Sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            int positive_votes = 0;
                            for (int j = 0; j < n_models; j++)
                                if (best[stepwise_method][combination_method][j])
                                    if (outputs[i][j] >= desc[j].threshold)
                                        positive_votes++;

                            positive_predicted_seq[i] = (positive_votes >= used_models-positive_votes);
                            output_summary[i] = positive_votes;
                        }
                        break;

                    case Maximum:
                        // MAXIMUM OUTPUT
                        // Positive sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            scalar max_output = 0.0;
                            for (int j = 0; j < n_models; j++)
                                if (best[stepwise_method][combination_method][j])
                                    if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                        max_output = outputs[i][j] - desc[j].threshold;

                            positive_predicted_seq[i] = (max_output >= 0.0);
                            output_summary[i] = max_output;
                        }
                        break;
                }

                TP = TN = FP = FN = 0;
                // Positive sequences
                for (int i = 0; i < n_positives; i++)
                    if (positive_predicted_seq[i])
                        TP++;
                    else
                        FN++;

                // Negative sequences
                for (int i = n_positives; i < n_positives+n_negatives; i++)
                    if (positive_predicted_seq[i])
                        FP++;
                    else
                        TN++;

                // auROC
                sprintf(file_name, "ROC-%s-%s", combination_method_label[combination_method], stepwise_method_label[stepwise_method]);
                AddModelsToFileName(file_name, best[stepwise_method][combination_method], n_models);
                printf("Obtaining auROC for %s %s model\n", combination_method_label[combination_method], stepwise_method_label[stepwise_method]);
                fprintf(parm.out_file, "%s auROC: %6.4lf ", stepwise_method_label[stepwise_method], AUROCFromOutputs(output_summary, n_positives, n_negatives, true, file_name, false));

                // auPRC
                sprintf(file_name, "PRC-%s-%s", combination_method_label[combination_method], stepwise_method_label[stepwise_method]);
                AddModelsToFileName(file_name, best[stepwise_method][combination_method], n_models);
                printf("Obtaining auPRC for %s %s model\n", combination_method_label[combination_method], stepwise_method_label[stepwise_method]);
                fprintf(parm.out_file, "auPRC: %6.4lf ", AUPRCFromOutputs(output_summary, n_positives, n_negatives, true, file_name, true));

                sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
                sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);
                fprintf(parm.out_file, "%s TP: %4d FN: %3d TN: %7d FP: %7d Sp: %6.4lf Sn: %6.4lf G-mean: %6.4lf ", combination_method_label[combination_method], TP, FN, TN, FP, sp, sn, sqrt(sp*sn));
                fprintf(parm.out_file, "Models:");
                for (int i = 0; i < n_models; i++)
                    if (best[stepwise_method][combination_method][i])
                        fprintf(parm.out_file, " %2d", i+1);
                fprintf(parm.out_file, "\n");
                fflush(parm.out_file);
                fsync(fileno(parm.out_file));
            }
    } else if (best_combination) {
        // Best combination method using all the models
        t_CombinationMethod best_combination_method = -1;
        scalar best_performance = -1.0;

        for (int combination_method = SumOfOutputs; combination_method < Maximum; combination_method++) {
            int used_models = 0;
            // Evaluate
            TP = TN = FP = FN = 0;

            switch (combination_method) {
                case SumOfOutputs:
                    // SUM OF OUTPUTS
                    // Positive sequences
                    for (int i = 0; i < n_positives+n_negatives; i++) {
                        scalar sum = 0.0;
                        for (int j = 0; j < n_models; j++)
                            if (use_model[j])
                                sum += outputs[i][j] - desc[j].threshold;

                        positive_predicted_seq[i] = (sum >= 0.0);
                        output_summary[i] = sum;
                    }
                    break;

                case Majority:
                    // MAJORITY VOTING OUTPUT
                    for (int i = 0; i < n_models; i++)
                        if (use_model[i])
                            used_models++;

                    // Sequences
                    for (int i = 0; i < n_positives+n_negatives; i++) {
                        int positive_votes = 0;
                        for (int j = 0; j < n_models; j++)
                            if (use_model[j])
                                if (outputs[i][j] >= desc[j].threshold)
                                    positive_votes++;

                        positive_predicted_seq[i] = (positive_votes >= used_models-positive_votes);
                        output_summary[i] = positive_votes;
                    }
                    break;

                case Maximum:
                    // MAXIMUM OUTPUT
                    // Positive sequences
                    for (int i = 0; i < n_positives+n_negatives; i++) {
                        scalar max_output = 0.0;
                        for (int j = 0; j < n_models; j++)
                            if (use_model[j])
                                if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                    max_output = outputs[i][j] - desc[j].threshold;

                        positive_predicted_seq[i] = (max_output >= 0.0);
                        output_summary[i] = max_output;
                    }
                    break;
            }

            // auROC, auPRC or G?
            scalar performance = 0.0;
            if (use_auroc)
                performance = AUROCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, false);
            else if (use_auprc)
                performance = AUPRCFromOutputs(output_summary, n_positives, n_negatives, false, NULL, true);
            else {
                // Positive sequences
                for (int i = 0; i < n_positives; i++)
                    if (positive_predicted_seq[i])
                        TP++;
                    else
                        FN++;

                // Negative sequences
                for (int i = n_positives; i < n_positives+n_negatives; i++)
                    if (positive_predicted_seq[i])
                        FP++;
                    else
                        TN++;

                sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
                sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);

                performance = sqrt(sn*sp);
            }

            // Best?
            // SFS must improve to add, destructive must deteriorate to keep
            if (performance > best_performance) {
                best_performance = performance;
                best_combination_method = combination_method;
            }
        }

        // Free training data
        for (int i = 0; i < n_positives + n_negatives; i++)
            FREE(outputs[i]);
        FREE(outputs);

        // Results with test model
        // Read outputs
        int n_lines = -1;
        if (!access(test_models_output_file, R_OK)) {
            // Open and count lines
            FILE *modelf;
            if ((modelf = fopen(test_models_output_file, "rt")) == NULL)
                OnErrorExit(1, "Cannot open models output file for reading");
            char line[MAX_LENGTH], *next_output;
            do {
                if (fgets(line, MAX_LENGTH, modelf) == NULL && !feof(modelf))
                    OnErrorExit(ERR_FILE, "Error reading model outputs file");
                n_lines++;
            } while (!feof(modelf));

            MALLOC(outputs, scalar *, n_lines, "main: outputs");
            for (int i = 0; i < n_lines; i++)
                MALLOC(outputs[i], scalar, n_models, "main: outputs[i]");

            // Read first line and count previous models
            rewind(modelf);
            if (fgets(line, MAX_LENGTH, modelf) == NULL)
                OnErrorExit(ERR_FILE, "Error reading model outputs file");
            int n_generated_models = 0;
            next_output = strtok(line, " \t");
            while (next_output != NULL && strncmp(next_output, "\n", 1)) {
                n_generated_models++;
                next_output = strtok(NULL, " \t");
            }

            if (n_generated_models != n_models)
                OnErrorExit(1, "Not enough columns in test data outputs");

            // Read outputs
            rewind(modelf);
            for (int i = 0; i < n_lines; i++)
                for (int j = 0; j < n_generated_models; j++)
                    if (!fscanf(modelf, SCALAR_FORMAT, &outputs[i][j]))
                        OnErrorExit(ERR_FILE, "Error reading model outputs file");

            fclose(modelf);
        } else
            OnErrorExit(1, "No file for outputs for test data");

        // Count instances of test files and check
        LoadGlobalParameters(&parm, desc[0].parameters_file);

        // Count instances
        sprintf(file_name, "%s", test_pos_file_name);
        if (access(file_name, R_OK))
            sprintf(file_name, "%s/%s/%s", home, parm.datad, test_pos_file_name);
        // No file, no positive sequences
        if (access(file_name, R_OK))
            n_positives = 0;
        else
            ReadFastaFile(file_name, NULL, &n_positives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                          parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

        fprintf(parm.out_file, "Test positive sequences: %d\n", n_positives);

        sprintf(file_name, "%s", test_neg_file_name);
        if (access(file_name, R_OK))
            sprintf(file_name, "%s/%s/%s", home, parm.datad, test_neg_file_name);
        // No file, no negative sequences
        if (access(file_name, R_OK))
            n_negatives = 0;
        else
            ReadFastaFile(file_name, NULL, &n_negatives, &length, 0, false, 0, true, NULL, 0, false, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters,
                          parm.do_not_check_sequences, letters, &parm.length, NULL, compress_sequences);

        fprintf(parm.out_file, "Test negative sequences: %d\n", n_negatives);

        if (n_lines != n_positives + n_negatives) {
            printf("Generated outputs: %d\tRead from file: %d\n", n_lines, n_positives+n_negatives);
            OnErrorExit(1, "Generated outputs and read data files do not match");
        }

        REALLOC(positive_predicted_seq, bool, n_positives+n_negatives, "main: positive_predicted_seq");
        REALLOC(output_summary, scalar, n_positives+n_negatives, "main: output_summary");

        // Evaluate best models
        switch (best_combination_method) {
                int used_models;

            case SumOfOutputs:
                // SUM OF OUTPUTS
                // Positive sequences
                for (int i = 0; i < n_positives+n_negatives; i++) {
                    scalar sum = 0.0;
                    for (int j = 0; j < n_models; j++)
                        if (use_model[j])
                            sum += outputs[i][j] - desc[j].threshold;

                    positive_predicted_seq[i] = (sum >= 0.0);
                    output_summary[i] = sum;
                }
                break;

            case Majority:
                // MAJORITY VOTING OUTPUT
                // Count models
                used_models = 0;
                for (int i = 0; i < n_models; i++)
                    if (use_model[i])
                        used_models++;

                // Sequences
                for (int i = 0; i < n_positives+n_negatives; i++) {
                    int positive_votes = 0;
                    for (int j = 0; j < n_models; j++)
                        if (use_model[j])
                            if (outputs[i][j] >= desc[j].threshold)
                                positive_votes++;

                    positive_predicted_seq[i] = (positive_votes >= used_models-positive_votes);
                    output_summary[i] = positive_votes;
                }
                break;

            case Maximum:
                // MAXIMUM OUTPUT
                // Positive sequences
                for (int i = 0; i < n_positives+n_negatives; i++) {
                    scalar max_output = 0.0;
                    for (int j = 0; j < n_models; j++)
                        if (use_model[j])
                            if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                max_output = outputs[i][j] - desc[j].threshold;

                    positive_predicted_seq[i] = (max_output >= 0.0);
                    output_summary[i] = max_output;
                }
                break;
        }

        TP = TN = FP = FN = 0;
        // Positive sequences
        for (int i = 0; i < n_positives; i++)
            if (positive_predicted_seq[i])
                TP++;
            else
                FN++;

        // Negative sequences
        for (int i = n_positives; i < n_positives+n_negatives; i++)
            if (positive_predicted_seq[i])
                FP++;
            else
                TN++;

        // auROC
        sprintf(file_name, "ROC-%s", combination_method_label[best_combination_method]);
        AddModelsToFileName(file_name, use_model, n_models);
        fprintf(parm.out_file, "auROC: %6.4lf ", AUROCFromOutputs(output_summary, n_positives, n_negatives, true, file_name, false));

        // auPRC
        sprintf(file_name, "PRC-%s", combination_method_label[best_combination_method]);
        AddModelsToFileName(file_name, use_model, n_models);
        fprintf(parm.out_file, "auPRC: %6.4lf ", AUPRCFromOutputs(output_summary, n_positives, n_negatives, true, file_name, true));

        sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
        sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);
        fprintf(parm.out_file, "%8s TP: %4d FN: %3d TN: %7d FP: %7d Sp: %6.4lf Sn: %6.4lf G-mean: %6.4lf ", combination_method_label[best_combination_method], TP, FN, TN, FP, sp, sn, sqrt(sp*sn));
        fprintf(parm.out_file, "Models:");
        for (int i = 0; i < n_models; i++)
            if (use_model[i])
                fprintf(parm.out_file, " %2d", i+1);
        fprintf(parm.out_file, "\n");
        fflush(parm.out_file);
    } else {
        // Count usable models
        int n_usable_models = 0, usable_models_list[n_models];
        for (int i = 0; i < n_models; i++)
            if (use_model[i]) {
                usable_models_list[n_usable_models] = i;
                n_usable_models++;
            }
        bool usable_selected_models[n_usable_models];

        // All possible combinations of usable models
        for (int m = 1; m < pow(2.0, n_usable_models); m++) {
            // Set models to use
            int number = m, used_models = 0;
            for (int i = 0; i < n_usable_models; i++) {
                if ((usable_selected_models[i] = number % 2))
                    used_models++;
                number /= 2;
            }

            // Convert selected to actual models
            for (int i = 0; i < n_models; i++)
                selected_models[i] = false;
            for (int i = 0; i < n_usable_models; i++)
                selected_models[usable_models_list[i]] = usable_selected_models[i];

            // Print models
            printf("Models: ");
            for (int i = 0; i < n_models; i++)
                if (selected_models[i])
                    printf("%d ", i+1);
            printf("\n");

            // Check whether ROC data must be written
            int converted = ConvertToNumberWithoutMask(m, usable_models_list, n_usable_models);
            write_roc_prc_data = false;
            for (int i = 0; i < n_roc; i++)
                if (roc_models[i] == converted) {
                    write_roc_prc_data = true;
                    break;
                }

            // Check whether FP data must be written
            write_fp_data = false;
            for (int i = 0; i < n_fp; i++)
                if (fp_models[i] == converted) {
                    write_fp_data = true;
                    break;
                }

            for (int combination_method = SumOfOutputs; combination_method < Maximum; combination_method++) {
                TP = TN = FP = FN = 0;

                switch (combination_method) {
                    case SumOfOutputs:
                        // SUM OF OUTPUTS
                        // Positive sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            scalar sum = 0.0;
                            for (int j = 0; j < n_models; j++)
                                if (selected_models[j])
                                    sum += outputs[i][j] - desc[j].threshold;

                            positive_predicted_seq[i] = (sum >= 0.0);
                            output_summary[i] = sum;
                        }
                        break;

                    case Majority:
                        // MAJORITY VOTING OUTPUT
                        // Sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            int positive_votes = 0;
                            for (int j = 0; j < n_models; j++)
                                if (selected_models[j])
                                    if (outputs[i][j] >= desc[j].threshold)
                                        positive_votes++;

                            positive_predicted_seq[i] = (positive_votes >= used_models-positive_votes);
                            output_summary[i] = positive_votes;
                        }
                        break;

                    case Maximum:
                        // MAXIMUM OUTPUT
                        // Positive sequences
                        for (int i = 0; i < n_positives+n_negatives; i++) {
                            scalar max_output = 0.0;
                            for (int j = 0; j < n_models; j++)
                                if (selected_models[j])
                                    if (fabs(outputs[i][j] - desc[j].threshold) > fabs(max_output))
                                        max_output = outputs[i][j] - desc[j].threshold;
                            positive_predicted_seq[i] = (max_output >= 0.0);
                            output_summary[i] = max_output;
                        }
                        break;
                }

                // Positive sequences
                for (int i = 0; i < n_positives; i++)
                    if (positive_predicted_seq[i])
                        TP++;
                    else
                        FN++;

                // Negative sequences
                ex = in = ig = 0;       // For false positives distribution
                for (int i = n_positives; i < n_positives+n_negatives; i++)
                    if (positive_predicted_seq[i]) {
                        FP++;
                        if (i - n_positives < offset[0])
                            ex++;
                        else if (i - n_positives < offset[1])
                            in++;
                        else
                            ig++;
                    } else
                        TN++;

                // auROC
                // Filename for auROC data if apply
                if (write_roc_prc_data) {
                    sprintf(file_name, "ROC-%s", combination_method_label[combination_method]);
                    AddModelsToFileName(file_name, selected_models, n_models);
                }
                fprintf(parm.out_file, "auROC: %6.4lf\t", AUROCFromOutputs(output_summary, n_positives, n_negatives, write_roc_prc_data, file_name, false));

                // auPRC
                // Filename for auPRC data if apply
                if (write_roc_prc_data) {
                    sprintf(file_name, "PRC-%s", combination_method_label[combination_method]);
                    AddModelsToFileName(file_name, selected_models, n_models);
                }
                fprintf(parm.out_file, "auPRC: %6.4lf\t", AUPRCFromOutputs(output_summary, n_positives, n_negatives, write_roc_prc_data, file_name, true));

                // FP data
                if (write_fp_data) {
                    FILE *fp_file;

                    sprintf(file_name, "FP-%s", combination_method_label[combination_method]);
                    AddModelsToFileName(file_name, selected_models, n_models);

                    if ((fp_file = fopen(file_name, "w")) == NULL)
                        OnErrorExit(1, "Cannot open file for false positives");

                    // Write negative samples index classified as positives: FP
                    for (int i = n_positives; i < n_positives+n_negatives; i++)
                        if (positive_predicted_seq[i])
                            fprintf(fp_file, "%d\n", i-n_positives+1);
                    fclose(fp_file);
                }

                sn = (TP+FN > 0 ? TP/((scalar) TP+FN) : 0.0);
                sp = (TN+FP > 0 ? TN/((scalar) TN+FP) : 0.0);
                fprintf(parm.out_file, "%8s TP: %4d FN: %3d TN: %7d FP: %7d\tSp: %6.4lf Sn: %6.4lf G-mean: %6.4lf\t", combination_method_label[combination_method], TP, FN, TN, FP, sp, sn,
                        sqrt(sp*sn));
                fprintf(parm.out_file, "FP(ex): %5d FP(in): %6d FP(ig): %7d\t", ex, in, ig);
                fprintf(parm.out_file, "Models:");
                for (int i = 0; i < n_models; i++)
                    if (selected_models[i])
                        fprintf(parm.out_file, " %2d", i+1);
                fprintf(parm.out_file, "\n");
                fflush(parm.out_file);

                // If just one model do not repeat for the same results
                if (used_models == 1)
                    break;
            }
        }
    }

    // Used time
    time_t end_time = time(NULL) - init_time;
    fprintf(parm.out_file, "Time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);
    fprintf(parm.out_file, "Ok\n");

    // Close output file
    fclose(parm.out_file);

    // Sort out file by G-mean
    // Run algorithm
    char command[1000];
    if (!best_combination && !stepwise_methods) {
        sprintf(command, "grep \"auROC:\" %s | sort -n -r -k 19 > %s.sorted.g", argv[2], argv[2]);
        system(command);
        sprintf(command, "grep \"auROC:\" %s | sort -n -r -k 2 > %s.sorted.auroc", argv[2], argv[2]);
        system(command);
        sprintf(command, "grep \"auROC:\" %s | sort -n -r -k 4 > %s.sorted.auprc", argv[2], argv[2]);
        system(command);
    }

    return 0;
}
