/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <time.h>
#include "type.h"
#include "macros.h"

void TrainNeuralNetwork(const t_Parameters *parm, t_Recognizer *R, char **positives, int n_positives, char **negatives, int n_negatives, int length, int n_hidden)
{
    scalar HyperbolicTangent(scalar, scalar, scalar);
    int CompareDouble(const void *, const void *);

    scalar **data, *momentum_ho, **momentum_ih, **change_ih, *change_ho, *h_outputs, *h_delta, error, *desired, change, delta;
    int *labels, instances, features;

    // Convert data for training
    ConvertToScalarData(positives, n_positives, negatives, n_negatives, length, &data, &labels, &instances, &features, 0);

    // Allocate network
    R->h_nodes = n_hidden;
    R->A = parm->htan_a;
    R->B = parm->htan_b;
    MALLOC(R->ih, scalar *, n_hidden, "TrainNeuralNetwork: R->ih");
    for (int i = 0; i < n_hidden; i++)
        MALLOC(R->ih[i], scalar, features + 1, "TrainNeuralNetwork: R->ih[i]");
    MALLOC(R->ho, scalar, n_hidden + 1, "TrainNeuralNetwork: R->ho");

    // Random initialization of weights
    for (int i = 0; i < n_hidden; i++)
        for (int j = 0; j < features + 1; j++)
            R->ih[i][j] = RANDOM(-0.1, 0.1);
    for (int i = 0; i < n_hidden + 1; i++)
        R->ho[i] = RANDOM(-0.1, 0.1);

    // Momentum term and cumulative changes
    MALLOC(momentum_ih, scalar *, n_hidden, "TrainNeuralNetwork: momentum_ih");
    MALLOC(change_ih, scalar *, n_hidden, "TrainNeuralNetwork: change_ih");
    for (int i = 0; i < n_hidden; i++) {
        MALLOC(momentum_ih[i], scalar, features + 1, "TrainNeuralNetwork: momentum_ih[i]");
        MALLOC(change_ih[i], scalar, features + 1, "TrainNeuralNetwork: change_ih[i]");
        for (int j = 0; j < features + 1; j++)
            momentum_ih[i][j] = change_ih[i][j] = 0.0;
    }

    MALLOC(momentum_ho, scalar, n_hidden + 1, "TrainNeuralNetwork: momentum_ho");
    MALLOC(change_ho, scalar, n_hidden + 1, "TrainNeuralNetwork: change_ho");
    for (int j = 0; j < n_hidden + 1; j++)
        momentum_ho[j] = change_ho[j] = 0.0;

    MALLOC(h_outputs, scalar, n_hidden, "TrainNeuralNetwork: h_outputs");
    MALLOC(h_delta, scalar, n_hidden, "TrainNeuralNetwork: delta");

    // Fill desired outputs
    MALLOC(desired, scalar, instances, "TrainNeuralNetwork: desired");
    for (int i = 0; i < instances; i++)
        if (labels[i] == 0)      // Positive instance
            desired[i] = 1.0;
        else
            desired[i] = -1.0;


    for (int iter = 0; iter < parm->n_epochs; iter++) {
        for (int k = 0; k < parm->epoch_size; k++) {
            // Choose random pattern
            int pattern = RANDOM(0, instances);

            for (int i = 0; i < n_hidden; i++) {
                scalar sum = 0.0;
                for (int j = 0; j < features; j++)
                    sum += R->ih[i][j] * data[pattern][j];

                // Bias
                sum += R->ih[i][features];

                h_outputs[i] = HyperbolicTangent(sum, R->A, R->B);
            }

            // Output node
            scalar output = 0.0;
            for (int j = 0; j < n_hidden; j++)
                output += h_outputs[j] * R->ho[j];
            // Bias
            output += R->ho[n_hidden];

            output = HyperbolicTangent(output, R->A, R->B);

            // Obtain error for output nodes.
            error = desired[pattern] - output;

            // Compute delta's for output and hidden nodes.
            delta = error * (parm->flat_spot + (R->B / R->A) * (R->A - output) * (R->A + output));

            for (int i = 0; i < n_hidden; i++) {
                h_delta[i] = delta * R->ho[i];
                h_delta[i] *= parm->flat_spot + (R->B / R->A) * (R->A - h_outputs[i]) * (R->A + h_outputs[i]);
            }

            // Obtain weight update
            // Input
            for (int i = 0; i < n_hidden; i++) {
                for (int j = 0; j < features; j++)
                    change_ih[i][j] += parm->eta * h_delta[i] * data[pattern][j]; // - parm->lambda * PENALTY ( R->ih[i][j] );

                // Bias unit connection
                change_ih[i][features] += parm->eta * h_delta[i];
            }

            // Output
            for (int j = 0; j < n_hidden; j++)
                change_ho[j] +=
                    parm->eta * delta * h_outputs[j]; //- parm->lambda * PENALTY ( classifier->ho[i][j] );
            // Bias unit connection
            change_ho[n_hidden] = parm->eta * delta;
        }

        // Perform weight update
        // Input -> Hidden
        for (int i = 0; i < n_hidden; i++)
            for (int j = 0; j < features + 1; j++) {
                change = change_ih[i][j] + parm->alpha * momentum_ih[i][j];
                momentum_ih[i][j] = change;
                change_ih[i][j] = 0.0;
                R->ih[i][j] += change;
            }


        // Hidden -> Output
        for (int j = 0; j < n_hidden + 1; j++) {
            change = change_ho[j] + parm->alpha * momentum_ho[j];
            momentum_ho[j] = change;
            change_ho[j] = 0.0;
            R->ho[j] += change;
        }

    }

    // Free training data
    FREE(labels);
    for (int i = 0; i < instances; i++)
        FREE(data[i]);
    FREE(data);
    FREE(momentum_ho);
    FREE(change_ho);
    for (int i = 0; i < n_hidden; i++) {
        FREE(momentum_ih[i]);
        FREE(change_ih[i]);
    }
    FREE(momentum_ih);
    FREE(change_ih);
    FREE(h_outputs);
    FREE(h_delta);
    FREE(desired);

    // Get threshold to obtain desired sensitivity = TP/(TP+FN)
    // Get all scores and ...
    scalar *sc;
    ALLOCA(sc, scalar, n_positives, "GetPositionWeightMatrix: sc");
    for (int i = 0; i < n_positives; i++)
        sc[i] = NeuralNetworkOutput(R, positives[i], length);

    // ...sort them
    qsort(sc, n_positives, sizeof(scalar), CompareDouble);

    // Get threshold for desired sensitivity
    if (parm->sensitivity > 0.0)
        R->threshold = sc[(int)(n_positives * (1.0 - parm->sensitivity)) ];
    else    // Use 80% as default
        R->threshold = sc[(int)(n_positives * 0.2) ];
}


scalar NeuralNetworkOutput(t_Recognizer *R, char *sequence, int length)
{
    scalar instance[4*length], h_outputs[R->h_nodes];

    ConvertToScalarSequence(sequence, length, instance);

    for (int i = 0; i < R->h_nodes; i++) {
        scalar sum = 0.0;
        for (int j = 0; j < length*4; j++)
            sum += R->ih[i][j] * instance[j];

        // Bias
        sum += R->ih[i][length*4];

        h_outputs[i] = HyperbolicTangent(sum, R->A, R->B);
    }

    // Output node
    scalar output = 0.0;
    for (int j = 0; j < R->h_nodes; j++)
        output += h_outputs[j] * R->ho[j];
    // Bias
    output += R->ho[R->h_nodes];

    output = HyperbolicTangent(output, R->A, R->B);

    return output;
}

scalar HyperbolicTangent(scalar x, scalar a, scalar b)
{
    if (x < -45.0)
        return -a;
    else if (x > 45.0)
        return a;
    else
        return a * tanh(b * x);
}
