/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "type.h"
#include "macros.h"

// Functions for noise filtering

/*
 * Koplowitz, J., Brown, T.A., 1981. On the relation of performance to editing in nearest neighbor rules. Pattern Recognition 13, 251–255.
 */

void GeneralizedEditingFilter (const t_Parameters *parm, char ***positives, char ***negatives, int *n_positive, int *n_negative, int k, int kprime, int length)
{
    t_SequenceClass neighbors[k];
    int np, nn, removed, reassigned;

    removed = reassigned = 0;

    // For all instances, positives and negatives
    for (int i = 0; i < *n_positive; i++) {
        // Get k neighbors
        GetKNeighbors((*positives)[i], *positives, *negatives, *n_positive, *n_negative, length, k, neighbors);

        // Count positive and negative neighbors
        np = nn = 0;
        for (int j = 0; j < k; j++)
            if (neighbors[j] == POSITIVE)
                np++;
            else
                nn++;

        // If count greater or equal than k' add to class, else remove
        if (np < kprime && nn < kprime) {
            removed++;

            // Remove
            (*n_positive)--;
            FREE((*positives)[i]);
            (*positives)[i] = (*positives)[*n_positive];
            REALLOC(*positives, char *, *n_positive, "GeneralizedEditingFilter: *positives");

            // Revisit
            i--;
        } else if (nn >= kprime) {
            reassigned++;

            // Reassign
            (*n_positive)--;
            (*n_negative)++;
            REALLOC(*negatives, char *, *n_negative, "GeneralizedEditingFilter: *negatives")
            (*negatives)[*n_negative-1] = (*positives)[i];
            (*positives)[i] = (*positives)[*n_positive];
            REALLOC(*positives, char *, *n_positive, "GeneralizedEditingFilter: *positives");

            // Revisit
            i--;
        }
    }

    fprintf(parm->out_file, "Noise filtering POSITIVE removed instances: %d\n", removed);
    fprintf(parm->out_file, "Noise filtering POSITIVE reassigned instances: %d\n", reassigned);

    removed = reassigned = 0;

    for (int i = 0; i < *n_negative; i++) {
        // Get k neighbors
        GetKNeighbors((*negatives)[i], *positives, *negatives, *n_positive, *n_negative, length, k, neighbors);

        // Count positive and negative neighbors
        np = nn = 0;
        for (int j = 0; j < k; j++)
            if (neighbors[j] == POSITIVE)
                np++;
            else
                nn++;

        // If count greater or equal than k' add to class, else remove
        if (np < kprime && nn < kprime) {
            removed++;

            // Remove
            (*n_negative)--;
            FREE((*negatives)[i]);
            (*negatives)[i] = (*negatives)[*n_negative];
            REALLOC(*negatives, char *, *n_negative, "GeneralizedEditingFilter: *negatives");

            // Revisit
            i--;
        } else if (np >= kprime) {
            reassigned++;

            // Reassign
            (*n_negative)--;
            (*n_positive)++;
            REALLOC(*positives, char *, *n_positive, "GeneralizedEditingFilter: *positives")
            (*positives)[*n_positive-1] = (*negatives)[i];
            (*negatives)[i] = (*negatives)[*n_negative];
            REALLOC(*negatives, char *, *n_negative, "GeneralizedEditingFilter: *negative");

            // Revisit
            i--;
        }
    }

    fprintf(parm->out_file, "Noise filtering NEGATIVE removed instances: %d\n", removed);
    fprintf(parm->out_file, "Noise filtering NEGATIVE reassigned instances: %d\n", reassigned);
}

void GetKNeighbors(char *seq, char **poss, char **negs, int n_pos, int n_neg, int length, int k, t_SequenceClass *neighbors)
{
    scalar distance[k], dmin, dist;

    // Select k first as k initial neighbors
    dmin = 2.0;
    for (int i = 0; i < k; i++)
        distance[i] = 2.0;

    // Get k-neighbors neighbors using Hamming distance
    for (int i = 0; i < n_pos; i++) {
        dist = HammingDistance(seq, poss[i], length);

        // Test if is one of the neighbors
        if (dist < dmin) {
            // Look for its place
            int j = 0;
            while (dist > distance[j])
                j++;

            // Sets in j-th place
            for (int l = k-1; l > j; l--) {
                neighbors[l] = neighbors[l-1];
                distance[l] = distance[l-1];
            }
            neighbors[j] = POSITIVE;
            distance[j] = dist;
            dmin = distance[k-1];
        }
    }

    for (int i = 0; i < n_neg; i++) {
        dist = HammingDistance(seq, negs[i], length);

        // Test if is one of the neighbors
        if (dist < dmin) {
            // Look for its place
            int j = 0;
            while (dist > distance[j])
                j++;

            // Sets in j-th place
            for (int l = k-1; l > j; l--) {
                neighbors[l] = neighbors[l-1];
                distance[l] = distance[l-1];
            }
            neighbors[j] = NEGATIVE;
            distance[j] = dist;
            dmin = distance[k-1];
        }
    }
}
