/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "type.h"
#include "macros.h"

void GetPositionWeightMatrix(const t_Parameters *parm, t_Recognizer *R, char **positives, int n_positives, char **negatives, int n_negatives, int length)
{
    extern int dna_index[];

    scalar bf[18];

    // Allocate memory
    for (int i = 0; i < 18; i++) {
        MALLOC(R->M[i], scalar, length, "GetPositionWeightMatrix: R->M[i]");
        bf[i] = 1.0;
    }

    // Base frequency in negative sequences
    for (int i = 0; i < n_negatives; i++)
        for (int j = 0; j < length; j++)
            bf[dna_index[(int) negatives[i][j]]]++;

    // Base frequency in positive sequences, and frequency in each position
    for (int j = 0; j < 18 ; j++)
        for (int i = 0; i < length; i++)
            R->M[j][i] = 1.0;
    for (int i = 0; i < n_positives; i++)
        for (int j = 0; j < length; j++) {
            bf[dna_index[(int) positives[i][j]]]++;
            R->M[dna_index[(int) positives[i][j]]][j]++;
        }

    // Frequency, divide by number of positives + pseudo-count
    // Also base frequency
    for (int j = 0; j < 18; j++) {
        bf[j] /= (n_negatives + n_positives) * length + 18;
        for (int i = 0; i < length; i++)
            R->M[j][i] /= n_positives + 18;
    }

    fprintf(parm->out_file, "Base frequency: A: %4.2lf%% T: %4.2lf%% G: %4.2lf%% C: %4.2lf%% Other: %4.2lf%%\n", bf[dna_index['A']], bf[dna_index['T']], bf[dna_index['G']],
            bf[dna_index['C']], bf[dna_index[0]]);

    // M(b,i) = log2 (f(b,i)/p(b))
    for (int j = 0; j < 18; j++)
        for (int i = 0; i < length; i++)
            R->M[j][i] = log2(R->M[j][i]/bf[j]);

    // Get threshold to obtain desired sensitivity = TP/(TP+FN)
    // Get all scores and ...
    scalar *sc;
    ALLOCA(sc, scalar, n_positives, "GetPositionWeightMatrix: sc");
    for (int i = 0; i < n_positives; i++)
        sc[i] = ScorePositionWeightMatrix(R, positives[i], length);

    // ...sort them
    qsort(sc, n_positives, sizeof(scalar), CompareDouble);

    // Get threshold for desired sensitivity
    if (parm->sensitivity > 0.0)
        R->threshold = sc[(int)(n_positives * (1.0 - parm->sensitivity)) ];
    else    // Use 80% as default
        R->threshold = sc[(int)(n_positives * 0.2) ];
}

scalar ScorePositionWeightMatrix(t_Recognizer *R, char *sequence, int length)
{
    extern int dna_index[];

    scalar score = 0.0;

    for (int i = 0; i < length; i++)
        score += R->M[dna_index[(int) sequence[i]]][i];

    return score;
}
