/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "type.h"

t_SequenceClass ClassifySequence(const t_Parameters *parm, t_Recognizer *rec, char *seq, int s_length, int tis_location, scalar *sc_seq, struct spectrum_kernel_info *spectrum)
{
    t_SequenceClass result = NEGATIVE;

    // Test method or stacking
    if (parm->stacking)
        result = StackingOutput(parm, rec, seq, s_length, tis_location) > rec->threshold ? POSITIVE : NEGATIVE;
    else
        switch (parm->r_method) {
            case STOP:
                result = GetSequenceScore(seq, s_length, rec->scores, rec->n_positions, tis_location, parm->reverse_stop_codon) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case C45:
                result = C45Output(rec->tree, seq, s_length) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case SVM:
                result = SVMOutput(rec->lsvm, seq, s_length, sc_seq, parm->use_salzberg_method, spectrum) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case kNN:
                result = ClassifySequencekNN(rec, seq, s_length, parm->knn_weight_by_distance) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case EasyEnsembleEnsemble:
                result = MultiEnsembleOutput(rec->easy, seq, s_length, parm->base_learner, parm->use_salzberg_method) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case BalanceCascadeEnsemble:
                result = MultiEnsembleOutput(rec->cascade, seq, s_length, parm->base_learner, parm->use_salzberg_method) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case PWM:
                result = ScorePositionWeightMatrix(rec, seq, s_length) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case NeuralNet:
                result = NeuralNetworkOutput(rec, seq, s_length) > rec->threshold ? POSITIVE : NEGATIVE;
                break;

            case LearningFromBites:
                result = EnsembleOutput(rec->ensemble, seq, s_length, parm->base_learner, parm->use_salzberg_method) > rec->ensemble->threshold ? POSITIVE : NEGATIVE;
                break;

            default:
                OnErrorExit(1, "Bad recognition method");
        }

    return result;
}

void EvaluateRecognizer(const t_Parameters *parm, t_Recognizer *rec, t_RecInfo *info, char **pos, int n_pos, int length, char **neg, int n_neg, int tis_location, scalar *positive_outputs,
                        scalar *negative_outputs, scalar **sc_positives, scalar **sc_negatives, struct spectrum_kernel_info *spectrum)
{
    t_SequenceClass r;
    scalar ok = 0.0;

    // Evaluate
    info->TP = info->TN = info->FP = info->FN = 0.0;

    // Positive sequences
    // If output yet obtained used them
    if (positive_outputs != NULL) {
        for (int i = 0; i < n_pos; i++) {
            r  = positive_outputs[i] > rec->threshold ? POSITIVE : NEGATIVE;

            if (r == POSITIVE) {
                ok++;
                info->TP++;
            } else
                info->FN++;
        }
    } else  {
        // Otherwise get them
        for (int i = 0; i < n_pos; i++) {
            r = ClassifySequence(parm, rec, pos[i], length, tis_location, sc_positives ? sc_positives[i] : NULL, spectrum);

            if (r == POSITIVE) {
                ok++;
                info->TP++;
            } else
                info->FN++;
        }
    }

    // Negative sequences
    // If output yet obtained used them
    if (negative_outputs != NULL) {
        for (int i = 0; i < n_neg; i++) {
            r  = negative_outputs[i] > rec->threshold ? POSITIVE : NEGATIVE;

            if (r == NEGATIVE) {
                ok++;
                info->TN++;
            } else
                info->FP++;
        }
    } else  {
        // Otherwise get them
        for (int i = 0; i < n_neg; i++) {
            r = ClassifySequence(parm, rec, neg[i], length, tis_location, sc_negatives ? sc_negatives[i] : NULL, spectrum);

            if (r == NEGATIVE) {
                ok++;
                info->TN++;
            } else
                info->FP++;
        }
    }

    info->sensitivity = info->TP / (info->TP + info->FN);
    info->specificity = info->TN / (info->TN + info->FP);
    info->bio_specificity = (info->TP + info->FP) > 0.0 ? info->TP / (info->TP + info->FP) : 0.0;
    info->precision = (info->TP + info->FP > 0) ? info->TP / (info->TP + info->FP) : 0.0;

    // F-measure
    info->F = (info->sensitivity + info->precision > 0.) ? (2.*info->sensitivity*info->precision)/(info->sensitivity + info->precision) : 0.0;

    // G-measure
    info->G = pow(info->sensitivity * info->specificity, 0.5);

    info->error = 1.0 - ok/ (n_pos + n_neg);
}

scalar ClassifierOutput(const t_Parameters *parm, t_Recognizer *rec, char *seq, int s_length, int tis_location, scalar *sc_seq, struct spectrum_kernel_info *spectrum)
{
    // Test method or stacking
    if (parm->stacking)
        return StackingOutput(parm, rec, seq, s_length, tis_location);
    else
        switch (parm->r_method) {
            case STOP:
                return GetSequenceScore(seq, s_length, rec->scores, rec->n_positions, tis_location, parm->reverse_stop_codon);
                break;

            case C45:
                return C45Output(rec->tree, seq, s_length);
                break;

            case SVM:
                return SVMOutput(rec->lsvm, seq, s_length, sc_seq, parm->use_salzberg_method, spectrum);
                break;

            case kNN:
                return ClassifySequencekNN(rec, seq, s_length, parm->knn_weight_by_distance);
                break;

            case EasyEnsembleEnsemble:
                return MultiEnsembleOutput(rec->easy, seq, s_length, parm->base_learner, parm->use_salzberg_method);
                break;

            case BalanceCascadeEnsemble:
                return MultiEnsembleOutput(rec->cascade, seq, s_length, parm->base_learner, parm->use_salzberg_method);
                break;

            case PWM:
                return ScorePositionWeightMatrix(rec, seq, s_length);
                break;

            case NeuralNet:
                return NeuralNetworkOutput(rec, seq, s_length);
                break;

            case LearningFromBites:
                return EnsembleOutput(rec->ensemble, seq, s_length, parm->base_learner, parm->use_salzberg_method);
                break;

            default:
                OnErrorExit(1, "Bad recognition method");
        }

    return 0.0;
}
