/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"

void WriteROCResults(const t_Parameters *parm, t_Recognizer *R, t_RecognitionMethod method, char **positives, int n_positives, char **negatives, int n_negatives, int length, char **train_positives, 
                     int train_n_positives, scalar **sc_positives, scalar **sc_negatives, FILE *out_file)
{
    // Testing error
    scalar *sc_pos, *sc_neg, *train_sc_pos;        // Output for testing positive and negative instances
    MALLOC(sc_pos, scalar, n_positives, "WriteROCResults: sc_pos");
    MALLOC(sc_neg, scalar, n_negatives, "WriteROCResults: sc_neg");
    MALLOC(train_sc_pos, scalar, train_n_positives, "WriteROCResults: train_sc_pos");
    t_RecInfo info;

    // To set the threshold we need training positive outputs
    for (int i = 0; i < train_n_positives; i++)
        train_sc_pos[i] = ClassifierOutput(parm, R, train_positives ? train_positives[i] : NULL, length, 0, sc_positives ? sc_positives[i] : NULL, NULL);

    // Output for all positive instances
    for (int i = 0; i < n_positives; i++)
        sc_pos[i] = ClassifierOutput(parm, R, positives ? positives[i] : NULL, length, 0, sc_positives ? sc_positives[i] : NULL, NULL);
    // Output for all negative instances
    for (int i = 0; i < n_negatives; i++)
        sc_neg[i] = ClassifierOutput(parm, R, negatives ? negatives[i] : NULL, length, 0, sc_negatives ? sc_negatives[i] : NULL, NULL);

    EvaluateRecognizer(parm, R, &info, positives, n_positives, length, negatives, n_negatives, 0, sc_pos, sc_neg, sc_positives, sc_negatives, NULL);

    fprintf(out_file, "TESTING results\nError: %lf\n", info.error);
    fprintf(out_file, "gl_sensitivity (TP rate): %lf\nSpecificity (TN rate): %lf\n", info.sensitivity, info.specificity);
    fprintf(out_file, "Bioinformatics specificity: %lf\n", info.bio_specificity);
    fprintf(out_file, "G-measure: %lf\nF-measure: %lf\nPrecision: %lf\n", info.G, info.F, info.precision);
    fprintf(out_file, "FP rate: %lf\n", info.FP/(info.FP + info.TN));
    fprintf(out_file, "==> FN rate: %lf\n", info.FN/n_positives);

    // Save results of sensitivity/specificity for testing and 10 intervals, from 0% to 100%
    fprintf(out_file, "Error\tSn\tSp\tSp(Bio)\tG\tF\tPrec\tFPrate\tFNrate\n");

    // Results for the whole interval
    // First 0%/100%
    // Sort outputs
    qsort(sc_pos, n_positives, sizeof(scalar), CompareDouble);
    qsort(sc_neg, n_negatives, sizeof(scalar), CompareDouble);
    qsort(train_sc_pos, train_n_positives, sizeof(scalar), CompareDouble);

    fprintf(out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar) n_positives)/(n_positives+n_negatives), 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0);
    for (scalar sn = 0.1; sn < 0.91; sn += 0.1) {
        scalar specificity, sensitivity, threshold, TN, TP, FN, FP, precision, bio_specificity;

        // Get threshold for specified sensitivity
        threshold = train_sc_pos[(int)(train_n_positives * (1.0 - sn)) ];

        // Obtain sensitivity
        // Positive sequences
        TP = 0.0;
        for (int i = n_positives - 1; sc_pos[i] >= threshold && i >= 0; i--)
            TP++;
        FN = n_positives - TP;

        // Obtain specificity
        // Negative sequences
        TN = 0.0;
        for (int i = 0; sc_neg[i] < threshold && i < n_negatives; i++)
            TN++;
        FP = n_negatives - TN;

        sensitivity = TP/(TP + FN);
        bio_specificity = (TP + FP) > 0.0 ? TP/(TP + FP) : 0.0;
        specificity = TN/(TN + FP);
        precision = (TP + FP > 0) ? TP/(TP + FP) : 0.0;

        fprintf(out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", (FP + FN)/(n_positives + n_negatives), sensitivity, specificity, bio_specificity,
                pow(sensitivity * specificity, 0.5), ((sensitivity + precision > 0.0) ? (2.0*sensitivity*precision)/(sensitivity + precision) : 0.0), precision, FP/n_negatives,
                FN/n_positives);
    }

    fprintf(out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar) n_negatives)/(n_positives+n_negatives), 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

    FREE(sc_neg);
    FREE(sc_pos);
    FREE(train_sc_pos);
}

void WriteStackingComponentsResults(const t_Parameters *parm, t_Recognizer *R, char **positives, int n_positives, char **negatives, int n_negatives, int length, char **train_positives,
                                    int train_n_positives, int stacking_methods, FILE *out_file)
{
    if (stacking_methods & stSTOP) {
        fprintf(out_file, "STOP codon stacking component results\n");
        WriteROCResults(parm, R, STOP, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stC45) {
        fprintf(out_file, "C4.5 stacking component results\n");
        WriteROCResults(parm, R, C45, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stSVM) {
        fprintf(out_file, "SVM stacking component results\n");
        WriteROCResults(parm, R, SVM, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stSVMSTR) {
        struct svm_model *backup = R->lsvm;
        R->lsvm = R->lsvm_string;
        fprintf(out_file, "SVM with string kernel stacking component results\n");
        WriteROCResults(parm, R, SVM, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
        R->lsvm = backup;
    }

    if (stacking_methods & stKNN) {
        fprintf(out_file, "k-NN stacking component results\n");
        WriteROCResults(parm, R, kNN, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stEasy) {
        fprintf(out_file, "EasyEnsembleEnsemble stacking component results\n");
        WriteROCResults(parm, R, EasyEnsembleEnsemble, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stCascade) {
        fprintf(out_file, "BalanceCascadeEnsemble stacking component results\n");
        WriteROCResults(parm, R, BalanceCascadeEnsemble, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stPWM) {
        fprintf(out_file, "PWM stacking component results\n");
        WriteROCResults(parm, R, PWM, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }

    if (stacking_methods & stNN) {
        fprintf(out_file, "NeuralNet stacking component results\n");
        WriteROCResults(parm, R, NeuralNet, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, NULL, NULL, out_file);
    }
}
