/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include "type.h"
#include "macros.h"
#include "common.h"

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal;
int MaxAtt;
ItemNo MaxItem;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY = 0, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float CF = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

int main(int argc, char *argv[])
{
    extern char letters[];

    time_t init_time, end_time;
    int n_positives, n_negatives, length, p_max_length = -1, n_max_length = -1;
    char **positives, **negatives, file_name[500], **train_positives = NULL;
    t_Recognizer R;
    scalar **vdm;
    t_Parameters parm;
    struct spectrum_kernel_info *spectrum_train = NULL, *spectrum_test = NULL;

    // Test enough parameters.
    if (argc >= 7) {
        // File for output messages.
        if ((parm.out_file = fopen(argv[6], "w")) == NULL)
            OnErrorExit(1, "main: Cannot open output file\n");

        LoadGlobalParameters(&parm, argv[5]);
    } else {
        fprintf(stdout, "site-recognizer train-positive train-negative {test-positive|NIL} {test-negative|NIL} parameters output\n");
        fprintf(stdout, "Options: -a additional-positive-instances-file -s outputs -s model -r repeated-seqs-file\n");
        exit(1);
    }

    // Use predefined seed or get a new one
    if (parm.seed != 0)
        // Fixed seed to make the evolution reproducible.
        fprintf(parm.out_file, "Random seed: %d (fixed)\n", parm.seed);
    else {
        parm.seed = - time(NULL);
        fprintf(parm.out_file, "Random seed: %d\n", parm.seed);
    }

    srandom(parm.seed);

    // Read training data
    // Positive sequences
    char *home = getenv("HOME");
    sprintf(file_name, "%s", argv[1]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s", home, parm.datad, argv[1]);
    ReadFastaFile(file_name, &positives, &n_positives, &length, parm.sampling_ratio, false, 0, false, NULL, 0, parm.allow_different_lengths, parm.offset, parm.max_length, parm.sampling_ratio,
                  parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL, false);
    fprintf(parm.out_file, "Positive sequences: %d\n", n_positives);
    printf("Read positive sequences training file: \"%s\"\n", argv[1]);

    if (parm.allow_different_lengths)
        p_max_length = length;

    // Optional arguments
    bool save_outputs = false, save_model = false;
    int opt, shift = 0;
    while ((opt = getopt(argc, argv, ":a:s:r:")) != EOF) {
        // Check options
        switch (opt) {
                char **additional;
                int n_additional;

            case 'a':
                if (parm.allow_different_lengths)
                    OnErrorExit(1, "Additional files not yet allowed with different lengths");

                // Additional positive instances training file
                // Check if file name given
                if (optarg == NULL)
                    OnErrorExit(1, "No additional data file for positive instances given");

                // Read FASTA data
                sprintf(file_name, "%s", optarg);
                if (access(file_name, R_OK))
                    sprintf(file_name, "%s/%s/%s", home, parm.datad, optarg);
                ReadFastaFile(file_name, &additional, &n_additional, &length, parm.sampling_ratio, false, 0, false, NULL, 0, parm.allow_different_lengths, parm.offset, parm.max_length,
                              parm.sampling_ratio, parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL, false);
                fprintf(parm.out_file, "Positive sequences added: %d\n", n_additional);
                printf("Read positive sequences from additional training file: \"%s\"\n", optarg);

                // Reallocate memory
                REALLOC(positives, char *, n_positives+n_additional, "main: positives");
                for (int i = n_positives; i < n_positives+n_additional; i++)
                    positives[i] = additional[i-n_positives];

                FREE(additional);
                n_positives += n_additional;

                // Shift remaining arguments
                shift += 2;
                break;

            // Use this file to remove repeated entries but not for training
            case 'r':
                if (parm.allow_different_lengths)
                    OnErrorExit(1, "Additional files not yet allowed with different lengths");

                // Additional positive instances training file
                // Check if file name given
                if (optarg == NULL)
                    OnErrorExit(1, "No additional data file for removing repeated positive instances given");

                // Read FASTA data
                sprintf(file_name, "%s", optarg);
                if (access(file_name, R_OK))
                    sprintf(file_name, "%s/%s/%s", home, parm.datad, optarg);
                ReadFastaFile(file_name, &additional, &n_additional, &length, parm.sampling_ratio, false, 0, false, NULL, 0, parm.allow_different_lengths, parm.offset, parm.max_length,
                              parm.sampling_ratio, parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL, false);
                printf("Read positive sequences from additional file for removing repeated entries: \"%s\"\n", optarg);

                // Remove repeated entries
                int removed = 0;
                for (int i = 0; i < n_positives; i++) {
                    // If repeated exchange with last one and decrease number of positive sequences
                    for (int j = 0; j < n_additional; j++)
                        if (!strcmp(positives[i], additional[j])) {
                            n_positives--;

                            FREE(positives[i]);
                            positives[i] = positives[n_positives];
                            removed++;
                            i--;    // Check again instance j

                            break;
                        }
                }

                printf("Removed %d repeated instances\n", removed);

                // Free memory
                for (int i = 0; i < n_additional; i++)
                    FREE(additional[i]);
                FREE(additional);

                // Shift remaining arguments
                shift += 2;
                break;

            case 's':
                // Save information
                // What to save
                if (optarg == NULL)
                    OnErrorExit(1, "What should I save?");

                if (!strncasecmp("outputs", optarg, 7)) {
                    save_outputs = true;
                    printf("Outputs for TESTING data will be saved after training\n");
                } else if (!strncasecmp("model", optarg, 5)) {
                    save_model = true;
                    printf("Model will be saved after training\n");
                } else
                    OnErrorExit(1, "Bad argument for save option");

                // Shift remaining arguments
                shift += 2;
                break;

            case ':':
                OnErrorExit(1, "Option needs an argument");
                break;

            case '?':
                OnErrorExit(1, "Unknown option");
        }
    }

    // Shifts arguments
    argv += shift;

    // Only useful for SVM just now
    if (save_model && parm.r_method != SVM && parm.r_method != kNN && parm.r_method != STOP && parm.r_method != PWM && parm.r_method != C45)
        OnErrorExit(1, "Not yet implemented");

    // Remove repeated entries
    int removed = 0;
    for (int i = 0; i < n_positives; i++) {
        // If repeated exchange with last one and decrease number of positive sequences
        for (int j = i+1; j < n_positives; j++)
            if (!strcmp(positives[i], positives[j])) {
                n_positives--;

                FREE(positives[j]);
                positives[j] = positives[n_positives];
                removed++;
                j--;    // Check again instance j
            }
    }

    printf("Removed %d repeated instances\n", removed);
    printf("Total positive instances: %d\n", n_positives);

    // Limit positive instances
    if (parm.max_positive_instances && parm.max_positive_instances < n_positives) {
        RandomizeSequences(positives, n_positives);
        n_positives = parm.max_positive_instances;
        REALLOC(positives, char *, n_positives, "main: positives");
        printf("Positive instances limited to: %d\n", n_positives);
    }

    // Negative sequences
    // File with a selection list?
    int *selection_list = NULL, n_selected = 0;
    sprintf(file_name, "%s.selection", argv[2]);
    if (!access(file_name, R_OK)) {
        FILE *selection_file;
        char line[100];
        if ((selection_file = fopen(file_name, "r")) == NULL)
            OnErrorExit(1, "Cannot open selection list file");

        // Count selected
        if (fgets(line, 100, selection_file) == NULL)
            OnErrorExit(1, "Error reading selection file");
        while (!feof(selection_file)) {
            n_selected++;
            if (fgets(line, 100, selection_file) == NULL)
                OnErrorExit(1, "Error reading selection file");
        }

        // Read selection
        MALLOC(selection_list, int, n_selected, "main: selection_list");
        rewind(selection_file);
        n_selected = 0;
        if (!fscanf(selection_file, "%d", &selection_list[n_selected]))
            OnErrorExit(1, "Error reading selection file");
        while (!feof(selection_file)) {
            n_selected++;
            if (!fscanf(selection_file, "%d", &selection_list[n_selected]) && !feof(selection_file))
                OnErrorExit(1, "Error reading selection file");
        }
        fclose(selection_file);
    }

    sprintf(file_name, "%s", argv[2]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s", home, parm.datad, argv[2]);
    ReadFastaFile(file_name, &negatives, &n_negatives, &length, parm.sampling_ratio, parm.imbalance_method == iUndersampling, n_positives, false, selection_list, n_selected,
                  parm.allow_different_lengths, parm.offset, parm.max_length, parm.sampling_ratio, parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL,
                  false);
    fprintf(parm.out_file, "Negative sequences: %d\n", n_negatives);
    printf("Read negative sequences training file: \"%s\"\n", argv[2]);

    if (parm.allow_different_lengths) {
        n_max_length = length;

        length = MAX(p_max_length, n_max_length);
    }

    // Salzberg method?
    scalar **scalar_positives = NULL, **scalar_negatives = NULL, **freqs = NULL;
    if (parm.use_salzberg_method)
        ConvertToDinucleotideFrequency(&positives, n_positives, &negatives, n_negatives, &scalar_positives, &scalar_negatives, &length, &freqs);

    // Protein sequence?
    if (parm.use_protein_sequence) {
        ConvertToProtein(positives, n_positives, length);
        length = ConvertToProtein(negatives, n_negatives, length);
    }

    // Store length
    R.length = length;

    // Mask inputs if apply
    int *mask = NULL;
    if (strncmp(parm.mask_file, "nil", 3) && !parm.use_salzberg_method) {
        // Read mask file, testing there is a value for each base
        FILE *maskf;
        char line[100];

        MALLOC(mask, int, length, "main: mask");

        if ((maskf = fopen(parm.mask_file, "r")) == NULL)
            OnErrorExit(1, "Cannot open mask file");

        int base = 0, c = 0;
        if (fgets(line, 100, maskf) == NULL)
            OnErrorExit(1, "Error reading mask file");
        while (!feof(maskf) && base < length) {
            mask[base] = atoi(line);

            // Count masked variables
            if (!mask[base])
                c++;

            base++;
            if (fgets(line, 100, maskf) == NULL && !feof(maskf))
                OnErrorExit(1, "Error reading mask file");
        }
        fclose(maskf);

        if (base != length)
            OnErrorExit(1, "Unexpected finish of mask file");

        if (c == length)
            OnErrorExit(1, "All bases masked out");

        // Mask
        for (int i = 0; i < n_positives; i++)
            for (int j = 0; j < length; j++)
                if (!mask[j])
                    positives[i][j] = 'N';

        for (int i = 0; i < n_negatives; i++)
            for (int j = 0; j < length; j++)
                if (!mask[j])
                    negatives[i][j] = 'N';
    }

    // For measuring expended time
    init_time = time(NULL);

    // Noise filters
    switch (parm.noise_filter) {
        case nfGEditing:
            GeneralizedEditingFilter((const t_Parameters *) &parm, &positives, &negatives, &n_positives, &n_negatives, parm.ge_k, parm.ge_kprime, length);
            break;

        case nfNone:
            break;
    }

    // Imbalance method?
    int all_negatives = n_negatives;
    switch (parm.imbalance_method) {
        case iNone:
            break;

        case iUndersampling:
            // Do nothing, under-sampling already done in reading negative file
            break;

        case iSmote:
            MALLOC(vdm, scalar *, LETTERS, "main: vdm");
            for (int i = 0; i < LETTERS; i++)
                MALLOC(vdm[i], scalar, LETTERS, "main: vdm[i]");
            ConstructValueDifferencesMatrix(vdm, positives, n_positives, negatives, n_negatives, length);
            SMOTEOverSampling(&positives, &n_positives, parm.smote_n, parm.smote_k, length, vdm);
            UndersampleData(&negatives, &n_negatives, n_positives, !parm.eval_training && !save_outputs);
            break;

        default:
            OnErrorExit(1, "Bad imbalance class method");
            break;
    }

    // Feature selection?
    t_Feat *f_info, *base_info = NULL;
    scalar **data;
    int *labels, instances, features, S[length], n_features;
    char aux[length + 1];

    switch (parm.fsalgorithm) {
        case fsRFE:
            ConvertToScalarData(positives, n_positives, negatives, n_negatives, length, &data, &labels, &instances, &features, parm.kernel);

            // SVM-RFE algorithm
            MALLOC(f_info, t_Feat, features, "main: f_info");

            // Get parameters
            if (parm.cv_parameters)
                CrossValidatedParameters(&parm, positives, n_positives, length, negatives, n_negatives);

            RFE((const t_Parameters *) &parm, f_info, features, data, instances, labels);

            //Sort features from worst to best ranking
            qsort(f_info, features, sizeof(t_Feat), CompareRanking);

            // Ranking 0 is worst
            fprintf(parm.out_file, "Feature selection info\nRanking\tFeature\tWeight\n");
            for (int i = 0; i < features; i++)
                fprintf(parm.out_file, "%d\t%d\t%lf\n",  i+1, f_info[i].index + 1, f_info[i].weight);

            // If string kernel, the same features are used
            if (parm.kernel != WD && parm.kernel != WDS && parm.kernel != LI) {
                // Average ranking for the 4 features associated with each position
                // Sequence info
                MALLOC(base_info, t_Feat, length, "main: base_info");
                for (int i = 0; i < length; i++) {
                    base_info[i].index = i;
                    base_info[i].w = 0.0;
                }

                int l = features/length;
                for (int i = 0; i < features; i++) {
                    base_info[f_info[i].index/l].w += i/l;
                    base_info[f_info[i].index/l].weight += f_info[i].weight/l;
                }
                qsort(base_info, length, sizeof(t_Feat), CompareW);

                FREE(f_info);

                // Sequence position info
                fprintf(parm.out_file, "Sequence position info\nRanking\tPosition\tAverage ranking\tAverage weight\n");
                for (int i = 0; i < length; i++)
                    fprintf(parm.out_file, "%d\t%d\t\t%lf\t%lf\n", i+1, base_info[i].index+1, base_info[i].w, base_info[i].weight);
            } else
                base_info = f_info;

            // Free data
            FREE(labels);
            for (int i = 0; i < instances; i++)
                FREE(data[i]);
            FREE(data);

            // Exit?
            if (parm.exit_after_fs) {
                end_time = time(NULL) - init_time;
                fprintf(parm.out_file, "FEATURE SELECTION time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);
                fclose(parm.out_file);
                exit(1);
            }

            // Check whether to cross-validate percentage
            if (EQUAL(parm.fs_percent_to_keep, 0.0)) {
                CrossValidateFeatureSelection(&parm, positives, n_positives, length, negatives, n_negatives, base_info, 0.1);
                fprintf(parm.out_file, "Percentage to keep obtained using cross-validation: %lf\n", parm.fs_percent_to_keep);
            }

            n_features = length * parm.fs_percent_to_keep;

            // Positive sequences
            length = n_features;
            for (int i = 0; i < n_positives; i++) {
                for (int j = 0; j < n_features; j++)
                    aux[j] = positives[i][base_info[j].index];

                aux[n_features] = '\0';
                REALLOC(positives[i], char, n_features + 1, "main: positives[i]");
                strcpy(positives[i], aux);
            }

            // Negative sequences
            for (int i = 0; i < n_negatives; i++) {
                for (int j = 0; j < n_features; j++)
                    aux[j] = negatives[i][base_info[j].index];

                aux[n_features] = '\0';
                REALLOC(negatives[i], char, n_features + 1, "main: negatives[i]");
                strcpy(negatives[i], aux);
            }

            fprintf(parm.out_file, "Selected %d bases (features)\n", n_features);
            break;

        case fsGA:
            // Get parameters
            if (parm.cv_parameters)
                CrossValidatedParameters(&parm, positives, n_positives, length, negatives, n_negatives);

            GAFeatureSelection((const t_Parameters *) &parm, S, positives, n_positives, negatives, n_negatives, length, parm.gafs_popsize, parm.gafs_generations, parm.gafs_set_probability,
                               parm.gafs_mutation, parm.gafs_bit_mutation, NULL);

            // Print results
            fprintf(parm.out_file, "GA feature selection\nFeature\tselected\n");
            for (int i = 0; i < length; i++)
                fprintf(parm.out_file, "%d\t%d\n", i+1, S[i]);

            // Exit?
            if (parm.exit_after_fs) {
                end_time = time(NULL) - init_time;
                fprintf(parm.out_file, "FEATURE SELECTION time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);
                fclose(parm.out_file);
                exit(1);
            }

            // Select features
            for (int i = 0; i < n_positives; i++)
                for (int j = 0; j < length; j++) {
                    int index = 0;
                    if (S[j]) {
                        positives[i][index] = positives[i][j];
                        index++;
                    }
                }

            for (int i = 0; i < n_negatives; i++)
                for (int j = 0; j < length; j++) {
                    int index = 0;
                    if (S[j]) {
                        negatives[i][index] = negatives[i][j];
                        index++;
                    }
                }

            // New length
            int new_length = 0;
            for (int j = 0; j < length; j++)
                if (S[j])
                    new_length++;

            length = new_length;
            break;

        case fsNONE:
        default:
            // Do nothing
            break;
    }

    // Training data not released yet
    bool training_data_released = false;

    if (parm.stacking || parm.r_method == C45 || ((parm.r_method == EasyEnsembleEnsemble || parm.r_method == BalanceCascadeEnsemble || parm.r_method == LearningFromBites) && parm.base_learner
            == C45)) {
        extern String FileName;

        // Create dummy names file
        FILE *fp = fopen("dummy.names", "w");
        fprintf(fp, "0, 1.    | classes\n");
        for (int i = 0; i < length; i++)
            fprintf(fp, "x%d: A,T,U,G,C,R,Y,K,M,S,W,B,D,H,V,N,X,-\n", i);
        fclose(fp);

        //  Initialize  values for c4.5
        ALLOCA(FileName, char, 100, "main: FileName");
        sprintf(FileName, "dummy");
        GetNames();

        remove("dummy.names");
    }

    // Recognition method or stacking?
    if (parm.stacking)
        training_data_released = TrainStackerAndItsClassifiers(&parm, &R, positives, n_positives, negatives, n_negatives, length);
    else
        switch (parm.r_method) {
            case STOP:
                R.length = length;
                StopCodonMethod(positives, n_positives, length, &R.scores, &R.n_positions, &R.threshold, parm.reverse_stop_codon, parm.sensitivity);
                break;

            case C45:
                R.tree = TrainC45Classifier((const t_Parameters *) &parm, positives, n_positives, length, negatives, n_negatives, false, &(R.threshold));

                // Training data released yet?
                if (!parm.eval_training)
                    training_data_released = true;

                // Tree size
                fprintf(parm.out_file, "Tree nodes: %d\n", TreeSize(R.tree));
                break;

            case SVM:
                if (parm.kernel == SPECTRUM || parm.kernel == MISMATCH || parm.kernel == WILDCARD) {
                    MALLOC(spectrum_train, struct spectrum_kernel_info, n_positives+n_negatives, "main: spectrum_train");
                    for (int i = 0; i < n_positives+n_negatives; i++) {
                        spectrum_train[i].n_k_mers = -1;
                        spectrum_train[i].list = NULL;
                        spectrum_train[i].count = NULL;
                        spectrum_train[i].k_xx = -1.0;
                    }
                }

                // Validate parameters?
                if (parm.cv_parameters)
                    CrossValidatedParameters(&parm, positives, n_positives, length, negatives, n_negatives);
                TrainSVMClassifier((const t_Parameters *) &parm, &(R.lsvm), positives, n_positives, length, negatives, n_negatives, false, &(R.threshold), scalar_positives, scalar_negatives, spectrum_train);

                // Training data released yet?
                if (!parm.eval_training)
                    training_data_released = true;

                // Support vectors
                fprintf(parm.out_file, "Support vectors: %d\n", (R.lsvm)->l);

                break;

            case kNN:
                // Validate parameters?
                R.threshold = 0.0;
                if (parm.cv_parameters)
                    CrossValidatedParameters(&parm, positives, n_positives, length, negatives, n_negatives);
                R.k = parm.k;
                R.n_poss = n_positives;
                R.pos_l = length;
                MALLOC(R.poss, char *, n_positives, "main: R.poss");
                for (int i = 0; i < n_positives; i++)
                    R.poss[i] = positives[i];

                R.n_negs = n_negatives;
                R.neg_l = length;
                MALLOC(R.negs, char *, n_negatives, "main: R.negs");
                for (int i = 0; i < n_negatives; i++)
                    R.negs[i] = negatives[i];

                // Do not release training data
                training_data_released = true;

                // Set threshold if needed
                if (parm.sensitivity > 0.0)
                    SetkNNThreshold((const t_Parameters *) &parm, &R, parm.out_file);
                break;

            case EasyEnsembleEnsemble:
                R.threshold = 0.0;
                R.easy = TrainEasyEnsemble((const t_Parameters *) &parm, positives, negatives, length, n_positives, n_negatives, parm.ensemble_t, parm.easy_s_i);

                if (parm.verbose > 1) {
                    // Print ensemble info
                    fprintf(parm.out_file, "Easy multiensemble of %d ensembles\n", R.easy->T);
                    for (int i = 0; i < R.easy->T; i++) {
                        fprintf(parm.out_file, "Ensemble %d/%d.\tThreshold: %5.2lf\n", i + 1, R.easy->T, R.easy->ensemble[i].threshold);
                        for (int j = 0; j < R.easy->ensemble[i].s; j++)
                            fprintf(parm.out_file, "Classifier %3d/%3d. Nodes: %3d\tAlpha = %5.2lf\n",j + 1, R.easy->ensemble[i].s, parm.base_learner == C45 ?
                                    TreeSize(R.easy->ensemble[i].trees[j]) :  SVMSize(R.easy->ensemble[i].lsvm[j]), R.easy->ensemble[i].alpha[j]);
                    }
                }
                if (parm.sensitivity > 0.0)
                    SetMultiEnsembleThreshold((const t_Parameters *) &parm, &R, positives, n_positives, length);
                break;

            case BalanceCascadeEnsemble:
                R.threshold = 0.0;
                R.cascade = TrainBalanceCascadeEnsemble((const t_Parameters *) &parm, positives, negatives, length, n_positives, n_negatives, parm.ensemble_t, parm.easy_s_i);

                if (parm.verbose > 1) {
                    // Print ensemble info
                    fprintf(parm.out_file, "Cascade cascade of %d ensembles\n", R.cascade->T);
                    for (int i = 0; i < R.cascade->T; i++) {
                        fprintf(parm.out_file, "Ensemble %d/%d.\tThreshold: %5.2lf\n", i + 1, R.cascade->T, R.cascade->ensemble[i].threshold);
                        for (int j = 0; j < R.cascade->ensemble[i].s; j++)
                            fprintf(parm.out_file, "Classifier %3d/%3d. Nodes: %3d\tAlpha = %5.2lf\n", j + 1, R.cascade->ensemble[i].s, parm.base_learner == C45 ?
                                    TreeSize(R.cascade->ensemble[i].trees[j]) : SVMSize(R.cascade->ensemble[i].lsvm[j]), R.cascade->ensemble[i].alpha[j]);
                    }
                }
                if (parm.sensitivity > 0.0)
                    SetMultiEnsembleThreshold((const t_Parameters *) &parm, &R, positives, n_positives, length);
                break;

            case LearningFromBites:
                R.ensemble = TrainLearningFromBites((const t_Parameters *) &parm, positives, negatives, length, n_positives, n_negatives, parm.n_bites, parm.bite_size);
                break;

            case PWM:
                GetPositionWeightMatrix((const t_Parameters *) &parm, &R, positives, n_positives, negatives, n_negatives, length);
                break;

            case NeuralNet:
                TrainNeuralNetwork((const t_Parameters *) &parm, &R, positives, n_positives, negatives, n_negatives, length, parm.nn_hidden_nodes);
                break;

            case NoOne:
            default:
                OnErrorExit(1, "Bad recognition method");
        }

    printf("Training of the model finished\n");

    // If save model write model and output
    if (save_model) {
        printf("Saving the model\n");

        // Save SVM model
        char model_file[1000];
        sprintf(model_file, "%s.%s", /*argv[1],*/ argv[6], "model");
        SaveModel((const t_Parameters *) &parm, &R, model_file);
    }

    if (false /*save_outputs*/) {
        printf("Saving outputs for training data\n");

        // Write classifier outputs
        char file_name[1000];
        FILE *save;

        // Include under-sampled negatives
        int backup = n_negatives;
        n_negatives = all_negatives;

        sprintf(file_name, "%s%s", argv[1], argv[6]);

        if (parm.stacking)
            SaveStackingOutputs((const t_Parameters *) &parm, file_name, &R, positives, n_positives, negatives, n_negatives, length, 0);
        else {
            if ((save = fopen(file_name, "w")) == NULL)
                OnErrorExit(1, "Cannot open file for saving outputs");
            for (int i = 0; i < n_positives; i++)
                fprintf(save, "%7.4lf 1\n", ClassifierOutput((const t_Parameters *) &parm, &R, positives ? positives[i] : NULL, length, 0, scalar_positives ? scalar_positives[i] : NULL,
                        &spectrum_train[i]));
            for (int i = 0; i < n_negatives; i++) {
                fprintf(save, "%7.4lf 0\n", ClassifierOutput((const t_Parameters *) &parm, &R, negatives ? negatives[i] : NULL, length, 0, scalar_negatives ? scalar_negatives[i] : NULL,
                        &spectrum_train[i+n_positives]));
            }
            fclose(save);
        }

        n_negatives = backup;
    }

    // Used time
    end_time = time(NULL) - init_time;
    fprintf(parm.out_file, "TRAINING time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);

    // Train error
    t_RecInfo info;
    if (parm.eval_training) {
        printf("Evaluating training performance\n");

        EvaluateRecognizer((const t_Parameters *) &parm, &R, &info, positives, n_positives, length, negatives, n_negatives, 0, NULL, NULL, scalar_positives, scalar_negatives, spectrum_train);

        fprintf(parm.out_file, "Recognition method: %d\n", parm.r_method);
        fprintf(parm.out_file, "TRAINING results\nError: %lf\n", info.error);
        fprintf(parm.out_file, "Sensitivity (TP rate): %lf\nSpecificity (TN rate): %lf\n", info.sensitivity, info.specificity);
        fprintf(parm.out_file, "G-measure: %lf\nF-measure: %lf\nPrecision: %lf\n", info.G, info.F, info.precision);
        fprintf(parm.out_file, "FP rate: %lf\n", info.FP/(info.FP + info.TN));
        fprintf(parm.out_file, "==> FN rate: %lf\n", info.FN/n_positives);
    }

    // ROC for training data?
    if (parm.roc_training) {
        printf("Obtaining ROC for training data\n");

        scalar *train_sc_pos, *train_sc_neg;

        MALLOC(train_sc_pos, scalar, n_positives, "main: train_sc_pos");
        MALLOC(train_sc_neg, scalar, n_negatives, "main: train_sc_neg");

        // Output for all positive instances
        for (int i = 0; i < n_positives; i++)
            train_sc_pos[i] = ClassifierOutput((const t_Parameters *) &parm, &R, positives ? positives[i] : NULL, length, 0, scalar_positives ? scalar_positives[i] : NULL, &spectrum_train[i]);
        for (int i = 0; i < n_negatives; i++)
            train_sc_neg[i] = ClassifierOutput((const t_Parameters *) &parm, &R, negatives ? negatives[i] : NULL, length, 0, scalar_negatives ? scalar_negatives[i] : NULL, &spectrum_train[i+n_positives]);

        qsort(train_sc_pos, n_positives, sizeof(scalar), CompareDouble);
        qsort(train_sc_neg, n_negatives, sizeof(scalar), CompareDouble);

        // Save results of sensitivity/specificity for testing and 10 intervals, from 0% to 100%
        fprintf(parm.out_file, "ROC data with TRAINING sequences\n");
        fprintf(parm.out_file, "Th\tError\tSn\tSp\tSp(Bio)\tG\tF\tPrec\tFPrate\tFNrate\n");

        fprintf(parm.out_file, "------\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar)n_positives)/(n_positives+n_negatives), 0.0, 1.0, 1.0, 0.0, 0.0, 0.0,
                0.0, 1.0);
        bool zeroed = false;
        for (scalar sn = 0.01; sn < 0.99; sn += 0.01) {
            scalar specificity, sensitivity, threshold, TN, TP, FN, FP, precision, bio_specificity;

            // Get threshold for specified sensitivity
            threshold = train_sc_pos[(int)(n_positives * (1.0-sn))];

            // Make sure zero is used
            if (threshold < 0.0 && !zeroed) {
                threshold = 0.0;
                zeroed = true;
                sn -= 0.01;
            }

            // Obtain sensitivity
            // Positive sequences
            TP = 0.0;
            for (int i = n_positives - 1; train_sc_pos[i] >= threshold && i >= 0; i--)
                TP++;
            FN = n_positives - TP;

            // Obtain specificity
            // Negative sequences
            TN = 0.0;
            for (int i = 0; train_sc_neg[i] < threshold && i < n_negatives; i++)
                TN++;
            FP = n_negatives - TN;

            sensitivity = TP/(TP + FN);
            bio_specificity = (TP + FP) > 0.0 ? TP/(TP + FP) : 0.0;
            specificity = TN/(TN + FP);
            precision = (TP + FP > 0) ? TP/(TP + FP) : 0.0;

            fprintf(parm.out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", threshold, (FP + FN)/(n_positives + n_negatives), sensitivity,
                    specificity, bio_specificity, pow(sensitivity * specificity, 0.5), ((sensitivity+precision > 0.0) ? (2.0*sensitivity*precision)/(sensitivity+precision) : 0.0),
                    precision, FP/n_negatives, FN/n_positives);
        }

        fprintf(parm.out_file, "------\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar)n_negatives)/(n_positives+n_negatives), 1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                1.0, 0.0);

        FREE(train_sc_neg);
        FREE(train_sc_pos);
    }

    int train_n_positives = n_positives;

    if (parm.use_salzberg_method && !training_data_released) {
        // Free training data
        for (int i = 0; i < n_positives; i++)
            FREE(scalar_positives[i]);
        FREE(scalar_positives);
        for (int i = 0; i < n_negatives; i++)
            FREE(scalar_negatives[i]);
        FREE(scalar_negatives);
    } else if (!training_data_released && !parm.stacking) {
        // Free training data
        for (int i = 0; i < n_positives; i++)
            FREE(positives[i]);
        FREE(positives);
        for (int i = 0; i < n_negatives; i++)
            FREE(negatives[i]);
        FREE(negatives);
    } else if (parm.stacking) {
        // Store positives for ROC outputs and free negatives
        train_positives = positives;
        for (int i = 0; i < n_negatives; i++)
            FREE(negatives[i]);
        FREE(negatives);
    }

    // If test file = NIL, exit
    if (!strncasecmp(argv[3], "NIL", 3))
        exit(1);

    // Read testing data
    sprintf(file_name, "%s", argv[3]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s", home, parm.datad, argv[3]);
    ReadFastaFile(file_name, &positives, &n_positives, &length, 1.0, false, 0, false, NULL, 0, parm.allow_different_lengths, parm.offset, parm.max_length, parm.sampling_ratio,
                  parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL, false);
    fprintf(parm.out_file, "Positive sequences: %d\n", n_positives);
    printf("Read positive sequences testing file: \"%s\"\n", argv[3]);

    sprintf(file_name, "%s", argv[4]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s", home, parm.datad, argv[4]);
    ReadFastaFile(file_name, &negatives, &n_negatives, &length, 1.0, false, 0, false, NULL, 0, parm.allow_different_lengths, parm.offset, parm.max_length, parm.sampling_ratio,
                  parm.allow_only_5_letters, parm.do_not_check_sequences, letters, &parm.length, NULL, false);
    fprintf(parm.out_file, "Negative sequences: %d\n", n_negatives);
    printf("Read negative sequences testing file: \"%s\"\n", argv[4]);

    if (parm.use_salzberg_method)
        ConvertToDinucleotideFrequency(&positives, n_positives, &negatives, n_negatives, &scalar_positives, &scalar_negatives, &length, &freqs);

    // Protein sequence?
    if (parm.use_protein_sequence) {
        ConvertToProtein(positives, n_positives, length);
        length = ConvertToProtein(negatives, n_negatives, length);
    }

    if (mask != NULL) {
        // Mask
        for (int i = 0; i < n_positives; i++)
            for (int j = 0; j < length; j++) {
                int index = 0;
                if (mask[j]) {
                    positives[i][index] = positives[i][j];
                    index++;
                }
            }

        for (int i = 0; i < n_negatives; i++)
            for (int j = 0; j < length; j++) {
                int index = 0;
                if (mask[j]) {
                    negatives[i][index] = negatives[i][j];
                    index++;
                }
            }

        // New length
        int new_length = 0;
        for (int j = 0; j < length; j++)
            if (mask[j])
                new_length++;

        length = new_length;
    }

    // Feature selection?
    if (parm.fsalgorithm == fsRFE) {
        // Keep desired percentage of inputs
        char aux[length + 1];
        int n_features = length * parm.fs_percent_to_keep;

        // Positive sequences
        length  = n_features;
        for (int i = 0; i < n_positives; i++) {
            for (int j = 0; j < n_features; j++)
                aux[j] = positives[i][base_info[j].index];

            aux[n_features] = '\0';
            REALLOC(positives[i], char, n_features + 1, "main: positives[i]");
            strcpy(positives[i], aux);
        }

        // Negative sequences
        for (int i = 0; i < n_negatives; i++) {
            for (int j = 0; j < n_features; j++)
                aux[j] = negatives[i][base_info[j].index];

            aux[n_features] = '\0';
            REALLOC(negatives[i], char, n_features + 1, "main: negatives[i]");
            strcpy(negatives[i], aux);
        }

        FREE(base_info);
    } else if (parm.fsalgorithm == fsGA) {
        // Select features
        for (int i = 0; i < n_positives; i++)
            for (int j = 0; j < length; j++) {
                int index = 0;
                if (S[j]) {
                    positives[i][index] = positives[i][j];
                    index++;
                }
            }

        for (int i = 0; i < n_negatives; i++)
            for (int j = 0; j < length; j++) {
                int index = 0;
                if (S[j]) {
                    negatives[i][index] = negatives[i][j];
                    index++;
                }
            }

        // New length
        int new_length = 0;
        for (int j = 0; j < length; j++)
            if (S[j])
                new_length++;

        length = new_length;
    }

    // If stacking output member results
    if (parm.stacking)
        WriteStackingComponentsResults((const t_Parameters *) &parm, &R, positives, n_positives, negatives, n_negatives, length, train_positives, train_n_positives, parm.stacking_methods, parm.out_file);

    // Testing error
    scalar *sc_pos, *sc_neg;        // Output for testing positive and negative instances
    MALLOC(sc_pos, scalar, n_positives, "main: sc_pos");
    MALLOC(sc_neg, scalar, n_negatives, "main: sc_neg");

    // For measuring expended time
    init_time = time(NULL);

    if (parm.kernel == SPECTRUM || parm.kernel == MISMATCH || parm.kernel == WILDCARD) {
        MALLOC(spectrum_test, struct spectrum_kernel_info, n_positives+n_negatives, "main: spectrum_test");
        for (int i = 0; i < n_positives+n_negatives; i++) {
            spectrum_test[i].n_k_mers = -1;
            spectrum_test[i].list = NULL;
            spectrum_test[i].count = NULL;
            spectrum_test[i].k_xx = -1.0;
        }
    }

    // Output for all positive instances
    for (int i = 0; i < n_positives; i++)
        sc_pos[i] = ClassifierOutput((const t_Parameters *) &parm, &R, positives ? positives[i] : NULL, length, 0, scalar_positives ? scalar_positives[i] : NULL, &spectrum_test[i]);
    // Output for all negative instances
    for (int i = 0; i < n_negatives; i++)
        sc_neg[i] = ClassifierOutput((const t_Parameters *) &parm, &R, negatives ? negatives[i] : NULL, length, 0, scalar_negatives ? scalar_negatives[i] : NULL, &spectrum_test[i+n_positives]);

    // Results for the whole interval
    if (save_outputs) {
        printf("Saving outputs for testing data\n");

        // Write classifier outputs
        char file_name[1000];
        FILE *save;

        sprintf(file_name, "%s+%s.%s.out", argv[3], argv[4], argv[6]);

        if (parm.stacking)
            SaveStackingOutputs((const t_Parameters *) &parm, file_name, &R, positives, n_positives, negatives, n_negatives, length, 0);
        else {
            if ((save = fopen(file_name, "w")) == NULL)
                OnErrorExit(1, "Cannot open file for saving outputs");
            for (int i = 0; i < n_positives; i++)
                fprintf(save, "%s %7.4lf 1\n", positives[i], sc_pos[i]);
            for (int i = 0; i < n_negatives; i++)
                fprintf(save, "%s %7.4lf 0\n", negatives[i], sc_neg[i]);
            fclose(save);
        }
    }

    // First 0%/100%
    // Sort outputs
    qsort(sc_pos, n_positives, sizeof(scalar), CompareDouble);
    qsort(sc_neg, n_negatives, sizeof(scalar), CompareDouble);

    printf("Evaluating testing performance\n");
    EvaluateRecognizer((const t_Parameters *) &parm, &R, &info, positives, n_positives, length, negatives, n_negatives, 0, sc_pos, sc_neg, scalar_positives, scalar_negatives, spectrum_test);

    // Used time
    end_time = time(NULL) - init_time;
    fprintf(parm.out_file, "TESTING time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);
    fprintf(parm.out_file, "TESTING results\n@Error: %lf\n", info.error);
    fprintf(parm.out_file, "@TP: %d\n@FN: %d\n@TN: %d\n@FP: %d\n", (int) info.TP, (int) info.FN, (int) info.TN, (int) info.FP);
    fprintf(parm.out_file, "@Sensitivity (TP rate): %lf\n@Specificity (TN rate): %lf\n", info.sensitivity, info.specificity);
    fprintf(parm.out_file, "@Bioinformatics specificity: %lf\n", info.bio_specificity);
    fprintf(parm.out_file, "@G-measure: %lf\n@F-measure: %lf\n@Precision: %lf\n", info.G, info.F, info.precision);
    fprintf(parm.out_file, "@FP rate: %lf\n", info.FP/(info.FP + info.TN));
    fprintf(parm.out_file, "@FN rate: %lf\n", info.FN/n_positives);
    // Reallocate sc_pos to include negative outputs
    REALLOC(sc_pos, scalar, n_positives+n_negatives, "main: sc_pos");
    memcpy(sc_pos+n_positives, sc_neg, n_negatives*sizeof(scalar));
    fprintf(parm.out_file, "@auROC: %lf\n", AUROCFromOutputs(sc_pos, n_positives, n_negatives, false, NULL, true));
    fprintf(parm.out_file, "@auPRC: %lf\n", AUPRCFromOutputs(sc_pos, n_positives, n_negatives, false, NULL, true));

    // Save results of sensitivity/specificity for testing and 10 intervals, from 0% to 100%
    fprintf(parm.out_file, "ROC data with TESTING sequences\n");
    fprintf(parm.out_file, "Th\tError\tSn\tSp\tSp(Bio)\tG\tF\tPrec\tFPrate\tFNrate\n");

    FILE *auc = NULL;
    if (parm.write_auc_data) {
        if ((auc = fopen(strcat(argv[6], "-auc"), "w")) == NULL)
            OnErrorExit(1, "Error opening AUC file");
        // Add always (0, 0)
        fprintf(auc, "0.0\t0.0\n");
    }

    printf("Evaluating testing performance for all thresholds\n");
    fprintf(parm.out_file, "------\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar)n_positives)/(n_positives+n_negatives), 0.0, 1.0, 1.0, 0.0, 0.0, 0.0,
            0.0, 1.0);
    bool zeroed = false;
    for (scalar sn = 0.01; sn < 0.99; sn += 0.01) {
        scalar specificity, sensitivity, threshold, TN, TP, FN, FP, precision, bio_specificity;

        // Get threshold for specified sensitivity using training data
        // threshold = train_sc_pos[(int) (train_n_positives * (1.0-sn))];

        // Get threshold for specified sensitivity using testing data
        threshold = sc_pos[(int)(n_positives * (1.0 - sn))];

        // Make sure zero is used
        if (threshold < 0.0 && !zeroed) {
            threshold = 0.0;
            zeroed = true;
            sn -= 0.01;
        }

        // Obtain sensitivity
        // Positive sequences
        TP = 0.0;
        for (int i = n_positives - 1; sc_pos[i] >= threshold && i >= 0; i--)
            TP++;
        FN = n_positives - TP;

        // Obtain specificity
        // Negative sequences
        TN = 0.0;
        for (int i = 0; sc_neg[i] < threshold && i < n_negatives; i++)
            TN++;
        FP = n_negatives - TN;

        sensitivity = TP/(TP + FN);
        // MODIFIED 4/November/2010
        // Use Bioinformatics specificity to avoid false large values
        bio_specificity = (TP + FP) > 0.0 ? TP/(TP + FP) : 0.0;
        specificity = TN/(TN + FP);
        precision = (TP + FP > 0) ? TP/(TP + FP) : 0.0;

        // MODIFIED 7/December/2010
        // Write AUC data file
        if (parm.write_auc_data)
            fprintf(auc, "%6.4lf\t%6.4lf\n", 1.0 - specificity, sensitivity);

        fprintf(parm.out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", threshold, (FP + FN)/(n_positives + n_negatives), sensitivity,
                specificity, bio_specificity, pow(sensitivity * specificity, 0.5), ((sensitivity + precision > 0.0) ? (2.0*sensitivity*precision)/(sensitivity + precision) : 0.0),
                precision, FP/n_negatives, FN/n_positives);
    }

    if (parm.write_auc_data) {
        // Add always (1, 1)
        fprintf(auc, "1.0\t1.0\n");
        fclose(auc);
    }

    fprintf(parm.out_file, "------\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", ((scalar)n_negatives)/(n_positives+n_negatives), 1.0, 0.0, 0.0, 0.0,
            0.0, 0.0, 1.0, 0.0);

    // If SVM same model
    if (parm.r_method == SVM) {
        char name[100];

        sprintf(name, "%s-smodel", argv[6]);
    }

    return 0;
}

// Auxiliary function for qsort: from worst to best ranking
int CompareRanking(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) y)->ranking - ((t_Feat *) x)->ranking) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort: from worst to best weight
int CompareW(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_Feat *) x)->w - ((t_Feat *) y)->w) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}
