/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "type.h"
#include "macros.h"

// Oversampling minority class following SMOTE method [Chawla, 2002]
// Create artificial instances from adjacent instances

// Each instance will be oversampled N% times

// Output: (N/100)T synthetic minority class samples are added to the sample
void SMOTEOverSampling(char ***minority, int *T, scalar N, int k, int length, scalar **vdm)
{
    int list[*T], to_smote;

    // If N < 100 randomize as not all instances will be SMOTEd
    for (int i = 0; i < *T; i++)
        list[i] = i;
    if (N < 100.0) {
        RandomizeVector(list, *T);
        to_smote = 100.0 * (*T/N);
        N = 100.0;
    } else
        to_smote = *T;

    N = (int) N/100.0;

    // Reallocate memory
    REALLOC(*minority, char *, *T + N * *T, "SMOTEOverSampling: *data");

    // Oversample minority class instances
    int n = *T;
    for (int i = 0; i < to_smote; i++) {
        Populate(list[i], *minority, *T, n, N, k, length, vdm);

        // Added N synthetic samples
        n += N;
    }

    *T = n;
}

void Populate(int sample, char **data, int T, int n, int N, int k, int length, scalar **vdm)
{
    void GetkNearestNeighbours(char *, char **, int, int, int, int *, scalar **);

    int nn[k];

    GetkNearestNeighbours(data[sample], data, T, length, k + 1, nn, vdm);

    // Remove itself
    for (int i = 0; i < k+1; i++)
        if (nn[i] == sample) {
            nn[i] = nn[k];
            break;
        }

    // Generate N synthetic samples
    for (int i = 0; i < N; i++) {
        // Create synthetic instance
        MALLOC(data [n + i], char , length, "Populate: data)[n + i]");

        // Choose randomly between its feature and a random neighbor
        for (int j = 0; j < length; j++)
            if (RANDOM(0.0, 1.0) < 0.5)
                data[n + i][j] = data[sample][j];
            else
                data[n + i][j] = data[nn[(int) RANDOM(0, k) ]][j];
    }
}

void GetkNearestNeighbours(char *instance, char **data, int n, int length, int k, int *nn, scalar **vdm)
{
    scalar ModifiedVDN(char *, char *, int, scalar **);

    scalar dist[k];
    scalar value, dmin;

    // Select k first as not known
    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nn[i] = -1;
    }

    // Get k nearest neighbours
    for (int i = 0; i < n && dmin > 0.0; i++) {
        value = ModifiedVDN(instance, data[i], length, vdm);

        // Test if is one of the nearest
        if (value < dmin) {
            // Look for its place
            int j = 0;
            while (value > dist[j])
                j++;

            // Sets in j-th place
            for (int l = k - 1; l > j; l--) {
                nn[l] = nn[l - 1];
                dist[l] = dist[l - 1];
            }
            nn[j] = i;
            dist[j] = value;
            dmin = dist[k - 1];
        }
    }
}

// Order of the matrix: ATGC x ATGC
// #define LETTERS        17
void ConstructValueDifferencesMatrix(scalar **vdm, char **pos, int n_pos, char **neg, int n_negs, int length)
{
    // Vector to convert ATGC to integer index
    extern int dna_index[];

    for (int i = 0; i < LETTERS; i++)
        vdm[i][i] = 0.0;

    // Count for all 17 values and 2 classes
    scalar gl_c[LETTERS][2];
    for (int i = 0; i < LETTERS; i++)
        gl_c[i][0] = gl_c[i][1] = 0.0;

    // Count occurrences in positive class
    for (int k = 0; k < n_pos; k++)
        for (int l = 0; l < length; l++)
            gl_c[dna_index[(int) pos[k][l]]][0] ++;

    // Count occurrences in negative class
    for (int k = 0; k < n_negs; k++)
        for (int l = 0; l < length; l++)
            gl_c[dna_index[(int) neg[k][l]]][1] ++;

    // Iterate for all possible combinations
    for (int i = 1; i < LETTERS; i++)
        for (int j = 0; j < i; j++)
            vdm[i][j] = vdm[j][i] =
                fabs(gl_c[i][0]/ (gl_c[i][0]+gl_c[i][1]) - gl_c[j][0]/ (gl_c[j][0]+gl_c[j][1])) + fabs(gl_c[i][1]/ (gl_c[i][0]+gl_c[i][1]) - gl_c[j][1]/ (gl_c[j][0]+gl_c[j][1]));
}

#define R       2.0       // R=1 Manhattan distance, R=2 Euclidean distance
scalar ModifiedVDN(char *x, char *y, int length, scalar **vdm)
{
    // Vector to convert ATGC to integer index
    extern int dna_index[];

    scalar distance = 0.0;

    for (int i = 0; i < length; i++)
        distance += vdm[dna_index[(int) x[i]]][(int) dna_index[(int) y[i]]];

    return distance;
}
