/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdio.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"

bool TrainStackerAndItsClassifiers(t_Parameters *parm, t_Recognizer *R, char **positives, int n_positives, char **negatives, int n_negatives, int length)
{
    bool training_data_released = false;
    int inputs = 0;

    // Train individuals classifiers
    if (parm->stacking_methods & stSTOP) {
        StopCodonMethod(positives, n_positives, length, &R->scores, &R->n_positions, &R->threshold, parm->reverse_stop_codon, parm->sensitivity);
        inputs++;
    }

    if (parm->stacking_methods & stC45) {
        R->tree = TrainC45Classifier(parm, positives, n_positives, length, negatives, n_negatives, false, NULL);
        inputs++;

        // Tree size
        fprintf(parm->out_file, "Tree nodes: %d\n", TreeSize(R->tree));
    }

    // Use string kernels: WD or WDS. Now: WD
    if (parm->stacking_methods & stSVMSTR) {
        parm->kernel = WD;
        inputs++;

        TrainSVMClassifier(parm, & (R->lsvm_string), positives, n_positives, length, negatives, n_negatives, false, NULL, NULL, NULL, NULL);

        // Support vectors
        fprintf(parm->out_file, "Support vectors: %d\n", (R->lsvm_string)->l);
    }

    if (parm->stacking_methods & stSVM) {
        inputs++;

        // Validate parameters, do not use string kernels
        parm->r_method = SVM;
        CrossValidatedParameters(parm, positives, n_positives, length, negatives, n_negatives);
        TrainSVMClassifier(parm, & (R->lsvm),positives, n_positives, length, negatives, n_negatives, false, NULL, NULL, NULL, NULL);

        // Support vectors
        fprintf(parm->out_file, "Support vectors: %d\n", (R->lsvm)->l);

        parm->r_method = NoOne;

    }

    if (parm->stacking_methods & stKNN) {
        inputs++;

        // Validate parameters?
        R->threshold = 0.0;
        parm->r_method = kNN;
        if (parm->cv_parameters)
            CrossValidatedParameters(parm, positives, n_positives, length, negatives, n_negatives);
        R->k = parm->k;
        R->n_poss = n_positives;
        R->pos_l = length;
        MALLOC(R->poss, char *, n_positives, "main: R->poss");
        for (int i = 0; i < n_positives; i++) {
            // MALLOC ( R->poss[i], char, R->pos_l + 1, "main: R->poss[i]" );
            // strcpy ( R->poss[i], positives[i] );
            R->poss[i] = positives[i];
        }
        R->n_negs = n_negatives;
        R->neg_l = length;
        MALLOC(R->negs, char *, n_negatives, "main: R->negs");
        for (int i = 0; i < n_negatives; i++) {
            // MALLOC ( R->negs[i], char, R->neg_l + 1, "main: R->negs[i]" );
            // strcpy ( R->negs[i], negatives[i] );
            R->negs[i] = negatives[i];
        }

        // Do not release training data
        training_data_released = true;

        parm->r_method = NoOne;
    }

    if (parm->stacking_methods & stEasy) {
        inputs++;
        parm->r_method = EasyEnsembleEnsemble;
        R->threshold = 0.0;
        R->easy = TrainEasyEnsemble(parm, positives, negatives, length, n_positives, n_negatives, parm->ensemble_t, parm->easy_s_i);

        if (parm->verbose > 1) {
            // Print ensemble info
            fprintf(parm->out_file, "Easy multiensemble of %d ensembles\n", R->easy->T);
            for (int i = 0; i < R->easy->T; i++) {
                fprintf(parm->out_file, "Ensemble %d/%d.\tThreshold: %5.2lf\n", i + 1, R->easy->T, R->easy->ensemble[i].threshold);
                for (int j = 0; j < R->easy->ensemble[i].s; j++)
                    fprintf(parm->out_file, "Classifier %3d/%3d. Nodes: %3d\tAlpha = %5.2lf\n", j + 1, R->easy->ensemble[i].s, parm->base_learner == C45 ?
                        TreeSize(R->easy->ensemble[i].trees[j]) : SVMSize(R->easy->ensemble[i].lsvm[j]), R->easy->ensemble[i].alpha[j]);
            }
        }

        parm->r_method = NoOne;
    }

    if (parm->stacking_methods & stCascade) {
        inputs++;
        parm->r_method = BalanceCascadeEnsemble;
        R->threshold = 0.0;
        R->cascade = TrainBalanceCascadeEnsemble(parm, positives, negatives, length, n_positives, n_negatives, parm->ensemble_t, parm->easy_s_i);

        if (parm->verbose > 1) {
            // Print ensemble info
            fprintf(parm->out_file, "Cascade cascade of %d ensembles\n", R->cascade->T);
            for (int i = 0; i < R->cascade->T; i++) {
                fprintf(parm->out_file, "Ensemble %d/%d.\tThreshold: %5.2lf\n", i + 1, R->cascade->T, R->cascade->ensemble[i].threshold);
                for (int j = 0; j < R->cascade->ensemble[i].s; j++)
                    fprintf(parm->out_file, "Classifier %3d/%3d. Nodes: %3d\tAlpha = %5.2lf\n", j + 1, R->cascade->ensemble[i].s, parm->base_learner == C45 ?
                        TreeSize(R->cascade->ensemble[i].trees[j]) : SVMSize(R->cascade->ensemble[i].lsvm[j]), R->cascade->ensemble[i].alpha[j]);
            }
        }

        parm->r_method = NoOne;
    }

    if (parm->stacking_methods & stPWM) {
        GetPositionWeightMatrix(parm, R, positives, n_positives, negatives, n_negatives, length);
        inputs++;
    }

    if (parm->stacking_methods & stNN) {
        TrainNeuralNetwork(parm, R, positives, n_positives, negatives, n_negatives, length, parm->nn_hidden_nodes);
        inputs++;
    }

    // Train stacking classifier
    // NOW: SVM with CV parameters
    // Create data
    scalar **data;
    int *labels, *mask;

    MALLOC(mask, int, inputs, "TrainStackerAndItsClassifiers: S");
    for (int i = 0; i < inputs; i++)
        mask[i] = 1;
    MALLOC(labels, int, n_positives + n_negatives, "TrainStackerAndItsClassifiers: labels");
    MALLOC(data, scalar *, n_positives + n_negatives, "TrainStackerAndItsClassifiers: data");

    // Positive instances and negative instances are interlaced
    if (n_positives != n_negatives)
        OnErrorExit(1, "parm->stacking must be used with undersampling");
    for (int i = 0; i < n_positives + n_negatives; i += 2) {
        MALLOC(data[i], scalar, inputs, "TrainStackerAndItsClassifiers: data[i]");
        labels[i] = POSITIVE;

        // Outputs of the classifiers
        int index = 0;
        if (parm->stacking_methods & stSTOP) {
            data[i][index] = GetSequenceScore(positives[i/2], length, R->scores, R->n_positions, 0, parm->reverse_stop_codon);
            index++;
        }

        if (parm->stacking_methods & stC45) {
            data[i][index] = C45Output(R->tree, positives[i/2], length);
            index++;
        }

        if (parm->stacking_methods & stSVM) {
            data[i][index] = SVMOutput(R->lsvm, positives[i/2], length, NULL, parm->use_salzberg_method, NULL);
            index++;
        }

        if (parm->stacking_methods & stSVMSTR) {
            data[i][index] = SVMOutput(R->lsvm_string, positives[i/2], length, NULL, parm->use_salzberg_method, NULL);
            index++;
        }

        if (parm->stacking_methods & stKNN) {
            data[i][index] = ClassifySequencekNN(R, positives[i/2], length, parm->knn_weight_by_distance);
            index++;
        }

        if (parm->stacking_methods & stEasy) {
            data[i][index] = MultiEnsembleOutput(R->easy, positives[i/2], length, parm->base_learner, parm->use_salzberg_method);
            index++;
        }

        if (parm->stacking_methods & stCascade) {
            data[i][index] = MultiEnsembleOutput(R->cascade, positives[i/2], length, parm->base_learner, parm->use_salzberg_method);
            index++;
        }

        if (parm->stacking_methods & stPWM) {
            data[i][index] = ScorePositionWeightMatrix(R, positives[i/2], length);
            index++;
        }

        if (parm->stacking_methods & stNN) {
            data[i][index] = NeuralNetworkOutput(R, positives[i/2], length);
            index++;
        }

        // Negative instances
        MALLOC(data[i + 1], scalar, inputs, "TrainStackerAndItsClassifiers: data[i + i]");
        labels[i + 1] = NEGATIVE;

        // Outputs of the classifiers
        index = 0;
        if (parm->stacking_methods & stSTOP) {
            data[i + 1][index] = GetSequenceScore(negatives[i/2], length, R->scores, R->n_positions, 0, parm->reverse_stop_codon);
            index++;
        }

        if (parm->stacking_methods & stC45) {
            data[i + 1][index] = C45Output(R->tree, negatives[i/2], length);
            index++;
        }

        if (parm->stacking_methods & stSVM) {
            data[i + 1][index] = SVMOutput(R->lsvm, negatives[i/2], length, NULL, parm->use_salzberg_method, NULL);
            index++;
        }

        if (parm->stacking_methods & stSVMSTR) {
            data[i + 1][index] = SVMOutput(R->lsvm_string, negatives[i/2], length, NULL, parm->use_salzberg_method, NULL);
            index++;
        }

        if (parm->stacking_methods & stKNN) {
            data[i + 1][index] = ClassifySequencekNN(R, negatives[i/2], length, parm->knn_weight_by_distance);
            index++;
        }

        if (parm->stacking_methods & stEasy) {
            data[i + 1][index] = MultiEnsembleOutput(R->easy, negatives[i/2], length, parm->base_learner, parm->use_salzberg_method);
            index++;
        }

        if (parm->stacking_methods & stCascade) {
            data[i + 1][index] = MultiEnsembleOutput(R->cascade, negatives[i/2], length, parm->base_learner, parm->use_salzberg_method);
            index++;
        }

        if (parm->stacking_methods & stPWM) {
            data[i + 1][index] = ScorePositionWeightMatrix(R, negatives[i/2], length);
            index++;
        }

        if (parm->stacking_methods & stNN) {
            data[i + 1][index] = NeuralNetworkOutput(R, negatives[i/2], length);
            index++;
        }
    }

    // parm->scale?
    ScaleData(data, n_positives + n_negatives, inputs, -1.0, +1.0, &(R->a), &(R->b), parm->scale);

    if (parm->cv_parameters) {
        if (parm->stacker == SVM)
            CrossValidatedSVMRealParameters(parm, data, n_positives + n_negatives, inputs, labels);
        else
            R->stacking_k = CrossValidatedkNNRealParameters(parm, data, n_positives + n_negatives, inputs, labels);
    }

    R->stacking_classifers = inputs;
    if (parm->stacker == SVM)
        R->stacking = TrainSVMClassifierRealInputs(parm, data, mask, n_positives + n_negatives, labels, inputs);
    else if (parm->stacker == kNN) {
        // Copy prototypes
        R->n_protos = n_positives + n_negatives;
        MALLOC(R->protos, scalar *, n_negatives + n_positives, "TrainStackerAndItsClassifiers: R->protos");
        MALLOC(R->proto_labels, int, n_negatives + n_positives, "TrainStackerAndItsClassifiers: R->proto_labels");
        for (int i = 0; i < R->n_protos; i++) {
            R->proto_labels[i] = labels[i];
            MALLOC(R->protos[i], scalar, inputs, "TrainStackerAndItsClassifiers: R->protos[i]");
            for (int j = 0; j < inputs; j++)
                R->protos[i][j] = data[i][j];
        }
    } else
        OnErrorExit(1, "Bad stacker");


    // If a sensitivity level is required the threshold is adjusted, otherwise it is 0.0
    if (parm->sensitivity > 0.0) {
        // Get threshold to obtain desired sensitivity = TP/(TP+FN)
        // Get all outputs and ...
        scalar *sc;
        ALLOCA(sc, scalar, n_positives, "TrainStackerAndItsClassifiers: sc");
        for (int i = 0; i < n_positives; i++)
            sc[i] = StackingOutput(parm, R, positives[i], length, 0);

        // ...sort them
        qsort(sc, n_positives, sizeof(scalar), CompareDouble);

        // Get threshold for desired sensitivity
        R->threshold = sc[(int)(n_positives * (1.0 - parm->sensitivity)) ];
    } else
        R->threshold = 0.0;

    for (int i = 0; i < n_positives + n_negatives; i++)
        FREE(data[i]);
    FREE(data);
    FREE(labels);

    // Return whether training data should be released
    return training_data_released;
}

scalar StackingOutput(const t_Parameters *parm, t_Recognizer *rec, char *seq, int s_length, int tis_location)
{
    scalar instance[rec->stacking_classifers];

    // Create data: Outputs of the classifiers
    int index = 0;
    if (parm->stacking_methods & stSTOP) {
        instance[index] = GetSequenceScore(seq, s_length, rec->scores, rec->n_positions, tis_location, parm->reverse_stop_codon);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stC45) {
        instance[index] = C45Output(rec->tree, seq, s_length);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stSVM) {
        instance[index] = SVMOutput(rec->lsvm, seq, s_length, NULL, parm->use_salzberg_method, NULL);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stSVMSTR) {
        instance[index] = SVMOutput(rec->lsvm_string, seq, s_length, NULL, parm->use_salzberg_method, NULL);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stKNN) {
        instance[index] = ClassifySequencekNN(rec, seq, s_length, parm->knn_weight_by_distance);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stEasy) {
        instance[index] = MultiEnsembleOutput(rec->easy, seq, s_length, parm->base_learner, parm->use_salzberg_method);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stCascade) {
        instance[index] = MultiEnsembleOutput(rec->cascade, seq, s_length, parm->base_learner, parm->use_salzberg_method);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stPWM) {
        instance[index] = ScorePositionWeightMatrix(rec, seq, s_length);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacking_methods & stNN) {
        instance[index] = NeuralNetworkOutput(rec, seq, s_length);
        instance[index] = rec->a[index] * instance[index] + rec->b[index];
        index++;
    }

    if (parm->stacker == SVM)
        return SVMRealOutput(rec->stacking, instance, index);
    else
        return kNearestNeighbour(instance, rec->protos, rec->n_protos, rec->stacking_classifers, rec->stacking_k, rec->proto_labels, parm->knn_weight_by_distance);
}

// Parameter cross-validation for SVM real

#define SVM_PARAM_SETS  21

void CrossValidatedSVMRealParameters(t_Parameters *parm, scalar  **data, int instances, int inputs, int *labels)
{
    scalar best_so_far, sensitivity, specificity, TP, TN, FP, FN, G, **train, **test;
    int best_index = 0, *test_label, *train_label, S[inputs];

    // Sets of parameters
    scalar svm_p[SVM_PARAM_SETS][3] = {{2.0, 0.0001, 0.1}, {2.0, 0.0001, 1.0}, {2.0, 0.0001, 10.0}, {2.0, 0.001, 0.1}, {2.0, 0.001, 1.0}, {2.0, 0.001, 10.0}, {2.0, 0.01, 0.1},
        {2.0, 0.01, 1.0}, {2.0, 0.01, 10.0}, {2.0, 0.1, 0.1}, {2.0, 0.1, 1.0}, {2.0, .1, 10.0}, {2.0, 1.0, 0.1}, {2.0, 1.0, 1.0}, {2.0, 1.0, 10.0}, {2.0, 10.0, 0.1}, {2.0, 10.0, 1.0},
        {2.0, 10.0, 10.0}, {0.0, 0.0, 0.1}, {0.0, 0.0, 1.0}, {0.0, 0.0, 10.0}};

    // Do validation runs
    best_so_far = -1.0;

    // Auxiliary data
    MALLOC(train, scalar *, instances, "CrossValidatedSVMRealParameters: train");
    MALLOC(test, scalar *, instances, "CrossValidatedSVMRealParameters: test");
    MALLOC(train_label, int, instances, "CrossValidatedSVMRealParameters: train_label");
    MALLOC(test_label, int, instances, "CrossValidatedSVMRealParameters: test_label");
    for (int i = 0; i < inputs; i++)
        S[i] = 1;

    // G-measure is used to parameter setting
    for (int i = 0; i < SVM_PARAM_SETS; i++) {
        struct svm_model *lsvm = NULL;

        parm->kernel = (int) svm_p[i][0];
        parm->gamma = svm_p[i][1];
        parm->c = svm_p[i][2];

        // Train and evaluate classifier
        int test_size;
        test_size = instances / parm->cv_folds;

        TN = TP = FP = FN = 0.0;

        for (int fold = 0; fold < parm->cv_folds; fold++) {
            // Create train and test sets
            int index = test_size * fold;
            for (int j = 0; j < test_size; j++, index = (index + 1) % instances) {
                test[j] = data[index];
                test_label[j] = labels[index];
            }
            for (int j = 0; j < instances - test_size; j++, index = (index + 1) % instances) {
                train[j] = data[index];
                train_label[j] = labels[index];
            }

            // Train and evaluate classifier
            // Learn model
            lsvm = TrainSVMClassifierRealInputs(parm, train, S, instances - test_size, train_label, inputs);

            // All instances
            for (int j = 0; j < test_size; j++)
                if (test_label[j] == POSITIVE) {
                    // Positive instances
                    if (SVMRealOutput(lsvm, test[j], inputs) > 0.0)
                        TP++;
                    else
                        FN++;
                } else {
                    // Negative instances
                    if (SVMRealOutput(lsvm, test[j], inputs) <= 0.0)
                        TN++;
                    else
                        FP++;
                }

            svm_free_and_destroy_model(&lsvm);
        }
        // Better than current best?
        sensitivity = TP / (TP + FN);
        specificity = TN / (TN + FP);
        G = pow(sensitivity * specificity, 0.5);
        if (G > best_so_far) {
            best_so_far = G;
            best_index = i;
        } else if (G == best_so_far) {
            // In case of a tie choose higher parm->c
            if (svm_p[best_index][2] < svm_p[i][2])
                best_index = i;
        }

        if (parm->verbose > 1)
            fprintf(parm->out_file, "parm->kernel = %d\tC = %lf\tgamma = %lf\tAccuracy: %lf\n", parm->kernel, parm->c, parm->gamma, G);
    }

    // Accuracy in percentage
    parm->kernel = (int) svm_p[best_index][0];
    parm->gamma = svm_p[best_index][1];
    parm->c = svm_p[best_index][2];
    fprintf(parm->out_file, "Best parameters (accuracy: %lf): parm->kernel = %d\tC = %lf\tGamma: %lf\n", best_so_far, parm->kernel, parm->c, parm->gamma);

    FREE(train);
    FREE(test);
    FREE(train_label);
    FREE(test_label);
}

void SaveStackingOutputs(const t_Parameters *parm, char *name, t_Recognizer *R, char **positives, int n_positives, char **negatives, int n_negatives, int length, int tis_location)
{
    FILE *save;
    char *seq, *class;

    if ((save = fopen(name, "w")) == NULL)
        OnErrorExit(1, "Cannot open file for saving outputs");

    for (int i = 0; i < n_positives + n_negatives; i++) {
        if (i < n_positives) {
            seq = positives[i];
            class = "+1";
        } else {
            seq = negatives[i - n_positives];
            class = "-1";
        }

        // Stacking components
        int index = 0;
        if (parm->stacking_methods & stSTOP) {
            fprintf(save, "%7.4lf ", R->a[index] * GetSequenceScore(seq, length, R->scores, R->n_positions, tis_location, parm->reverse_stop_codon) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stC45) {
            fprintf(save, "%7.4lf ", R->a[index] * C45Output(R->tree, seq, length) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stSVM) {
            fprintf(save, "%7.4lf ", R->a[index] * SVMOutput(R->lsvm, seq, length, NULL, parm->use_salzberg_method, NULL) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stSVMSTR) {
            fprintf(save, "%7.4lf ", R->a[index] * SVMOutput(R->lsvm_string, seq, length, NULL, parm->use_salzberg_method, NULL) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stKNN) {
            fprintf(save, "%7.4lf ", R->a[index] * ClassifySequencekNN(R, seq, length, parm->knn_weight_by_distance) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stEasy) {
            fprintf(save, "%7.4lf ", R->a[index] * MultiEnsembleOutput(R->easy, seq, length, parm->base_learner, parm->use_salzberg_method) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stCascade) {
            fprintf(save, "%7.4lf ", R->a[index] * MultiEnsembleOutput(R->cascade, seq, length, parm->base_learner, parm->use_salzberg_method) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stPWM) {
            fprintf(save, "%7.4lf ", R->a[index] * ScorePositionWeightMatrix(R, seq, length) + R->b[index]);
            index++;
        }

        if (parm->stacking_methods & stNN) {
            fprintf(save, "%7.4lf ", R->a[index] * NeuralNetworkOutput(R, seq, length) + R->b[index]);
            index++;
        }

        // Stacking output
        fprintf(save, "%7.4lf %s\n", StackingOutput(parm, R, seq, length, tis_location), class);
    }

    fclose(save);
}
