/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <string.h>
#include "type.h"
#include "macros.h"

// TIS recognition method based on stop codon weight matrix described in:
// Y. Saeys, T. Abeel, S. Degroeve and Y. Van de Peer, "Translation initiation side prediction on a genomic scale: beauty in simplicity", Bioinformatics, vol. 23, pp. i418-i423.

void StopCodonMethod(char **pos_seqs, int n_seqs, int length, scalar **scores, int *n_scores, scalar *threshold, bool reverse, scalar sensitivity)
{
    int CompareDouble(const void *, const void *);

    const char *stop[] = {"TAG", "TAA", "TGA"};

    // Discount ATG itself and count codons
    *n_scores = (length - 3)/3;

    // Allocate memory
    MALLOC(*scores, scalar, *n_scores + 1, "StopCodonMethod: *scores");

    // Use pseudo-counts
    for (int i = 0; i < *n_scores; i++)
        (*scores)[i] = 1.0;

    // Calculate scores
    for (int i = 0; i < n_seqs; i++) {
        int pos = 0;
        if (!reverse) {
            // Look for stop codons
            for (int j = 3; j < length && pos < *n_scores; j += 3, pos++)
                if (!strncmp((const char *) pos_seqs[i]+j, stop[0], 3) ||  !strncmp((const char *) pos_seqs[i]+j, stop[1], 3) ||
                    !strncmp((const char *) pos_seqs[i]+j, stop[2], 3)) {
                    (*scores)[pos]++;
                    break;
                }
        } else
            // Look for stop codons
            for (int j = length - 6; j >= 0 && pos < *n_scores; j -= 3, pos++)
                if (!strncmp((const char *) pos_seqs[i]+j, stop[0], 3) ||  !strncmp((const char *) pos_seqs[i]+j, stop[1], 3) ||
                    !strncmp((const char *) pos_seqs[i]+j, stop[2], 3)) {
                    (*scores)[pos]++;
                    break;
                }
    }

    // Get cumulative probabilities
    // Use pseudo-counts
    (*scores)[0] /= n_seqs + *n_scores;
    for (int i = 1; i < *n_scores; i++)
        (*scores)[i] = (*scores)[i-1] + (*scores)[i]/(n_seqs + *n_scores);

    // Transform scores to: pred_stop(z) = ln(prob(z is an in-frame stop codon downstream))
    for (int i = 0; i < *n_scores; i++)
        (*scores)[i] = log((*scores)[i]);

    // Get threshold to obtain desired sensitivity = TP/(TP+FN)
    // Get all scores and ...
    scalar *sc;
    ALLOCA(sc, scalar, n_seqs, "StopCodonMethod: sc");
    for (int i = 0; i < n_seqs; i++)
        sc[i] = GetSequenceScore(pos_seqs[i], length, *scores, *n_scores, 0, reverse);

    // ...sort them
    qsort(sc, n_seqs, sizeof(scalar), CompareDouble);

    // Get threshold for desired sensitivity
    if (sensitivity > 0.0)
        *threshold = sc[(int)(n_seqs * (1.0 - sensitivity)) ];
    else    // Use 80% as default
        *threshold = sc[(int)(n_seqs * 0.2) ];
}

scalar GetSequenceScore(char *seq, int s_length, scalar *scores, int n_scores, int tis_loc, bool reverse)
{
    int pos = 0;
    const char *stop[] = {"TAG", "TAA", "TGA"};

    if (tis_loc)
        OnErrorExit(1, "Not yet implemented");

    // Look for stop codons
    if (!reverse) {
        for (int j = 3; j < s_length && pos < n_scores; j += 3, pos++)
            if (!strncmp((const char *) seq+j, stop[0], 3) || !strncmp((const char *) seq+j, stop[1], 3) || !strncmp((const char *) seq+j, stop[2], 3))
                break;
    } else
        for (int j = s_length - 6; j >= 0 && pos < n_scores; j -= 3, pos++)
            if (!strncmp((const char *) seq+j, stop[0], 3) || !strncmp((const char *) seq+j, stop[1], 3) || !strncmp((const char *) seq+j, stop[2], 3))
                break;

    // If no stop codon return last value
    if (pos == n_scores)
        pos--;

    return (scores[pos]);
}
