/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Functions that deal with support vector machines
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "type.h"
#include "macros.h"
#include "svm.h"

// VERY IMPORTANT: The library assigns the labels of classes as they are found in the training

void TrainSVMClassifier(const t_Parameters *parm, struct svm_model **svm, char **pseqs, int n_pseqs, int length, char **nseqs, int n_nseqs, bool releasedata, scalar *threshold, scalar **sc_positives,
                        scalar **sc_negatives, struct spectrum_kernel_info *spectrum)
{
    int CompareDouble(const void *, const void *);

    extern int dna_index[];

    struct svm_model *smodel;       // SVM model
    struct svm_problem prob;        // SVM data
    struct svm_parameter param;     // SVM parameters

    int in;
    const char *error_msg;

    // Parameter values
    param.svm_type = 0;
    param.kernel_type = parm->kernel;
    param.degree = parm->degree; // String kernel WD or polynomial kernel
    // default 1/k
    if (parm->gamma == 0.0)
        param.gamma = 1.0/length; // 1/k
    else
        param.gamma = parm->gamma;    // Also sigma for WDS kernel and l for LI
    param.coef0 = 0.0;
    param.cache_size = 2000;        // Large for large problems
    param.C = parm->c;
    param.eps = 0.1;      // Large: 0.1, small: 0.001
    param.p = 0.1;
    param.shrinking = 0;    // Epsilon small: 1, epsilon large: 0
    param.nr_weight = 0;
    param.weight_label = NULL;
    param.weight = NULL;
    param.nu = 0.1;
    param.probability = 0;

    // Create problem structure for the classifier
    // Allocate memory
    prob.l = n_pseqs + n_nseqs;

    MALLOC(prob.x, struct svm_node *, prob.l, "TrainSVMClassifier: prob.x");
    MALLOC(prob.y, double, prob.l, "TrainSVMClassifier: prob.y");

    // Create data for SVM
    int count = 0;

    // Positive sequences
    for (int j = 0; j < n_pseqs; j++) {
        MALLOC(prob.x[count], struct svm_node, length + 1, "TrainSVMClassifier: prob.x[count]");

        // Set inputs
        in = 0;
        if (parm->use_salzberg_method) {
            for (int i = 0; i < length; i++) {
                prob.x[count][in].index = i;
                prob.x[count][in].value = sc_positives[j][i];
                in++;
            }
        } else if (parm->kernel != WD && parm->kernel != WDS && parm->kernel != LI && param.kernel_type != SPECTRUM && param.kernel_type != MISMATCH && param.kernel_type != SMITHWATERMAN && 
                    param.kernel_type != WILDCARD) {
            // Codify A -> 1000, T -> 0100, G - > 0010, C -> 0001, disregard the rest
            for (int i = 0; i < length; i++)
                if (dna_index[(int) pseqs[j][i]] != -1) {
                    prob.x[count][in].index = i * 4 + dna_index[(int) pseqs[j][i]];
                    prob.x[count][in].value = 1.0;
                    in++;
                }
        } else {
            for (int i = 0; i < length && i < strlen(pseqs[j]); i++) {
                prob.x[count][in].index = i;
                prob.x[count][in].value = dna_index[(int) pseqs[j][i]];
                in++;
            }
        }
        prob.x[count][in].index = -1;

        // Set class label
        prob.y[count] = POSITIVE;
        
        if (param.kernel_type == SPECTRUM || param.kernel_type == MISMATCH || param.kernel_type == WILDCARD)
            prob.x[count][0].spectrum = &spectrum[j];
        else
            prob.x[count][0].spectrum = NULL;

        // Add an instance
        count++;
    }

    // Negative sequences
    for (int j = 0; j < n_nseqs; j++) {
        MALLOC(prob.x[count], struct svm_node, length + 1, "TrainSVMClassifier: prob.x[count]");

        // Set inputs
        in = 0;
        if (parm->use_salzberg_method) {
            for (int i = 0; i < length; i++) {
                prob.x[count][in].index = i;
                prob.x[count][in].value = sc_negatives[j][i];
                in++;
            }
        } else if (parm->kernel != WD && parm->kernel != WDS && parm->kernel != LI && param.kernel_type != SPECTRUM && param.kernel_type != MISMATCH && param.kernel_type != SMITHWATERMAN &&
                    param.kernel_type != WILDCARD) {
            // Codify A -> 1000, T -> 0100, G - > 0010, C -> 0001, disregard the rest
            for (int i = 0; i < length; i++)
                if (dna_index[(int) nseqs[j][i]] != -1) {
                    prob.x[count][in].index = i * 4 + dna_index[(int) nseqs[j][i]];
                    prob.x[count][in].value = 1.0;
                    in++;
                }
        } else {
            for (int i = 0; i < length && i < strlen(nseqs[j]); i++) {
                prob.x[count][in].index = i;
                prob.x[count][in].value = dna_index[(int) nseqs[j][i]];
                in++;
            }
        }

        prob.x[count][in].index = -1;

        // Set class label
        prob.y[count] = NEGATIVE;
        
        if (param.kernel_type == SPECTRUM || param.kernel_type == MISMATCH || param.kernel_type == WILDCARD)
            prob.x[count][0].spectrum = &spectrum[j+n_pseqs];
        else
            prob.x[count][0].spectrum = NULL;

        // Add an instance
        count++;

        // Release data
        if (releasedata) {
            if (nseqs != NULL) {
                FREE(nseqs[j]);
            } else
                FREE(sc_negatives[j]);
        }
    }

    if (releasedata) {
        if (nseqs != NULL) {
            FREE(nseqs);
        } else
            FREE(sc_negatives);
    }

    // Check parameters
    error_msg = svm_check_parameter(&prob, &param);
    if (error_msg)
        OnErrorExit(1, (char *) error_msg);

    // Create classifier
    smodel = svm_train(&prob, &param);

    // Free memory
    char *template = tempnam(".", NULL);
    svm_save_model(template, smodel);
    svm_free_and_destroy_model(&smodel);
    svm_destroy_param(&param);

    if ((smodel = svm_load_model(template)) == NULL) {
        remove(template);
        OnErrorExit(1, "Error loading SVM model");
    }
    remove(template);

    if ((parm->kernel == WD || parm->kernel == WDS || parm->kernel == LI) && parm->verbose > 1)
        fprintf(parm->out_file, "WD(S) degree = %d\n", smodel->param.degree);

    // Deallocate memory
    for (int i = 0; i < prob.l; i++)
        FREE(prob.x[i]);
    FREE(prob.x);
    FREE(prob.y);

    // If a sensitivity level is required the threshold is adjusted, otherwise is 0.0
    if (parm->sensitivity > 0.0 && threshold != NULL) {
        // Get threshold to obtain desired sensitivity = TP/(TP+FN)
        // Get all outputs and ...
        scalar *sc;
        ALLOCA(sc, scalar, n_pseqs, "TrainSVMClassifier: sc");
        for (int i = 0; i < n_pseqs; i++) {
            sc[i] = SVMOutput(smodel, pseqs != NULL ? pseqs[i] : NULL, length, sc_positives != NULL ? sc_positives[i] : NULL, parm->use_salzberg_method, &spectrum[i]);
            if (releasedata) {        // Release training data to save memory
                if (pseqs) {
                    FREE(pseqs[i]);
                } else
                    FREE(sc_positives[i]);
            }
        }
        // ...sort them
        qsort(sc, n_pseqs, sizeof(scalar), CompareDouble);

        // Get threshold for desired sensitivity
        *threshold = sc[(int)(n_pseqs * (1.0 - parm->sensitivity)) ];
    } else if (threshold != NULL)
        *threshold = 0.0;

    if (threshold != NULL)
        fprintf(parm->out_file, "SVM threshold: %lf\n", *threshold);

    if (releasedata) {
        FREE(pseqs);
    }

    *svm = smodel;
}

// Return number of support vectors
int SVMSize(struct svm_model *smodel)
{
    return smodel->l;
}
