/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <alloca.h>
#include "type.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"

struct svm_model *TrainSVMClassifierRealInputs(const t_Parameters *parm, scalar **data, int *S, int instances, int *label, int inputs) {
    struct svm_model *smodel;  // SVM model
    struct svm_problem prob;   // SVM data
    struct svm_parameter param;     // SVM parameters

    int in;
    const char *error_msg;

    // Count selected features
    int features = 0;
    for (int i = 0; i < inputs; i++)
        if (S[i])
            features++;

    // If no features do nothing
    if (!features)
        return NULL;

    // Parameter values
    param.svm_type = 0;
    param.kernel_type = parm->kernel;
    param.degree = 3;
    // default 1/k
    if (parm->gamma == 0.0)
        param.gamma = 1.0 / inputs;
    else
        param.gamma = parm->gamma;    // 1/k
    param.coef0 = 0.0;
    param.cache_size = 100000;
    param.C = parm->c;
    param.eps = 0.01;
    param.p = 0.1;
    param.shrinking = 1;
    param.nr_weight = 0;
    param.weight_label = NULL;
    param.weight = NULL;
    param.nu = 0.5;
    param.probability = 0;

    // Create problem structure for the classifier
    // Allocate memory
    prob.l = instances;
    MALLOC(prob.x, struct svm_node *, prob.l, "TrainSVMClassifier: prob.x");
    MALLOC(prob.y, double, prob.l, "TrainSVMClassifier: prob.y");
    for (int i = 0; i < prob.l; i++)
        MALLOC(prob.x[i], struct svm_node, features + 1, "TrainSVMClassifier: prob.x[i]");

    // Create data for SVM taking into account selected instances
    // Original data
    for (int j = 0; j < instances; j++) {
        // Set inputs
        in = 0;
        for (int i = 0; i < inputs; i++)
            if (S[i] && data[j][i] * data[j][i] > ZERO2) {
                prob.x[j][in].index = i;
                prob.x[j][in].value = data[j][i];
                in++;
            }
        prob.x[j][in].index = -1;

        // Set class label
        prob.y[j] = label[j];
    }

    // Check parameters
    error_msg = svm_check_parameter(& (prob), &param);
    if (error_msg)
        OnErrorExit(1, (char *) error_msg);

    // Create classifier
    smodel = svm_train(& (prob), &param);

    // Free memory
    char *template = tempnam(".", NULL);
    svm_save_model(template, smodel);
    svm_free_and_destroy_model(&smodel);
    svm_destroy_param(&param);

    if ((smodel = svm_load_model(template)) == NULL) {
        remove(template);
        OnErrorExit(1, "Error loading SVM model");
    }
    remove(template);

    // De-allocate memory
    for (int i = 0; i < prob.l; i++)
        FREE(prob.x[i]);
    FREE(prob.x);
    FREE(prob.y);

    return smodel;
}

scalar SVMRealOutput(struct svm_model *smodel, scalar *instance, int inputs)
{
    struct svm_node svm_x[inputs + 1];
    int in;
    double dec_values;

    // Create description
    in = 0;
    for (int i = 0; i < inputs; i++)
        if (instance[i] * instance[i] > ZERO2) {
            svm_x[in].index = i;
            svm_x[in].value = instance[i];
            in++;
        }
    svm_x[in].index = -1;

    // Get output
    svm_predict_values(smodel, svm_x, &dec_values);

    // Return output
    return (scalar) dec_values;
}
