/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "type.h"
#include "macros.h"

// Implements numerical integration using trapezoid rule
// Reads a plain text file. Integrates each pair of columns.
// Left column is X, right column is Y

scalar TrapezoidIntegration(char *file, int points)
{
    scalar qtrap(scalar **, scalar, scalar, int);
    void ReadDataForIntegration(char *, scalar ** *, int, int);

    scalar ***data;

    // Read file
    MALLOC(data, scalar **, 1, "TrapezoidIntegration: data");
    MALLOC(data[0], scalar *, points, "TrapezoidIntegration: data[0]");
    for (int j = 0; j < points; j++)
        MALLOC(data[0][j], scalar, 2, "TrapezoidIntegration: data[j]");

    // Read data
    ReadDataForIntegration(file, data, 1, points);

    // Do integration
    return qtrap(data[0], 0.0, 1.0, points);
}


// Interpolate value
scalar f(scalar x, scalar **data, int n)
{
    scalar x1, x2, y1, y2, y;

    // If outside limits return 0
    if (x < data[0][0] && x > data[n-1][0])
        return 0.0;

    // Otherwise linear interpolation
    // Get points to interpolate
    int i = 0;
    while (x > data[i][0])
        i++;

    // Check especial cases
    if (i == 0)
        return data[0][1];

    x1 = data[i-1][0];
    y1 = data[i-1][1];

    // Check especial cases
    if (x == x1)
        y = y1;
    else {
        x2 = data[i][0];
        y2 = data[i][1];

        if (x==x2)
            y=y2;
        // If x1=x2, return y1
        else if (x1==x2)
            y = y1;
        else {
            // Equation is: y = y1 + [(y2-y1)/(x2-x1)]·(x-x1) if x1 != x2
            y = y1 + ((y2-y1) / (x2-x1)) * (x-x1);
        }
    }

    return y;
}

scalar trapzd(scalar **data, scalar a, scalar b, int n, int nvalues)
{
    scalar  x, tnm, sum, del;
    static scalar s;
    int it, j;
    if (n == 1) {
        return (s = 0.5* (b-a) * (f(a, data, nvalues) + f(b, data, nvalues)));
    } else {
        for (it = 1,j = 1; j < n-1; j++)
            it <<= 1;
        tnm = it;
        del = (b - a) / tnm;   /*This is the spacing of points to be added. */
        x = a + 0.5 * del;
        for (sum = 0.0, j = 1; j <= it; j++, x += del)
            sum += f(x, data, nvalues);
        s = 0.5 * (s + (b-a) * sum/tnm);     /*This replaces s by its refined value.*/
        return s;
    }
}

void ReadDataForIntegration(char *file, scalar ***data, int functions, int n)
{
    FILE *f;

    if ((f = fopen(file, "r")) == NULL)
        OnErrorExit(1, "Cannot open AUC data file");

    for (int i = 0; i < n; i++) {
        for (int j = 0; j < functions; j++) {
            if (!fscanf(f, SCALAR_FORMAT, &data[j][i][0]))
                OnErrorExit(1, "Error reading data for integration file");
            if (!fscanf(f, SCALAR_FORMAT, &data[j][i][1]))
                OnErrorExit(1, "Error reading data for integration file");
        }
    }

    fclose(f);

}

#define EPS 1.0e-5
#define JMAX 20
scalar qtrap(scalar **data, scalar a, scalar b, int n)
{
    scalar trapzd(scalar **, scalar, scalar, int, int);

    int j;
    scalar s,olds;
    olds = -1.0e30;
    for (j=1; j<=JMAX; j++) {
        s=trapzd(data, a, b, j, n);
        if (j > 5)
            if (fabs(s-olds) < EPS*fabs(olds) || (s == 0.0 && olds == 0.0))
                return s;
        olds=s;
    }

    return 0.0;
}
