/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Types and prototypes of functions
#ifndef _TYPE_H
#define _TYPE_H
#include "types.i"
#include "common.h"
#include "svm.h"

// #define AUTOCHECK            // Code checks for errors, only for debugging process
// #define DEBUG                // Activate code for debugging purposes

#define LETTERS 17      // DNA alphabet

// Positive and negatives instances in TIS problem
typedef enum {
    NEGATIVE = 0, POSITIVE = 1
} t_SequenceClass;

// Type of recognition method
typedef enum {
    SVM, kNN, C45, STOP, EasyEnsembleEnsemble, BalanceCascadeEnsemble, PWM, NeuralNet, LearningFromBites, NoOne /* For stacking */
} t_RecognitionMethod;

typedef enum {
    nfNone, nfGEditing
} t_NoiseFilter;

// Ensemble of decision trees
typedef struct {
    int s;

    // Trees
    Tree *trees;

    // Support vector machine
    struct svm_model **lsvm;

    scalar *alpha, threshold;
} t_Ensemble;

typedef struct {
    int T;

    t_Ensemble *ensemble;
} t_MultiEnsemble;

// typedef enum {
//     fsNONE, fsRFE, fsGA
// } t_FSAlgorithm;

// Recognizer information structure
typedef struct {
    // Length of input sequence
    int length;

    // C4.5 decision tree
    Tree tree;

    // Easy ensemble
    t_MultiEnsemble *easy;

    // Balance cascade ensemble
    t_MultiEnsemble *cascade;

    // Ensemble for learning from bites
    t_Ensemble *ensemble;

    // Support vector machine
    struct svm_model *lsvm;

    // String kernel (only for stacking)
    struct svm_model *lsvm_string;

    // k-Nearest neighbor
    int k, n_poss, n_negs, pos_l, neg_l;
    char **poss, * *negs;

    // Stop codon method
    scalar *scores, threshold;
    int n_positions;

    // Position weight matrix
    scalar *M[18];

    // Neural network
    scalar **ih, *ho, A, B;
    int h_nodes;

    // Use for stacking classifiers
    scalar *a, *b;   // For scaling data
    int stacking_classifers;
    // SVM
    struct svm_model *stacking;
    // k-NN
    int stacking_k, n_protos;
    scalar **protos;
    int *proto_labels;
} t_Recognizer;

// Info of the performance of a classifier
typedef struct {
    scalar error, sensitivity, specificity, bio_specificity, TP, TN, FP, FN,  precision, F, G;
} t_RecInfo;

// Methods for dealing with imbalance two-class problems
typedef enum {
    iNone, iUndersampling, iSmote
} t_ImbalanceMethod;

// Structure for sorting features by their weight
typedef struct {
    scalar weight;
    scalar w;
    int index;
    int ranking;
} t_Feat;

// Global parameters
typedef struct {
    FILE *out_file;

    scalar sensitivity, gamma, c, coef0, smote_n, fs_percent_to_keep, fs_min_percent, eta, alpha, htan_a, htan_b, flat_spot, gafs_alpha, gafs_mutation, gafs_bit_mutation, gafs_set_probability,
        gafs_lsmutation, brute_force_threshold, sampling_ratio, *c45_weights;

    int seed, verbose, kernel, k, min_k, max_k, ensemble_t, cv_folds, degree, easy_s_i, smote_k, n_bites, bite_size, rfe_subset_size, offset, length, stacking_methods, nn_hidden_nodes, ge_k, n_epochs,
	epoch_size, offset_to_site, gafs_popsize, gafs_generations, left_window, right_window, max_length, ge_kprime, max_positive_instances, plus_l, minus_r, stepwise_max_attempts;

    short TRIALS;

    char datad[1000], mask_file[1000];

    t_RecognitionMethod r_method, base_learner, stacker;

    t_ImbalanceMethod imbalance_method;

    t_FSAlgorithm fsalgorithm;

    t_NoiseFilter noise_filter;

    bool prune_tree, soften_thresholds, eval_training, knn_weight_by_distance, exit_after_fs, cv_parameters, roc_training, write_auc_data, stacking, scale, use_protein_sequence,
        allow_only_5_letters, use_salzberg_method, reverse_stop_codon, do_not_check_sequences, allow_different_lengths;
} t_Parameters;


// PVM
#define SENT_BY_MASTER  1
#define SENT_BY_SLAVE   2

#define SEQUENCES_BLOCK     100 // Number of sequences sent to slaves each time
#define MAX_SEQUENCE_LENGTH 1100

// Methods to use for stacked classifiers
#define stSTOP          0x01
#define stC45           0x02
#define stSVM           0x04
#define stSVMSTR        0x08    // SVM with a string kernel
#define stEasy          0x10
#define stCascade       0x20
#define stKNN           0x40
#define stPWM           0x80
#define stNN            0x100

// Prototypes
// General purpose functions
void LoadGlobalParameters(t_Parameters *, char *);
void ReadFastaFile(char *, char ** *, int *, int *, scalar, bool, int, bool, int *, int, bool, int, int, scalar, bool, bool, char *, int *, char ** *, bool);
scalar HammingDistance(char *, char *, int/*, int*/);
int ConvertToProtein(char **, int, int);
void ReadRawFastaFile(char *, char **, int *, int);
int CountSequencesInFastaFile(char *);

// TIS recognition functions
scalar GetSequenceScore(char *, int, scalar *, int, int, bool);
t_SequenceClass ClassifySequence(const t_Parameters *, t_Recognizer *, char *, int, int, scalar *, struct spectrum_kernel_info *);
void EvaluateRecognizer(const t_Parameters *, t_Recognizer *, t_RecInfo *, char **, int, int, char **, int, int, scalar *, scalar *, scalar **, scalar **, struct spectrum_kernel_info *);
void StopCodonMethod(char **, int, int, scalar **, int *, scalar *, bool, scalar);
Tree TrainC45Classifier(const t_Parameters *, char **, int, int, char **, int, bool, scalar *);
scalar C45Output(Tree, char *, int);
void TrainSVMClassifier(const t_Parameters *, struct svm_model **, char **, int, int, char **, int, bool, scalar *, scalar **, scalar **, struct spectrum_kernel_info *);
scalar SVMOutput(struct svm_model *, char *, int, scalar *, bool, struct spectrum_kernel_info *);
void CrossValidatedParameters(t_Parameters *, char **, int, int, char **, int);
scalar ClassifySequencekNN(t_Recognizer *, char *, int, bool);
t_MultiEnsemble *TrainEasyEnsemble(const t_Parameters *, char **, char **, int, int, int, int, int);
scalar MultiEnsembleOutput(t_MultiEnsemble *, char *, int, t_RecognitionMethod, bool);
t_SequenceClass CascadeMultiEnsembleOutput(t_Ensemble *, char *, int);
scalar EnsembleOutput(t_Ensemble *, char *, int, t_RecognitionMethod, bool);
t_MultiEnsemble *TrainBalanceCascadeEnsemble(const t_Parameters *, char **, char **, int, int, int, int, int);
void SetkNNThreshold(const t_Parameters *, t_Recognizer *, FILE *);
void SetMultiEnsembleThreshold(const t_Parameters *, t_Recognizer *, char **, int, int);
scalar ClassifierOutput(const t_Parameters *, t_Recognizer *, char *, int, int, scalar *, struct spectrum_kernel_info *);
int SVMSize(struct svm_model *);
void SaveModel(const t_Parameters *, t_Recognizer *, char *);
void LoadModel(t_Parameters *, t_Recognizer *, char *);
void GetPositionWeightMatrix(const t_Parameters *, t_Recognizer *, char **, int, char **, int, int);
scalar ScorePositionWeightMatrix(t_Recognizer *, char *, int);
void ConvertToScalarSequence(char *, int, scalar *);
scalar NeuralNetworkOutput(t_Recognizer *, char *, int);
void TrainNeuralNetwork(const t_Parameters *, t_Recognizer *, char **, int, char **, int, int, int);

// Data
void ConvertToDinucleotideFrequency(char ** *, int, char ** *, int, scalar ** *, scalar ** *, int *, scalar ** *);
void RandomizeSequences(char **, int);

// Imbalance class methods
void UndersampleData(char ** *, int *, int, bool);
void SMOTEOverSampling(char ** *, int *, scalar, int, int, scalar **);
void ConstructValueDifferencesMatrix(scalar **, char **, int, char **, int, int);

// Feature selection
void RFE(const t_Parameters *, t_Feat *, int, scalar **, int, int *);
void ConvertToScalarData(char **, int, char **, int, int, scalar ** *, int **, int *, int *, int);
void CrossValidateFeatureSelection(t_Parameters *, char **, int, int, char **, int, t_Feat *, scalar);
struct svm_model *TrainSVMClassifierRealInputs(const t_Parameters *, scalar **, int *, int, int *, int);
void GAFeatureSelection(const t_Parameters *, int *, char **, int, char **, int, int, int, int,  scalar, scalar, scalar, struct spectrum_kernel_info *);
scalar EvaluateIndividual(const t_Parameters *, char **, int, int, char **, int, int *, t_RecognitionMethod, int, scalar, scalar *, scalar *, struct spectrum_kernel_info *);
struct svm_model *TrainSVMClassifierRealInpus(scalar **, int *, int, int *, int);
scalar **GetSVMW(struct svm_model *, int);
int CompareIndex(const void *, const void *);
int CompareWInv(const void *, const void *);

// String kernels
double WeightedDegreeKernel(const struct svm_node *, const struct svm_node *, int);

// Stacking
bool TrainStackerAndItsClassifiers(t_Parameters *, t_Recognizer *, char **, int, char **, int, int);
scalar SVMRealOutput(struct svm_model *, scalar *, int);
scalar StackingOutput(const t_Parameters *, t_Recognizer *, char *, int, int);
void CrossValidatedSVMRealParameters(t_Parameters *, scalar **, int, int, int *);
void SaveStackingOutputs(const t_Parameters *, char *, t_Recognizer *, char **, int, char **, int, int, int);
void ScaleData(scalar **, int, int, scalar, scalar, scalar **, scalar **, bool);
scalar kNearestNeighbour(scalar *, scalar **, int, int, int, int *, bool);
int CrossValidatedkNNRealParameters(const t_Parameters *, scalar **, int, int, int *);

// Output of results
void WriteStackingComponentsResults(const t_Parameters *, t_Recognizer *, char **, int, char **, int, int, char **, int, int, FILE *);
void WriteROCResults(const t_Parameters *, t_Recognizer *, t_RecognitionMethod, char **, int, char **, int, int, char **, int, scalar **, scalar **, FILE *);
scalar AUROCFromOutputs(scalar *, int, int, bool, char *, bool);
scalar AUPRCFromOutputs(scalar *, int, int, bool, char *, bool);

// PVM
int InitGuardian(int, int *, char *, char *);

// Learning from bites
t_Ensemble *TrainLearningFromBites(const t_Parameters *, char **, char **, int, int, int, int, int);

// Sequence compression
void CompressSequence(char **);
void UncompressSequence(char *, char *, int);
int CompressedSequenceLength(char *, int);

void AddModelsToFileName(char *, bool *, int);
int ConvertToNumberWithoutMask(int, int *, int);

int CompareDouble(const void *, const void *);
int CompareRanking(const void *, const void *);
int CompareW(const void *, const void *);
void Populate(int, char **, int, int, int, int, int, scalar **);
void RandomizeVector(int *, int);
scalar HyperbolicTangent(scalar, scalar, scalar);
int HammingDistanceInteger(int *, int *, int);
int OrderedSelection(scalar *, int, int);
void HuxCrossover(int *, int *, int *, int *, int);
bool RandomMutation(int *, int, scalar);
scalar LSMutation(const t_Parameters *, int *, int, char **, int, char **, int, t_RecognitionMethod, int, scalar, scalar *, scalar *, struct spectrum_kernel_info *);
int RouletteSelection(scalar *, int);
void SetBitPair(char *, int, int);
bool NextSetOfParameters(int, int [], int []);
void CopyFile(char *, char *);
scalar TrapezoidIntegration(char *, int);
void RemoveLF(char *);

// Noise filters
void GetKNeighbors(char *, char **, char **, int, int, int, int, t_SequenceClass *);
void GeneralizedEditingFilter (const t_Parameters *, char ***, char ***, int *, int *, int, int, int);

#endif // _TYPE_H
