/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
 * Input/Output functions for FASTA (or quasi fasta format)
 */
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include "common.h"
#include "macros.h"

/*
 * Read a data file. The structure of the file must be:
 * > ID
 * One sequence per line
 * Codification:
                A       0001
                T/U     0010
                G       0100
                C       1000
                R       0101 (A or G, purine)
                Y       1010 (T/U or C, pyrimidine)
                K       0110 (G or T/U, ketone)
                M       1001 (A or C, amino group)
                S       1100 (G or C, strong interaction)
                W       0011 (A or T, weak interaction)
                B       1110 (Not A)
                D       0111 (Not C)
                H       1011 (Not G)
                V       1101 (Not T, not U)
                N       0000 (Any)
                X       0000 (Masked)
                -       0000 (Gap)
 */

void ReadFastaData(scalar ***data, int **labels, int *instances, int *inputs, char *p_file, char *n_file, bool huge_dataset, int seq_length, int seq_offset, bool randomize, bool raw_inputs)
{
    FILE *f;
    int length;
    char line[10000], s[10000];

    // POSITIVE SEQUENCES
    // Opens the file and read number of instances
    if ((f = fopen(p_file, "rt")) == NULL)
        OnErrorExit(ERR_FILE, "ReadFastaData");

    // Count sequences
    int p_sequences = 0;

    if (!fgets(s, 10000, f))
        OnErrorExit(1, "ReadFastaData: Error reading FASTA file");
    while (!feof(f)) {
        if (s[0] == '>')   // New sequence
            p_sequences++;
        if (!fgets(s, 10000, f) && !feof(f))
            OnErrorExit(1, "ReadFastaData: Error reading FASTA file");
    }

    // Where to begin and end to read
    int start, end;
    end = length = strlen(s);
    start = seq_offset;

    // Check if smaller length
    if (seq_length == 0)
        seq_length = length - start;
    else if (length > seq_length)
        end = start + seq_length;

    // Check validity
    if (start >= end || end > length) {
        printf("Start: %d, End: %d, Length: %d\n", start, end, length);
        OnErrorExit(1, "ReadFastaData: Sequence length and offset do not match");
    }

    // Each nucleotide 4 inputs
    if (!raw_inputs)
        *inputs = 4 * seq_length;
    else
        *inputs = seq_length;

    // If huge dataset just count
    if (huge_dataset)
        goto next_negatives;

    // Memory allocation
    MALLOC(*data, scalar *, p_sequences, "ReadFastaData: *data");
    MALLOC(*labels, int, p_sequences, "ReadFastaData: *labels");
    for (int i = 0; i < p_sequences; i++)
        MALLOC((*data)[i], scalar, *inputs, "ReadFastaData: (*data)[i]");

    // Read data
    rewind(f);
    for (int i = 0; i < p_sequences; i++) {
        int n = 0;
        // Read "> ID"
        if (fgets(line, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadFastaData");

        // Read info
        if (fgets(line, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadFastaData");
        if (!raw_inputs) {
            // Codify
            for (int j = start; j < end; j++) {
                switch (toupper(line[j])) {
                    case 'A':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = 0;
                        (*data)[i][n+3] = 1;
                        break;
                    case 'T':
                    case 'U':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+3] = 0;
                        (*data)[i][n+2] = 1;
                        break;
                    case 'G':
                        (*data)[i][n] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n+1] = 1;
                        break;
                    case 'C':
                        (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = 1;
                        break;
                    case 'R':
                        (*data)[i][n] = (*data)[i][n+2] = 0;
                        (*data)[i][n+1] = (*data)[i][n+3] = 1;
                        break;
                    case 'Y':
                        (*data)[i][n+1] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+2] = 1;
                        break;
                    case 'K':
                        (*data)[i][n] = (*data)[i][n+3] = 0;
                        (*data)[i][n+1] = (*data)[i][n+2] = 1;
                        break;
                    case 'M':
                        (*data)[i][n+1] = (*data)[i][n+2] = 0;
                        (*data)[i][n] = (*data)[i][n+3] = 1;
                        break;
                    case 'S':
                        (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = 1;
                        break;
                    case 'W':
                        (*data)[i][n] = (*data)[i][n+1] = 0;
                        (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'B':
                        (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = 1;
                        break;
                    case 'D':
                        (*data)[i][n] = 0;
                        (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'H':
                        (*data)[i][n+1] = 0;
                        (*data)[i][n] = (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'V':
                        (*data)[i][n+2] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+3] = 1;
                        break;
                    case 'N':
                    case 'X':
                    case '-':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        break;
                }
                n += 4;
            }
        } else {
            // Original inputs
            for (int j = start; j < end; j++) {
                switch (toupper(line[j])) {
                    case 'A':
                        (*data)[i][n] = 1;
                        break;
                    case 'T':
                    case 'U':
                        (*data)[i][n] = 2;
                        break;
                    case 'G':
                        (*data)[i][n] = 3;
                        break;
                    case 'C':
                        (*data)[i][n] = 4;
                        break;
                    case 'R':
                    case 'Y':
                    case 'K':
                    case 'M':
                    case 'S':
                    case 'W':
                    case 'B':
                    case 'D':
                    case 'H':
                    case 'V':
                    case 'N':
                    case 'X':
                    case '-':
                        (*data)[i][n] = -1;     // DO NOT USE 0, it will break sequences
                        break;
                }
                n++;
            }
        }

        // Positive class
        (*labels) [i] = 1;
    }


next_negatives:

    fclose(f);

    // NEGATIVE SEQUENCES
    // Opens the file and read number of instances
    if ((f = fopen(n_file, "rt")) == NULL)
        OnErrorExit(ERR_FILE, "ReadFastaData");

    // Count sequences
    int n_sequences = 0;

    if (!fgets(s, 10000, f))
        OnErrorExit(1, "ReadFastaData: Error reading FASTA file");
    while (!feof(f)) {
        if (s[0] == '>')   // New sequence
            n_sequences++;
        if (!fgets(s, 10000, f) && !feof(f))
            OnErrorExit(1, "ReadFastaData: Error reading FASTA file");
    }

    // If huge dataset just count
    if (huge_dataset)
        goto next_end;

    // Memory allocation
    REALLOC(*data, scalar *, n_sequences + p_sequences, "ReadFastaData");
    REALLOC(*labels, int, n_sequences + p_sequences, "ReadFastaData");
    for (int i = p_sequences; i < p_sequences + n_sequences; i++)
        MALLOC((*data)[i], scalar, *inputs, "ReadFastaData");

    // Read data
    rewind(f);
    for (int i = p_sequences; i < p_sequences + n_sequences; i++) {
        int n = 0;
        // Read "> ID"
        if (fgets(line, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadFastaData");

        // Read info
        if (fgets(line, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadFastaData");
        // Codify
        if (!raw_inputs) {
            for (int j = start; j < end; j++) {
                switch (toupper(line[j])) {
                    case 'A':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = 0;
                        (*data)[i][n+3] = 1;
                        break;
                    case 'T':
                    case 'U':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+3] = 0;
                        (*data)[i][n+2] = 1;
                        break;
                    case 'G':
                        (*data)[i][n] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n+1] = 1;
                        break;
                    case 'C':
                        (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = 1;
                        break;
                    case 'R':
                        (*data)[i][n] = (*data)[i][n+2] = 0;
                        (*data)[i][n+1] = (*data)[i][n+3] = 1;
                        break;
                    case 'Y':
                        (*data)[i][n+1] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+2] = 1;
                        break;
                    case 'K':
                        (*data)[i][n] = (*data)[i][n+3] = 0;
                        (*data)[i][n+1] = (*data)[i][n+2] = 1;
                        break;
                    case 'M':
                        (*data)[i][n+1] = (*data)[i][n+2] = 0;
                        (*data)[i][n] = (*data)[i][n+3] = 1;
                        break;
                    case 'S':
                        (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = 1;
                        break;
                    case 'W':
                        (*data)[i][n] = (*data)[i][n+1] = 0;
                        (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'B':
                        (*data)[i][n+3] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = 1;
                        break;
                    case 'D':
                        (*data)[i][n] = 0;
                        (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'H':
                        (*data)[i][n+1] = 0;
                        (*data)[i][n] = (*data)[i][n+2] = (*data)[i][n+3] = 1;
                        break;
                    case 'V':
                        (*data)[i][n+2] = 0;
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+3] = 1;
                        break;
                    case 'N':
                    case 'X':
                    case '-':
                        (*data)[i][n] = (*data)[i][n+1] = (*data)[i][n+2] = (*data)[i][n+3] = 0;
                        break;
                }
                n += 4;
            }
        } else {
            // Original inputs
            for (int j = start; j < end; j++) {
                switch (toupper(line[j])) {
                    case 'A':
                        (*data)[i][n] = 1;
                        break;
                    case 'T':
                    case 'U':
                        (*data)[i][n] = 2;
                        break;
                    case 'G':
                        (*data)[i][n] = 3;
                        break;
                    case 'C':
                        (*data)[i][n] = 4;
                        break;
                    case 'R':
                    case 'Y':
                    case 'K':
                    case 'M':
                    case 'S':
                    case 'W':
                    case 'B':
                    case 'D':
                    case 'H':
                    case 'V':
                    case 'N':
                    case 'X':
                    case '-':
                        (*data)[i][n] = -1;     // DO NOT USE 0, it will break sequences
                        break;
                }
                n++;
            }
        }

        // Positive class
        (*labels) [i] = 0;
    }

    // Randomize to mix positive and negative instances?
    if (randomize)
        for (int i = 0; i < *instances; i++) {
            int r = RANDOM(0, *instances);

            // Exchange
            int i_swap = (*labels) [i];
            (*labels) [i] = (*labels) [r];
            (*labels) [r] = i_swap;

            scalar *s_swap = (*data)[i];
            (*data)[i] = (*data)[r];
            (*data)[r] = s_swap;
        }

next_end:

    fclose(f);

    *instances = n_sequences + p_sequences;
}

int ReadNextBlockOfFastaRawSequences(scalar **data, int *labels, char *p_file, char *n_file, int seq_length, int seq_offset, int block_size, bool first_time)
{
    static FILE *fp, *fn, *f;
    static int n_sequences = 0, p_sequences = 0, read_instances = 0;
    char s[10000];

    // Open the files and read number of instances
    if (first_time) {
        if ((fp = fopen(p_file, "rt")) == NULL)
            OnErrorExit(ERR_FILE, "ReadNextBlockOfFastaRawSequences");

        if (!fgets(s, 10000, fp))
            OnErrorExit(1, "ReadNextBlockOfFastaRawSequences: Error reading FASTA file");
        while (!feof(fp)) {
            if (s[0] == '>')   // New sequence
                p_sequences++;
            if (!fgets(s, 10000, fp) && !feof(fp))
                OnErrorExit(1, "ReadNextBlockOfFastaRawSequences: Error reading FASTA file");
        }

        if ((fn = fopen(n_file, "rt")) == NULL)
            OnErrorExit(ERR_FILE, "ReadNextBlockOfFastaRawSequences");

        if (!fgets(s, 10000, fn))
            OnErrorExit(1, "ReadNextBlockOfFastaRawSequences: Error reading FASTA file");
        while (!feof(fn)) {
            if (s[0] == '>')   // New sequence
                n_sequences++;
            if (!fgets(s, 10000, fn) && !feof(fn))
                OnErrorExit(1, "ReadNextBlockOfFastaRawSequences: Error reading FASTA file");
        }

        rewind(fp);
        rewind(fn);

        // Start reading positive sequences file
        f = fp;

        return p_sequences + n_sequences;
    }

    // Read data
    for (int i = 0; i < block_size && read_instances < p_sequences+n_sequences; i++, read_instances++) {
        // Change to negative instances file?
        if (read_instances == p_sequences) {
            f = fn;
            fclose(fp);
        }

        // Read "> ID"
        if (fgets(s, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadNextBlockOfFastaRawSequences");

        // Read info
        if (fgets(s, 10000, f) == NULL)
            OnErrorExit(ERR_READ, "ReadNextBlockOfFastaRawSequences");

        if (strlen(s) < seq_offset+seq_length)
            OnErrorExit(1, "Sequence too short");

        // Original inputs
        for (int j = seq_offset; j < seq_offset+seq_length; j++) {
            switch (toupper(s[j])) {
                case 'A':
                    data[i][j-seq_offset] = 1;
                    break;
                case 'T':
                case 'U':
                    data[i][j-seq_offset] = 2;
                    break;
                case 'G':
                    data[i][j-seq_offset] = 3;
                    break;
                case 'C':
                    data[i][j-seq_offset] = 4;
                    break;
                case 'R':
                case 'Y':
                case 'K':
                case 'M':
                case 'S':
                case 'W':
                case 'B':
                case 'D':
                case 'H':
                case 'V':
                case 'N':
                case 'X':
                case '-':
                    data[i][j-seq_offset] = -1;     // DO NOT USE 0, it will break sequences
                    break;
            }
        }

        // Positive class
        labels[i] = read_instances < p_sequences ? 1 : 0;
    }

    // Finished?
    if (read_instances == p_sequences+n_sequences)
        fclose(fn);

    // Return read sequences so far
    return read_instances;
}
